/*
 * Messages for OpenMoko
 * Copyright (C) 2007 Milko Krachounov <exabyte@3mhz.net>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */


#ifndef MESSAGES_ACCESS_H
#define MESSAGES_ACCESS_H
#include <sys/types.h>

/* Default folders */
enum {
	FOLDER_ROOT = 0,
	FOLDER_INBOX,
	FOLDER_OUTBOX,
	FOLDER_SENT,
	FOLDER_DRAFTS,
	FOLDER_TEMPLATES
};

/* Folder flags, used for choosing icons, among everything else */
enum {
	FOLDER_FLAG_DEFAULT = 0,
	FOLDER_FLAG_ACCOUNT = 1,
	FOLDER_FLAG_OUTGOING = 2,
	FOLDER_FLAG_DRAFTS = 4,
	FOLDER_FLAG_NO_MESSAGES = 8, /* The folder can't store messages */
	FOLDER_FLAG_NO_SUBFOLDERS = 16 /* Can't have child folders */
};

/* Message types */
enum {
	MESSAGE_TYPE_INVALID = 0,
	MESSAGE_TYPE_SMS,
	MESSAGE_TYPE_MMS,      /* Somewhere in the future */
	MESSAGE_TYPE_EMAIL
};

enum {
	MESSAGE_FLAG_DEFAULT = 0,
	MESSAGE_FLAG_NEW = 1,
	MESSAGE_FLAG_UNREAD = 2,
	MESSAGE_FLAG_UNSENT = 4,
	MESSAGE_FLAG_OUTGOING = 8,
	MESSAGE_FLAG_FORWARDED = 16,
	MESSAGE_FLAG_REPLIED = 32
};

/* Events */
enum {
	EVENT_INCOMING_MESSAGE = 1,
	EVENT_INCOMING_VOICEMAIL,
	EVENT_OUTGOING_SMS_STATUS,
	EVENT_OUTGOING_MAIL_STATUS
	
	
};

/* Structure for giving the folder list to the GUI */
typedef struct {
	gint id;
	gint parent_id;
	const gchar* name;
	gint flags;
} folder_list_info;

/* Structure containing the message for listing, when sent to the GUI */
typedef struct {
	gint type;
	gint id;
	gint ts;
	/* These are UTF-8 encoded strings */
	const gchar* subject;
	const gchar* tofrom;
	gint flags;

} message_list_info;

/* Structure containing the message for displaying */
typedef struct {
	GtkTextBuffer *message;	
	gint ts;	
	gint flags;
	gchararray tofrom;
} sms_display_info;

typedef struct{
	gint id;
	gchararray name;
	/* For sent mail: */
	gchararray path; 
	/* For received mail only: */
	off_t size;	 
	gint folder_id;
	gint message_id;
} mail_attachment;

typedef struct {	
	GtkTextBuffer *message;
	gint ts;
	gint flags;
	gchararray from;
	gchararray to;
	gchararray cc;
	gchararray bcc; /* Bcc is for sent messages :) */
	gchararray subject;
	GList *attachments;
	gint n_attachments;
	off_t attachments_size;
} mail_display_info;


typedef struct {
	gint folder;
	gint type;
	gint id;
	union {
		sms_display_info sms;
		/*mms_display_info mms;*/
		mail_display_info mail;
	} info;
} display_info;

/* Return the data from the events */
typedef struct {
	gint folder;
	message_list_info info;
} event_message_info;

typedef struct {
} event_voicemail_info;

typedef struct {
	gboolean sent;
} event_sms_status;

typedef struct {
	gint type;
	union {
		event_message_info message;
		event_voicemail_info voicemail;
		event_sms_status sms_status;
	} data;
} event_info;

typedef gboolean (*MessageListCallback) (const message_list_info*, gpointer data);
typedef gboolean (*FolderListCallback) (const folder_list_info*, gpointer data);
typedef gboolean (*EventHandler) (const event_info*, gpointer data);

void messages_access_init(EventHandler, gpointer data);

/* The following functions call cb for every message/folder in the list */
gboolean get_folder_list(gint parent, FolderListCallback cb, gboolean recursive, gpointer data);
gboolean get_message_list(gint folder_id, MessageListCallback cb, gpointer data);

/* The following function save the message in the struct pointed to by the last argument   *
 * The result should be cleaned with the corresponding unset function                      */
gboolean message_get_sms(gint folder_id, gint id, sms_display_info *value);
gboolean message_get_mail(gint folder_id, gint id, mail_display_info *value);
gboolean message_delete(gint folder_id, gint type, gint id);
gboolean message_send_sms(const sms_display_info *message, message_list_info *outbox_entry);
gboolean message_send_mail(const mail_display_info *message, message_list_info *outbox_entry);
gboolean message_move(gint src_fid, gint src_id, gint dst_fid, gint dst_id);

/* This one saves the attachment to path, if path is a directory, it uses the attachment name as filename */
gboolean message_mail_save_attachment(gint folder_id, gint mail_id, gint id, const gchar *path);

void display_info_unset(display_info*);
void sms_display_info_unset(sms_display_info*);
void mail_display_info_unset(mail_display_info*);


/* Folder manipulation functions */
/* This is for IMAP folders, returns true when one can perform operations on the folder */
gboolean folder_is_online(gint id);
gboolean folder_can_be_renamed(gint id);

/* The following function makes a copy of the folder name and returns the name of
   the folder, 0 on failure. */
gint folder_create_new(gint parent, const gchar* name);
gboolean folder_rename (gint id, const gchar* new_name);
gboolean folder_delete (gint id);

#endif
