/*
 * Messages for OpenMoko
 * Copyright (C) 2007 Milko Krachounov <exabyte@3mhz.net>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */


#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <time.h>
#include <moko-finger-scroll.h>
#include <moko-stock.h>


#include "messages-defines.h"
#include "messages-common.h"
#include "messages-access.h"
#include "messages-read.h"
#include "messages-write.h"
#include "messages-list.h"


/* Some prototypes */
static inline gboolean has_current_message(AppData *data);
static gboolean get_current_message (AppData *data);
static void messages_fill_message_list(AppData* app_data);




/* Fields in the folder list */
enum {
	FIELD_FOLDER_ID = 0,
	FIELD_FOLDER_ICON,
	FIELD_FOLDER_NAME,
	FIELD_FOLDER_FLAGS,
	N_FIELDS
};

/* Callback on folder selection */

static void
cb_folder_select (GtkComboBox *folder_list, AppData *data)
{	
	static GtkTreeIter iter;

	gtk_combo_box_get_active_iter(folder_list, &iter);
	data->current_folder = &iter;

	if (data->message_list)
		gtk_list_store_clear(GTK_LIST_STORE(data->message_list));

	/* We should free the currently displayed message */
	if (has_current_message(data))
		display_info_unset(&(data->selected));

	data->current_message = NULL;
	messages_fill_message_list(data);

	return;
}

static const gchar *
get_folder_icon (gint id, gint flags) {
	return "gtk-directory";
}

/* A callback for adding folders to the list */ 
static gboolean
cb_add_folder (folder_list_info* folder, AppData* app_data)
{
	gtk_list_store_append (GTK_LIST_STORE(app_data->folder_list), app_data->last_folder);
	gtk_list_store_set (GTK_LIST_STORE(app_data->folder_list), app_data->last_folder,
	                    FIELD_FOLDER_ID, folder->id,
	                    FIELD_FOLDER_ICON, get_folder_icon(folder->id, folder->flags),
	                    FIELD_FOLDER_NAME, folder->name,
	                    FIELD_FOLDER_FLAGS, folder->flags,
	                    -1);
	return TRUE;
}

static void
messages_init_folder_list (AppData *app_data)
{


	GtkListStore *folders;
	static GtkTreeIter iter;

	folders = gtk_list_store_new (N_FIELDS,
	                              G_TYPE_INT,
	                              G_TYPE_STRING,
	                              G_TYPE_STRING,
	                              G_TYPE_INT);

	app_data->last_folder = &iter;

	
	app_data->folder_list = GTK_TREE_MODEL(folders);

	get_folder_list(FOLDER_ROOT, (FolderListCallback) cb_add_folder, TRUE, app_data);
	

}


static GtkWidget *
messages_create_folder_list (AppData* app_data)
{
	GtkCellRenderer     *renderer;
	GtkWidget           *folder_combo;
	static GtkTreeIter  iter;

	folder_combo = gtk_combo_box_new_with_model (app_data->folder_list);

	renderer = gtk_cell_renderer_pixbuf_new();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (folder_combo), renderer, FALSE);
	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (folder_combo),
	                                            renderer,
	                                            "icon-name", FIELD_FOLDER_ICON,
	                                            NULL);
	renderer = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (folder_combo), renderer, TRUE);
	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (folder_combo),
	                                            renderer,
	                                            "text", FIELD_FOLDER_NAME,
	                                            NULL);

	g_object_unref (app_data->folder_list);

	/* Set the current folder and initialize the current folder iter */
	gtk_combo_box_set_active(GTK_COMBO_BOX(folder_combo), 0);
	gtk_combo_box_get_active_iter(GTK_COMBO_BOX(folder_combo), &iter);
	app_data->current_folder = &iter;

	g_signal_connect (G_OBJECT(folder_combo), "changed", G_CALLBACK(cb_folder_select), app_data);
	return folder_combo;
}

/* Message list */
/* Callback called, when a message is selected */
static void
cb_message_select (GtkTreeView *message_list, AppData *data)
{
	static GtkTreeIter current_item;
	/* If there is a selected message, we should free it */
	if (has_current_message(data))
		display_info_unset(&(data->selected));
	data->current_message = NULL;

	GtkTreeSelection *selection = gtk_tree_view_get_selection(message_list);
	if (gtk_tree_selection_get_selected(selection, NULL, &current_item)) {
		data->current_message = &current_item;
	
		get_current_message(data);
	
		container_remove_all_widgets(GTK_CONTAINER(data->display_pane));
	
		if (data->selected.type == MESSAGE_TYPE_SMS) {
			message_sms_display(GTK_BOX(data->display_pane), &(data->selected));
		}
		else if (data->selected.type == MESSAGE_TYPE_EMAIL) {
			message_email_display(GTK_BOX(data->display_pane), &(data->selected));
		}
	}
	else {
		data->current_message = NULL;
	}
}

static void
cb_message_activate (GtkTreeView *tree_view, GtkTreePath *path, GtkTreeViewColumn *column, AppData *data)
{
	/* NOTE: We rely on the fact that when a row is activated, it is already 
	   the currently selected one and that the respective callback is called
           before this one. */
	gint page_num;
	GtkWidget *page;
	if (has_current_message(data)) {
		page = messages_create_read_page(data);
		page_num = notebook_add_page_with_icon (data->notebook, page, "mail-read", 6);
		gtk_widget_show_all (GTK_WIDGET(data->notebook));
		gtk_notebook_set_current_page (data->notebook, page_num);
	}
	return;
}



static const gchar *
get_message_icon_by_flags (gint flags) {
	return "moko-contact-email";
}


/* Callback for adding a message to the message list */
/* Columns in the message list */
enum {
	COL_TYPE = 0,
	COL_ID,
	COL_TOFROM,
	COL_SUBJECT,
	COL_ICON,
	COL_TS,
	COL_UNREAD,
	N_COLUMNS	
};

gboolean
list_add_message (const message_list_info* message, AppData* app_data) {
	gtk_list_store_append (GTK_LIST_STORE(app_data->message_list), app_data->last_message);
	gtk_list_store_set (GTK_LIST_STORE(app_data->message_list), app_data->last_message,
	                    COL_TYPE, message->type,
	                    COL_ID, message->id,
	                    COL_TOFROM, message->tofrom,
	                    COL_SUBJECT, message->subject,
	                    COL_ICON, get_message_icon_by_flags(message->flags),
	                    COL_TS, message->ts,
	                    COL_UNREAD, message->flags & MESSAGE_FLAG_UNREAD,
	                    -1);
	return TRUE;
}

typedef struct {
	gint type;
	gint id;
	gboolean success;
} FindMessageData;

static gboolean
list_delete_current_message(AppData *app_data)
{
	GtkTreeSelection* selection;
	if (!gtk_list_store_remove(GTK_LIST_STORE(app_data->message_list), app_data->current_message))
	{
		return FALSE;
	}

	selection = gtk_tree_view_get_selection(app_data->message_view);
	gtk_tree_selection_select_iter(selection, app_data->current_message);
	return TRUE;
	
}

static gboolean
cb_delete_message (GtkListStore *list, GtkTreePath *p, GtkTreeIter *iter, FindMessageData *data)
{
	if (
		data->type == tree_get_value_int(GTK_TREE_MODEL(list), iter, COL_TYPE) &&
		data->id == tree_get_value_int(GTK_TREE_MODEL(list), iter, COL_ID)
	) {
		data->success = gtk_list_store_remove(GTK_LIST_STORE(list), iter);
		return TRUE;
	}
	return FALSE;
}

gboolean
list_delete_message (gint folder, gint type, gint id, AppData *app_data)
{
	FindMessageData data;
	if (folder != app_data->selected.folder)
		return TRUE;
	if (has_current_message(app_data) && type == app_data->selected.type && id == app_data->selected.id) {
		return list_delete_current_message(app_data);
	}
		
	data.type = type;
	data.id = id;
	data.success = FALSE;
	
	gtk_tree_model_foreach(app_data->message_list, (GtkTreeModelForeachFunc) cb_delete_message, &data);

	return data.success;
	
	
}
/* Fill the message list with the messages from the selected folder */
static void
messages_fill_message_list(AppData* app_data)
{
	gint folder;
	static GtkTreeIter iter;
	app_data->last_message = &iter;

	folder = tree_get_value_int(GTK_TREE_MODEL(app_data->folder_list), app_data->current_folder, FIELD_FOLDER_ID);
	get_message_list(folder, (MessageListCallback) list_add_message, app_data);

	
	app_data->current_message = NULL;

}

static void
messages_init_message_list (AppData* app_data)
{
	
	/* We use app_data->last_message as an iter */

	GtkListStore *messages;

	/* See messages-access.h */
	messages = gtk_list_store_new (N_COLUMNS,
	                               G_TYPE_INT,
	                               G_TYPE_INT,
	                               G_TYPE_STRING,
	                               G_TYPE_STRING,
	                               G_TYPE_STRING,
	                               G_TYPE_INT, /* FIXME: Don't send messages after 2038 */
	                               G_TYPE_INT);


	/* Sort messages by time */
	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (messages), COL_TS, GTK_SORT_DESCENDING);

	app_data->message_list = GTK_TREE_MODEL(messages);

	messages_fill_message_list (app_data);
	
	 
}

static inline gboolean has_current_message(AppData *data)
{
	return data->current_message && data->current_folder;
}

gint
message_list_get_folder (AppData *app_data)
{
	return tree_get_value_int(GTK_TREE_MODEL(app_data->folder_list), app_data->current_folder, FIELD_FOLDER_ID);
}
static gboolean
get_current_message (AppData *data)
{
	if (!has_current_message(data))
		return FALSE;
	data->selected.folder = tree_get_value_int(GTK_TREE_MODEL(data->folder_list), data->current_folder, FIELD_FOLDER_ID);
	data->selected.type = tree_get_value_int(GTK_TREE_MODEL(data->message_list), data->current_message, COL_TYPE);
	data->selected.id = tree_get_value_int(GTK_TREE_MODEL(data->message_list), data->current_message, COL_ID);
	return TRUE;
}

static GtkWidget *
messages_create_message_list (AppData* app_data)
{
	GtkCellRenderer     *renderer;
	GtkWidget           *message_view;
	

	message_view = gtk_tree_view_new ();
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW (message_view), FALSE);
	/* The icon */
	renderer = gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (message_view),
	                                            -1,
	                                            "",
	                                            renderer,
	                                            "icon-name", COL_ICON,
	                                            NULL);
	/* To/From */
	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (message_view),
	                                            -1,
	                                            "Peer", /* TODO */
	                                            renderer,
	                                            "text", COL_TOFROM,
	                                            NULL);
	/* Subject */
	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (message_view),
	                                            -1,
	                                            "Subject",
	                                            renderer,
	                                            "text", COL_SUBJECT,
	                                            NULL);
	gtk_tree_view_set_model (GTK_TREE_VIEW (message_view), app_data->message_list);
	g_object_unref (app_data->message_list);
	g_signal_connect (G_OBJECT(message_view), "cursor-changed", G_CALLBACK(cb_message_select), app_data);
	g_signal_connect (G_OBJECT(message_view), "row-activated", G_CALLBACK(cb_message_activate), app_data);

	app_data->message_view = GTK_TREE_VIEW(message_view);
	return message_view;
}

/* Callback that puts the gutter between the panes in a better position :) */
void cb_paned_size_allocate (GtkWidget *widget, GtkAllocation *allocation, gpointer data)
{
	/* FIXME: This is a crude hack */
	static gint last_height = 0;
	if (allocation->height > 0 && last_height != allocation->height) {
		gtk_paned_set_position(GTK_PANED(widget), allocation->height * 4 / 10);
		last_height = allocation->height;
	}
}

/* Button callbacks */
/* These could be assigned to a GtkToolButton, or a GtkMenuItem */
static void
cb_message_new_sms (GtkWidget *widget, AppData *data)
{
	gint page_num;
	GtkWidget *page;
	page = messages_create_write_sms_page(data);
	page_num = notebook_add_page_with_icon (data->notebook, page, "mail-message-new", 0);
	gtk_widget_show_all (GTK_WIDGET(data->notebook));
	gtk_notebook_set_current_page (data->notebook, page_num);
}

static void
cb_message_new_mail (GtkWidget *widget, AppData *data)
{
	gint page_num;
	GtkWidget *page;
	page = messages_create_write_email_page(data);
	page_num = notebook_add_page_with_icon (data->notebook, page, "mail-message-new", 0);
	gtk_widget_show_all (GTK_WIDGET(data->notebook));
	gtk_notebook_set_current_page (data->notebook, page_num);
}


static void
cb_message_reply_sender (GtkWidget *widget, AppData *data)
{

}

static void
cb_message_reply_all (GtkWidget *widget, AppData *data)
{

}

/* After asking several people, inline forwarding is prefered to attachment,
 * and also, I believe that it is better for a phone */
static void
cb_message_forward_inline (GtkWidget *widget, AppData *data)
{

}

static void
cb_message_forward_attachment (GtkWidget *widget, AppData *data)
{

}

static void
cb_message_delete (GtkToolButton *button, AppData *data)
{
	if (has_current_message(data)) {
		message_delete_dialog (data->selected.folder, data->selected.type, data->selected.id);
		list_delete_current_message(data);
	}
}

static GtkWidget*
get_menu_message_new (AppData* app_data)
{
	GtkWidget *menu, *item;
	menu = gtk_menu_new();

	item = gtk_menu_item_new_with_label (_("New SMS"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_new_sms), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	item = gtk_menu_item_new_with_label (_("New Email"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_new_mail), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_widget_show_all(menu);
	return menu;
}

static GtkWidget*
get_menu_message_reply (AppData* app_data)
{
	GtkWidget *menu, *item;
	menu = gtk_menu_new();

	item = gtk_menu_item_new_with_label (_("Reply to Sender"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_reply_sender), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	item = gtk_menu_item_new_with_label (_("Reply to All"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_reply_all), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_widget_show_all(menu);
	return menu;
}

static GtkWidget*
get_menu_message_forward (AppData* app_data)
{
	GtkWidget *menu, *item;
	menu = gtk_menu_new();

	item = gtk_menu_item_new_with_label (_("Forward as Inline"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_forward_inline), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	item = gtk_menu_item_new_with_label (_("Forward as Attachment"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_forward_attachment), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_widget_show_all(menu);
	return menu;
}

GtkWidget*
messages_create_list_page (AppData* app_data)
{
	GtkWidget *page, *toolbar, *paned, *top_pane,  *folders, *messages, *display_pane, *list_scroll, *preview_scroll, *list_viewport, *preview_viewport;

	page = gtk_vbox_new (FALSE, 0);

	/* We create the toolbar */
	toolbar = gtk_toolbar_new ();

	toolbar_insert_menu_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (MOKO_STOCK_MAIL_NEW, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_message_new_sms), app_data, get_menu_message_new(app_data));
	toolbar_insert_menu_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (MOKO_STOCK_MAIL_REPLY_SENDER, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_message_reply_sender), app_data, get_menu_message_reply(app_data));
	toolbar_insert_menu_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (MOKO_STOCK_MAIL_FORWARD, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_message_forward_inline), app_data, get_menu_message_forward(app_data));
	toolbar_insert_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (MOKO_STOCK_MAIL_DELETE, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_message_delete), app_data);

	gtk_box_pack_start (GTK_BOX (page), toolbar, FALSE, TRUE, 0);


	/* We create the two panes */
	paned = gtk_vpaned_new ();
	/* We fix the gutter position at size allocation */
	g_signal_connect(G_OBJECT(paned), "size-allocate", G_CALLBACK(cb_paned_size_allocate), NULL);

	top_pane = gtk_vbox_new (FALSE, 0);
	display_pane = gtk_vbox_new (FALSE, 0);

	app_data->display_pane = display_pane;
	
	/* We create the folder list */
	messages_init_folder_list(app_data);
	folders = messages_create_folder_list(app_data);
	gtk_box_pack_start (GTK_BOX(top_pane), folders, FALSE, TRUE, 0);

	/* We create the message list */
	messages_init_message_list(app_data);
	messages = messages_create_message_list(app_data);


	list_viewport = gtk_viewport_new(NULL, NULL);
	gtk_viewport_set_shadow_type (GTK_VIEWPORT (list_viewport), GTK_SHADOW_NONE);

	list_scroll = moko_finger_scroll_new();
	gtk_container_add (GTK_CONTAINER (list_viewport), messages);	
	gtk_container_add (GTK_CONTAINER (list_scroll), list_viewport);
	
	gtk_box_pack_end (GTK_BOX (top_pane), list_scroll, TRUE, TRUE, 0);

	/* We add the panes */
	gtk_paned_add1 (GTK_PANED (paned), top_pane);

	/* We create make the display pane scrollable */

	preview_viewport = gtk_viewport_new (NULL, NULL);
	gtk_viewport_set_shadow_type (GTK_VIEWPORT (preview_viewport), GTK_SHADOW_NONE);

	preview_scroll = moko_finger_scroll_new();	
	gtk_container_add (GTK_CONTAINER (preview_viewport), display_pane);
	gtk_container_add (GTK_CONTAINER (preview_scroll), preview_viewport);
	gtk_paned_add2 (GTK_PANED (paned), preview_scroll);

	

	gtk_box_pack_start (GTK_BOX (page), paned, TRUE, TRUE, 0);
	

	return page;

	

}
