/* 
 * Header Cell Renderer for libtinymail and GTK
 * Copyright (C) 2000  Red Hat, Inc.,  Jonathan Blandford <jrb@redhat.com>
 * Copyright (C) 2007 Milko Krachounov <exabyte@3mhz.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "mail-defines.h"
#include "mail-common.h"
#include "headercellrenderer.h"

#include <pango/pango.h>
#include <tny-header.h>
#include <gtk/gtkintl.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gtk/gtkprivate.h>
#include <gtk/gtkiconfactory.h>
#include <gtk/gtkicontheme.h>

G_DEFINE_TYPE (MokoHeaderCellRenderer, moko_header_cell_renderer, GTK_TYPE_CELL_RENDERER)

#define MOKO_HEADER_CELL_RENDERER_GET_PRIVATE(o) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((o), MOKO_TYPE_HEADER_CELL_RENDERER, \
	 MokoHeaderCellRendererPrivate))

typedef struct _MokoHeaderCellRendererPrivate MokoHeaderCellRendererPrivate;
struct _MokoHeaderCellRendererPrivate
{
	TnyHeaderFlags flags;
	gchararray from_text;
	gchararray to_text;
	gchararray subject_text;
	gboolean display_to;
	TnyHeader *header;
	guint xspacing;
	guint yspacing;
	

	PangoAttrList *normal_attr;
	PangoAttrList *unread_attr;
	PangoAttrList *deleted_attr;
	PangoAttrList *peer_attr;

	PangoAttrColor *peer_attr_color;
};

static void moko_header_cell_renderer_get_property  (GObject                    *object,
                                                     guint                       id,
                                                     GValue                     *value,
                                                     GParamSpec                 *pspec);

static void moko_header_cell_renderer_set_property  (GObject                    *object,
                                                     guint                       id,
                                                     const GValue               *value,
                                                     GParamSpec                 *pspec);

static void moko_header_cell_renderer_finalize (GObject *gobject);



static void moko_header_cell_renderer_get_size   (GtkCellRenderer            *cell,
                                                  GtkWidget                  *widget,
                                                  GdkRectangle               *cell_area,
                                                  gint                       *x_offset,
                                                  gint                       *y_offset,
                                                  gint                       *width,
                                                  gint                       *height);

static void moko_header_cell_renderer_render     (GtkCellRenderer            *cell,
                                                  GdkWindow                  *window,
                                                  GtkWidget                  *widget,
                                                  GdkRectangle               *background_area,
                                                  GdkRectangle               *cell_area,
                                                  GdkRectangle               *expose_area,
                                                  guint                       flags);



static GdkPixbuf* get_message_icon_by_flags (TnyHeaderFlags  flags);


static GdkPixbuf* create_colorized_pixbuf (GdkPixbuf *src, 
			                   GdkColor  *new_color);
enum {
	PROP_FROM = 1,
	PROP_TO,
	PROP_SUBJECT,
	PROP_FLAGS,
	PROP_HEADER,
	PROP_DISPLAY_TO,
	PROP_XSPACING,
	PROP_YSPACING
};

static void
moko_header_cell_renderer_finalize (GObject *object)
{
	MokoHeaderCellRendererPrivate *priv = MOKO_HEADER_CELL_RENDERER_GET_PRIVATE(object);
	if (priv->header)
		g_object_unref(G_OBJECT(priv->header));
	if (priv->from_text)
		g_free(priv->from_text);
	if (priv->to_text)
		g_free(priv->to_text);
	if (priv->subject_text)
		g_free(priv->subject_text);

	pango_attr_list_unref(priv->normal_attr);
	pango_attr_list_unref(priv->unread_attr);
	pango_attr_list_unref(priv->deleted_attr);
	pango_attr_list_unref(priv->peer_attr);
	(* G_OBJECT_CLASS (moko_header_cell_renderer_parent_class)->finalize) (object);
}

static void
moko_header_cell_renderer_class_init (MokoHeaderCellRendererClass *class)
{
	GtkCellRendererClass *cell_renderer_class;
	GObjectClass  *object_class;

	g_type_class_add_private (class, sizeof (MokoHeaderCellRendererPrivate));
	
	cell_renderer_class = GTK_CELL_RENDERER_CLASS(class);
	object_class = G_OBJECT_CLASS(class);
	
	cell_renderer_class->get_size = moko_header_cell_renderer_get_size;
	cell_renderer_class->render = moko_header_cell_renderer_render;
	
	object_class->finalize = moko_header_cell_renderer_finalize;
	object_class->get_property = moko_header_cell_renderer_get_property;
	object_class->set_property = moko_header_cell_renderer_set_property;

	g_object_class_install_property (object_class,
					 PROP_FROM,
        				 g_param_spec_string ("from",
							      P_("From"),
							      P_("The sender of the message"),
							      "",
							      GTK_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_TO,
        				 g_param_spec_string ("to",
							      P_("To"),
							      P_("The recipient of the message"),
							      "",
							      GTK_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_SUBJECT,
        				 g_param_spec_string ("subject",
							      P_("Subject"),
							      P_("The subject of the message"),
							      "",
							      GTK_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_FLAGS,
        				 g_param_spec_int ("flags",
							   P_("The TnyHeaderFlags"),
							   P_("The flags of the message"),
							   G_MININT,
							   G_MAXINT,
							   0,
							   GTK_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_HEADER,
        				 g_param_spec_object ("header",
							      P_("The TnyHeader object"),
							      P_("The header object, corresponding to the message"),
							      TNY_TYPE_HEADER,
							      GTK_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_DISPLAY_TO,
        				 g_param_spec_boolean ("display-to",
							       P_("Display the To field"),
							       P_("Display the To field instead of the from field"),
							       FALSE,
							       GTK_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_XSPACING,
        				 g_param_spec_uint ("icon-spacing",
							   P_("Icon spacing"),
							   P_("The spacing between the icon and the text"),
							   0,
							   G_MAXUINT,
							   0,
							   GTK_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_YSPACING,
        				 g_param_spec_uint ("text-spacing",
							   P_("Text spacing"),
							   P_("The spacing between the subject and the sender"),
							   0,
							   G_MAXUINT,
							   0,
							   GTK_PARAM_READWRITE));


}


static void
moko_header_cell_renderer_init (MokoHeaderCellRenderer *header_cell_renderer)
{

	MokoHeaderCellRendererPrivate *priv = MOKO_HEADER_CELL_RENDERER_GET_PRIVATE(header_cell_renderer);
	PangoAttribute *attr;

	
	
	
	priv->flags = 0;

	priv->from_text = NULL;
	priv->to_text = NULL;
	priv->subject_text = NULL;

	priv->header = NULL;

	priv->display_to = FALSE;

	/* Initialize attributes */
	priv->normal_attr = pango_attr_list_new();	
	attr = pango_attr_weight_new(PANGO_WEIGHT_NORMAL);
	attr->start_index = 0;
	attr->end_index = G_MAXUINT;
	pango_attr_list_insert(priv->normal_attr, attr);

	priv->unread_attr = pango_attr_list_new();
	attr = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
	attr->start_index = 0;
	attr->end_index = G_MAXUINT;
	pango_attr_list_insert(priv->unread_attr, attr);

	priv->deleted_attr = pango_attr_list_new();
	attr = pango_attr_strikethrough_new(TRUE);
	attr->start_index = 0;
	attr->end_index = G_MAXUINT;
	pango_attr_list_insert(priv->deleted_attr, attr);
	attr = pango_attr_weight_new(PANGO_WEIGHT_NORMAL);
	attr->start_index = 0;
	attr->end_index = G_MAXUINT;
	pango_attr_list_insert(priv->deleted_attr, attr);
	


	priv->peer_attr = pango_attr_list_new();
	attr = pango_attr_style_new(PANGO_STYLE_ITALIC);
	attr->start_index = 0;
	attr->end_index = G_MAXUINT;
	pango_attr_list_insert(priv->peer_attr, attr);
	attr = pango_attr_weight_new(PANGO_WEIGHT_ULTRALIGHT);
	attr->start_index = 0;
	attr->end_index = G_MAXUINT;
	pango_attr_list_insert(priv->peer_attr, attr);

	/* The colour here is arbitrary. Should be set before drawing! */
	attr = pango_attr_foreground_new (0,0,0);
	attr->start_index = 0;
	attr->end_index = G_MAXUINT;
	priv->peer_attr_color = (PangoAttrColor*) attr;
	pango_attr_list_insert(priv->peer_attr, attr);


}

static void
moko_header_cell_renderer_set_property (GObject      *object,
                                        guint         id,
                                        const GValue *value,
                                        GParamSpec   *pspec)
{
	MokoHeaderCellRendererPrivate *priv = MOKO_HEADER_CELL_RENDERER_GET_PRIVATE(object);
	switch (id) {
		case PROP_FROM:
			if (priv->from_text)
				g_free(priv->from_text);
			priv->from_text = g_value_dup_string(value);
			break;
		case PROP_TO:
			if (priv->to_text)
				g_free(priv->to_text);
			priv->to_text = g_value_dup_string(value);
			break;
		case PROP_SUBJECT:
			if (priv->subject_text)
				g_free(priv->subject_text);
			priv->subject_text = g_value_dup_string(value);
			break;
		case PROP_FLAGS:
			priv->flags = g_value_get_int(value);
			break;
		case PROP_HEADER:
			/* FIXME: I don't know if this is g_object_ref'd, but who cares, when somebody forgot
			 *        to mention it in the documentation, assume that it will work the way you 
		         *        like it
                         */
			priv->header = g_value_get_object(value);
			break;
		case PROP_DISPLAY_TO:
			priv->display_to = g_value_get_boolean(value);
			break;
		case PROP_XSPACING:
			priv->xspacing = g_value_get_uint(value);
			break;
		case PROP_YSPACING:
			priv->yspacing = g_value_get_uint(value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, id, pspec);
	}
}


static void
moko_header_cell_renderer_get_property (GObject    *object,
                                        guint       id,
                                        GValue     *value,
                                        GParamSpec *psec)
{
	MokoHeaderCellRendererPrivate *priv = MOKO_HEADER_CELL_RENDERER_GET_PRIVATE(object);
	switch (id) {
		case PROP_FROM:
			g_value_set_string(value, priv->from_text);
			break;
		case PROP_TO:
			g_value_set_string(value, priv->to_text);
			break;
		case PROP_SUBJECT:
			g_value_set_string(value, priv->subject_text);
			break;
		case PROP_FLAGS:
			g_value_set_int(value, priv->flags);
			break;
		case PROP_HEADER:
			g_value_set_object(value, priv->header);
			break;
		case PROP_DISPLAY_TO:
			g_value_set_boolean(value, priv->display_to);
			break;
		case PROP_XSPACING:
			g_value_set_uint(value, priv->xspacing);
			break;
		case PROP_YSPACING:
			g_value_set_uint(value, priv->yspacing);
			break;
	}
}

GtkCellRenderer *
moko_header_cell_renderer_new (void)
{
	return g_object_new(MOKO_TYPE_HEADER_CELL_RENDERER, NULL);
}


static inline PangoLayout*
get_layout(GtkWidget *widget, const gchar *text)
{
	PangoLayout* layout;
	layout = gtk_widget_create_pango_layout (widget, text);
	pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
	return layout;
}

/*
  I'm trying to do this:

  [ICON] Hello, my dear friend
  [ICON] John Doe
*/
static void
moko_header_cell_renderer_get_size (GtkCellRenderer *cell,
                                    GtkWidget       *widget,
                                    GdkRectangle    *cell_area,
                                    gint            *x_offset,
                                    gint            *y_offset,
                                    gint            *width,
                                    gint            *height)
{
	
	MokoHeaderCellRendererPrivate *priv = MOKO_HEADER_CELL_RENDERER_GET_PRIVATE(cell);

	PangoContext *context;
	PangoFontMetrics *metrics;
	GdkPixbuf *pixbuf = NULL;

	gint text_height, pixbuf_height, calc_height;
	gint pixbuf_width, char_width, calc_width;

	pixbuf = get_message_icon_by_flags(priv->flags);

	context = gtk_widget_get_pango_context (widget);

	
	metrics = pango_context_get_metrics (context, widget->style->font_desc, 
					     pango_context_get_language (context));
	
	char_width = PANGO_PIXELS(pango_font_metrics_get_approximate_char_width (metrics));
	text_height = PANGO_PIXELS(pango_font_metrics_get_ascent (metrics) +
		                   pango_font_metrics_get_descent (metrics));

	
	/* We save the height of both layouts, plus a padding */
	calc_height = text_height * 2 + priv->yspacing;
	
	if (pixbuf) {
		pixbuf_width  = gdk_pixbuf_get_width (pixbuf);
		pixbuf_height = gdk_pixbuf_get_height (pixbuf);
	}
	else {
		pixbuf_width = 0;
		pixbuf_height = 0;
	}
	
	/* We want the pixbuf to be centered in a square area */
#if 0
	if (calc_height > pixbuf_height)
		pixbuf_width += calc_height - pixbuf_height;
#endif
	
	pixbuf_width = MIN(pixbuf_width, 0);
	pixbuf_height = MIN(pixbuf_height, 0);
	
	
	calc_width = (gint) cell->xpad * 2 + pixbuf_width + priv->xspacing + char_width * 3;
	calc_height = (gint) cell->ypad * 2 + MAX(calc_height, pixbuf_height);
	

	if (width)
		*width = calc_width;

	if (height)
		*height = calc_height;

	

	if (cell_area) {
		if (x_offset) {
			*x_offset = (gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL ?
				      1.0 - cell->xalign : cell->xalign) * (cell_area->width - calc_width);
			*x_offset = MAX (*x_offset, 0);
		}

		if (y_offset) {
			*y_offset = cell->yalign * (cell_area->height - calc_height);
			*y_offset = MAX (*y_offset, 0);
		}
	}
	if (pixbuf)
		g_object_unref(pixbuf);


}

static void
moko_header_cell_renderer_render (GtkCellRenderer *cell,
                                  GdkWindow       *window,
                                  GtkWidget       *widget,
                                  GdkRectangle    *background_area,
                                  GdkRectangle    *cell_area,
                                  GdkRectangle    *expose_area,
                                  guint            flags)
{
	MokoHeaderCellRendererPrivate *priv = MOKO_HEADER_CELL_RENDERER_GET_PRIVATE(cell);

	GdkPixbuf *pixbuf = NULL;
	GdkPixbuf *orig_pixbuf = NULL;
	PangoLayout *peer, *subject;

	GtkIconSource *source;
	GdkRectangle pix_rect, draw_rect;
	GdkRectangle peer_rect, subject_rect;
	gint text_height;
	GtkStateType state;
	
	cairo_t *cr;


	

	if (GTK_WIDGET_STATE (widget) == GTK_STATE_INSENSITIVE || !cell->sensitive) {
		state = GTK_STATE_INSENSITIVE;
	}
	else if (flags & GTK_CELL_RENDERER_SELECTED) {
		if (GTK_WIDGET_HAS_FOCUS (widget))
			state = GTK_STATE_SELECTED;
		else
			state = GTK_STATE_ACTIVE;
	}
	else if (flags &  GTK_CELL_RENDERER_PRELIT) {
		state = GTK_STATE_PRELIGHT;
	}
	else {
		state = GTK_STATE_NORMAL;
	}


	orig_pixbuf = get_message_icon_by_flags(priv->flags);
	
	peer = get_layout(widget, priv->display_to ? priv->to_text : priv->from_text);
	subject = get_layout(widget, priv->subject_text);

	/* We set the attributes of the text */

	/* We want the peer field to be dimmed, so it can be easily distinguished */
	priv->peer_attr_color->color.red = widget->style->text_aa[state].red;
	priv->peer_attr_color->color.green = widget->style->text_aa[state].green;
	priv->peer_attr_color->color.blue = widget->style->text_aa[state].blue;
	pango_layout_set_attributes (peer, priv->peer_attr);
	if (priv->flags & TNY_HEADER_FLAG_DELETED)
		pango_layout_set_attributes (subject, priv->deleted_attr);
	else if (priv->flags & TNY_HEADER_FLAG_SEEN)
		pango_layout_set_attributes (subject, priv->normal_attr);
	else
		pango_layout_set_attributes (subject, priv->unread_attr);


	/* 1. The pixbuf position */
	if (orig_pixbuf) {
		pix_rect.width = gdk_pixbuf_get_width (orig_pixbuf) + 1*cell->xpad + priv->xspacing;
		pix_rect.height = gdk_pixbuf_get_height (orig_pixbuf) + 2*cell->ypad;
	}
	else {
		pix_rect.width = 1*cell->xpad;
		pix_rect.height = 2*cell->ypad;
	}
	pix_rect.x = cell_area->x + cell->xpad;
	pix_rect.y = cell_area->y + 0.5 * (cell_area->height - pix_rect.height);

	/* 2. The subject position */
	subject_rect.x = cell_area->x + pix_rect.width;
	subject_rect.y = cell_area->y;
	subject_rect.width = MAX(cell_area->width - pix_rect.width, 0);
	subject_rect.height = (cell_area->height - priv->yspacing) / 2 + (cell_area->height - priv->yspacing) % 2;
	
	/* 3. Our peer buddy position */
	peer_rect.x = subject_rect.x;
	peer_rect.y = subject_rect.y + subject_rect.height + priv->yspacing;
	peer_rect.width = subject_rect.width;
	peer_rect.height = (cell_area->height - priv->yspacing) / 2;

	/* Now we render everything */
	/* 1. Pixbuf */
	if (orig_pixbuf &&
	    gdk_rectangle_intersect (cell_area, &pix_rect, &draw_rect) &&
	    gdk_rectangle_intersect (expose_area, &draw_rect, &draw_rect))
	{
		switch (state) {
			case GTK_STATE_INSENSITIVE:
				source = gtk_icon_source_new ();
				gtk_icon_source_set_pixbuf (source, orig_pixbuf);

				gtk_icon_source_set_size (source, GTK_ICON_SIZE_SMALL_TOOLBAR);
				gtk_icon_source_set_size_wildcarded (source, FALSE);

				pixbuf =  gtk_style_render_icon (widget->style,
								 source,
								 gtk_widget_get_direction (widget),
								 GTK_STATE_INSENSITIVE,
								 (GtkIconSize) -1,
								 widget,
								 "mokoheadercellrenderer");
				gtk_icon_source_free (source);
				break;
			case GTK_STATE_SELECTED:
			case GTK_STATE_ACTIVE:
				pixbuf = create_colorized_pixbuf (orig_pixbuf, &widget->style->base[state]);
				break;
			case GTK_STATE_NORMAL:
			/* NOTE: We don't want to paint a prelighted pixbuf, you might
				 change the behavior, if you want */
			case GTK_STATE_PRELIGHT:
			default:
				pixbuf = orig_pixbuf;
		}
		cr = gdk_cairo_create (window);
		gdk_cairo_set_source_pixbuf (cr, pixbuf, pix_rect.x + cell->xpad, pix_rect.y + cell->ypad);
		gdk_cairo_rectangle (cr, &draw_rect);
		cairo_fill (cr);
		cairo_destroy (cr);
	}

	/* 2. The subject */
	if (gdk_rectangle_intersect (cell_area, &subject_rect, &subject_rect)) {
		pango_layout_set_width(subject, (subject_rect.width - 2*cell->xpad) * PANGO_SCALE);
		pango_layout_get_pixel_size(subject, NULL, &text_height);
		gtk_paint_layout (widget->style,
				    window,
				    state,
				    TRUE,
				    expose_area,
				    widget,
				    "mokoheadercellrenderer",
				    subject_rect.x + cell->xpad,
				    subject_rect.y + (subject_rect.height - text_height) - cell->ypad,
				    subject);
	}

	/* 2. The peer */
	if (gdk_rectangle_intersect (cell_area, &peer_rect, &peer_rect)) {
		pango_layout_set_width(peer, (peer_rect.width - 2*cell->xpad) * PANGO_SCALE);
		pango_layout_get_pixel_size(peer, NULL, &text_height);
		gtk_paint_layout (widget->style,
				    window,
				    state,
				    TRUE,
				    expose_area,
				    widget,
				    "mokoheadercellrenderer",
				    peer_rect.x + cell->xpad,
				    peer_rect.y + (peer_rect.height - text_height) - cell->ypad,
				    peer);
	}

	if (pixbuf && pixbuf != orig_pixbuf)
		g_object_unref(pixbuf);
	if (orig_pixbuf)
		g_object_unref(orig_pixbuf);

	g_object_unref(peer);
	g_object_unref(subject);
}

static void
pixbuf_add_decoration (GdkPixbuf *pixbuf, const gchar *dec_name)
{
	GdkPixbuf *decoration;
	GError *err = NULL;

	decoration = pixbuf_load_from_datadir (dec_name, &err);	
		
	if (!decoration)
		g_warning ("Could not load image: %s\n", err->message);
	else {
		gdk_pixbuf_composite (decoration,
				      pixbuf,
				      0,
				      0,
				      gdk_pixbuf_get_width(decoration),
				      gdk_pixbuf_get_height(decoration),
				      0.0, 0.0,
				      1.0, 1.0,
				      GDK_INTERP_NEAREST,
				      255);
		g_object_unref(decoration);
	}

	if (err)
		g_error_free (err);
}
static GdkPixbuf *
get_message_icon_by_flags (TnyHeaderFlags flags) {
	const gchar *icon_name;
	

#if 0
	gint width, height;
	GdkScreen *screen;
	GtkIconTheme *icon_theme;
	GtkSettings *settings;
#endif
	GError *err = NULL;
	
	GdkPixbuf *ret_val;

	/* FIXME: These should not be hardcoded */
	if (flags & TNY_HEADER_FLAG_SEEN)
		icon_name = "mail-message-read.png";
	else
		icon_name = "mail-message-unread.png";

#if 0
	screen = gtk_widget_get_screen (GTK_WIDGET (widget));
	icon_theme = gtk_icon_theme_get_for_screen (screen);
	settings = gtk_settings_get_for_screen (screen);

	

	if (!gtk_icon_size_lookup_for_settings (settings, GTK_ICON_SIZE_SMALL_TOOLBAR,
						&width, &height))
	{
		width = height = 48;
		
	}

	width = MIN(width, height);
#endif
	ret_val = pixbuf_load_from_datadir (icon_name, &err);

	if (ret_val) {
		if (flags & TNY_HEADER_FLAG_ATTACHMENTS)
			pixbuf_add_decoration(ret_val, "mail-message-attachment.png");
		if (flags & TNY_HEADER_FLAG_ANSWERED)
			pixbuf_add_decoration(ret_val, "mail-message-answered.png");
#if 0
		if (flags & TNY_HEADER_FLAG_ ? )
			pixbuf_add_decoration(ret_val, "mail-message-forwarded.png");
#endif
		if (flags & TNY_HEADER_FLAG_DELETED)
			pixbuf_add_decoration(ret_val, "mail-message-deleted.png");
		
	}


	if (!ret_val)
		g_warning ("Could not load image: %s\n", err->message);
	if (err)
		g_error_free (err);


	return ret_val;
}

/* The code below this line is copied from gtkcellrendererpixbuf.c
   and the copyright is owned by RedHat Inc. */

static GdkPixbuf *
create_colorized_pixbuf (GdkPixbuf *src, 
			 GdkColor  *new_color)
{
  gint i, j;
  gint width, height, has_alpha, src_row_stride, dst_row_stride;
  gint red_value, green_value, blue_value;
  guchar *target_pixels;
  guchar *original_pixels;
  guchar *pixsrc;
  guchar *pixdest;
  GdkPixbuf *dest;
  
  red_value = new_color->red / 255.0;
  green_value = new_color->green / 255.0;
  blue_value = new_color->blue / 255.0;
  
  dest = gdk_pixbuf_new (gdk_pixbuf_get_colorspace (src),
			 gdk_pixbuf_get_has_alpha (src),
			 gdk_pixbuf_get_bits_per_sample (src),
			 gdk_pixbuf_get_width (src),
			 gdk_pixbuf_get_height (src));
  
  has_alpha = gdk_pixbuf_get_has_alpha (src);
  width = gdk_pixbuf_get_width (src);
  height = gdk_pixbuf_get_height (src);
  src_row_stride = gdk_pixbuf_get_rowstride (src);
  dst_row_stride = gdk_pixbuf_get_rowstride (dest);
  target_pixels = gdk_pixbuf_get_pixels (dest);
  original_pixels = gdk_pixbuf_get_pixels (src);
  
  for (i = 0; i < height; i++) {
    pixdest = target_pixels + i*dst_row_stride;
    pixsrc = original_pixels + i*src_row_stride;
    for (j = 0; j < width; j++) {		
      *pixdest++ = (*pixsrc++ * red_value) >> 8;
      *pixdest++ = (*pixsrc++ * green_value) >> 8;
      *pixdest++ = (*pixsrc++ * blue_value) >> 8;
      if (has_alpha) {
	*pixdest++ = *pixsrc++;
      }
    }
  }
  return dest;
}
