/* 
 * OpenMoko functions for libtinymail
 * Copyright (C) 2006-2007 Philip Van Hoof <pvanhoof@gnome.org>
 * Copyright (C) 2007-2008 Milko Krachounov <exabyte@3mhz.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <sys/mman.h>

#include <config.h>
#include <glib/gi18n-lib.h>

#include <string.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>

#include <tny-platform-factory.h>
#include <tny-password-getter.h>
#include <tny-openmoko-platform-factory.h>
#include <tny-account-store.h>
#include <tny-openmoko-account-store.h>
#include <tny-account.h>
#include <tny-store-account.h>
#include <tny-transport-account.h>
#include <tny-device.h>
#include <tny-folder-store.h>
#include <tny-simple-list.h>

#include <tny-camel-account.h>
#include <tny-camel-store-account.h>
#include <tny-camel-imap-store-account.h>
#include <tny-camel-nntp-store-account.h>
#include <tny-camel-pop-store-account.h>
#include <tny-camel-transport-account.h>
#include <tny-openmoko-device.h>
#include <tny-session-camel.h>

#include <tny-gtk-lockable.h>
#include <tny-error.h>


static void tny_account_store_init (gpointer g, gpointer iface_data);

G_DEFINE_TYPE_WITH_CODE(TnyOpenMokoAccountStore, tny_openmoko_account_store, G_TYPE_OBJECT,
			G_IMPLEMENT_INTERFACE(TNY_TYPE_ACCOUNT_STORE, tny_account_store_init))

enum {
	ACCOUNT_ADDED,
	ACCOUNT_REMOVED,
	ACCOUNT_CHANGED,
	IDENTITY_ADDED,
	IDENTITY_REMOVED,
	IDENTITY_CHANGED,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

typedef struct _TnyOpenMokoAccountStorePriv TnyOpenMokoAccountStorePriv;

struct _TnyOpenMokoAccountStorePriv
{
	GConfClient *client;
	gchar *cache_dir;
	TnySessionCamel *session;
	TnyDevice *device;
	guint notify;
	/* Accounts are stored by ids here */
	GHashTable *accounts;
	/* Account data is stored by pointers here */
	GHashTable *account_data;
	/* Cached list of added accounts waiting to be added to the store */
	GSList *new_accounts; 
	GMutex *new_accounts_lock;
	GMutex *gconf_lock;
};


#define TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE(o)	\
	(G_TYPE_INSTANCE_GET_PRIVATE ((o), TNY_TYPE_OPENMOKO_ACCOUNT_STORE, TnyOpenMokoAccountStorePriv))

typedef struct {
	gchar* sent;
	gchar* drafts;
	gchar* trash;
	gchar* identity;
} AccountData;

static void
free_account_data(AccountData *data)
{
	g_free(data->sent);
	g_free(data->drafts);
	g_free(data->trash);
	g_slice_free(AccountData, data);
}
#define new_account_data() g_slice_new0(AccountData)


const gchar *gconf_path = "/apps/tinymail";

static gchar* 
per_account_get_pass_func (TnyAccount *account, const gchar *prompt, gboolean *cancel)
{
	TnyPlatformFactory *platfact = tny_openmoko_platform_factory_get_instance ();
	TnyPasswordGetter *pwdgetter;
	gchar *retval;

	pwdgetter = tny_platform_factory_new_password_getter (platfact);
	retval = (gchar*) tny_password_getter_get_password (pwdgetter, 
		tny_account_get_id (account), prompt, cancel);
	g_object_unref (G_OBJECT (pwdgetter));

	return retval;
}


static void
per_account_forget_pass_func (TnyAccount *account)
{
	TnyPlatformFactory *platfact = tny_openmoko_platform_factory_get_instance ();
	TnyPasswordGetter *pwdgetter;

	pwdgetter = tny_platform_factory_new_password_getter (platfact);
	tny_password_getter_forget_password (pwdgetter, tny_account_get_id (account));
	g_object_unref (G_OBJECT (pwdgetter));
}

static gboolean
tny_openmoko_account_store_alert (TnyAccountStore *self, TnyAccount *account, TnyAlertType type, gboolean question, const GError *error)
{
	GtkMessageType gtktype;
	gboolean retval = FALSE;
	GtkWidget *dialog;

	switch (type)
	{
		case TNY_ALERT_TYPE_INFO:
			gtktype = GTK_MESSAGE_INFO;
			break;
		case TNY_ALERT_TYPE_WARNING:
			gtktype = GTK_MESSAGE_WARNING;
			break;
		case TNY_ALERT_TYPE_ERROR:
		default:
			gtktype = GTK_MESSAGE_ERROR;
			break;
	}

	dialog = gtk_message_dialog_new (NULL, GTK_DIALOG_MODAL,
		gtktype, GTK_BUTTONS_YES_NO, error->message);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_YES)
		retval = TRUE;

	gtk_widget_destroy (dialog);

	return retval;
}

static inline void
kill_stored_accounts (TnyOpenMokoAccountStorePriv *priv)
{
	g_hash_table_remove_all (priv->accounts);
	return;
}


struct _CamelSessionPrivate {
	GMutex *lock;		/* for locking everything basically */
	GMutex *thread_lock;	/* locking threads */

	int thread_id;
	GHashTable *thread_active;
	GThreadPool *thread_pool;

	GHashTable *thread_msg_op;
};

gint account_list_find_by_id (TnyAccount *account, const gchar *id)
{
	return strcmp(tny_account_get_id(account), id);
}


static void
add_account_data (TnyOpenMokoAccountStorePriv *priv, 
                  TnyAccount *account,
                  AccountData *this_account_data)
{
	g_hash_table_insert (priv->account_data, account, this_account_data);
	/* I use g_object_weak_ref for the "opposite" of its intended purpose -- the data in the
	   hash table has to live as long as the account object exists to avoid race conditions,
	   causing segmentation faults. And when the account goes away, the data aboit it is 
	   removed from the hash table. Additionaly, we make each folder to keep a reference
	   to the hash table. */
	g_object_weak_ref(G_OBJECT(account), (GWeakNotify) g_hash_table_remove, priv->account_data);
	g_hash_table_ref(priv->account_data);
	g_object_weak_ref(G_OBJECT(account), (GWeakNotify) g_hash_table_unref, priv->account_data);
}

static gboolean
is_known_proto(const gchar *proto) {
	return  !g_ascii_strncasecmp (proto, "pop", 3) ||
		!g_ascii_strncasecmp (proto, "imap", 4) ||
		!g_ascii_strncasecmp (proto, "smtp", 4);
}

static void
add_account (TnyOpenMokoAccountStore *self, const gchar *id)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	gchar *proto, *type, *key, *name, *mech;
	gint port;
	TnyAccount *account = NULL;
	GSList *options;

	GSList *new_account;

	/* We wait if new accounts are currently updated
	   FIXME: Possibly not needed and not helping.
	*/
	g_mutex_lock (priv->gconf_lock);
	g_mutex_unlock (priv->gconf_lock);

	g_mutex_lock (priv->new_accounts_lock);
	if ((new_account = g_slist_find_custom (priv->new_accounts, id, (GCompareFunc) account_list_find_by_id))) {
		
		priv->new_accounts = g_slist_remove_link(priv->new_accounts, new_account);

		g_hash_table_insert(priv->accounts, g_strdup(id), new_account->data);
		g_slist_free_1(new_account);
		return;
	}
	g_mutex_unlock(priv->new_accounts_lock);

	key = g_strdup_printf("%s/accounts/%s", gconf_path, id);
	if (!gconf_client_dir_exists (priv->client, key, NULL)) {
		g_warning("The account doesn't exist in the configuration: %s", key);
		g_free(key);
		return;
	}
	g_free(key);


	key = g_strdup_printf ("%s/accounts/%s/disabled", gconf_path, id);
	if (gconf_client_get_bool (priv->client, (const gchar*) key, NULL))
	{
		g_free (key);
		return;
	}
	g_free (key);

	key = g_strdup_printf ("%s/accounts/%s/type", gconf_path, id);
	type = gconf_client_get_string (priv->client, (const gchar*) key, NULL);
	g_free (key);
	if (!type) {
		return;
	}

	key = g_strdup_printf ("%s/accounts/%s/proto", gconf_path, id);
	proto = gconf_client_get_string (priv->client, 
		(const gchar*) key, NULL);
	g_free (key);


	key = g_strdup_printf ("%s/accounts/%s/mech", gconf_path, id);
	mech = gconf_client_get_string (priv->client, 
		(const gchar*) key, NULL);
	g_free (key);

	



	if (!g_ascii_strncasecmp (proto, "smtp", 4))
		account = TNY_ACCOUNT (tny_camel_transport_account_new ());
	else if (!g_ascii_strncasecmp (proto, "imap", 4))
		account = TNY_ACCOUNT (tny_camel_imap_store_account_new ());
	else if (!g_ascii_strncasecmp (proto, "nntp", 4))
		account = TNY_ACCOUNT (tny_camel_nntp_store_account_new ());
	else if (!g_ascii_strncasecmp (proto, "pop", 3))
		account = TNY_ACCOUNT (tny_camel_pop_store_account_new ());
	else	/* Unknown, create a generic one? */
		account = TNY_ACCOUNT (tny_camel_store_account_new ());

	if (type)
		g_free (type);

	if (account)
	{
		tny_camel_account_set_session (TNY_CAMEL_ACCOUNT (account), priv->session);
		tny_account_set_proto (TNY_ACCOUNT (account), proto);


		key = g_strdup_printf ("%s/accounts/%s/name", gconf_path, id);
		name = gconf_client_get_string (priv->client, 
			(const gchar*) key, NULL);
		g_free (key);

		if (name)
		{
			tny_account_set_name (TNY_ACCOUNT (account), name);
			g_free (name);
		}

		if (mech)
			tny_account_set_secure_auth_mech (TNY_ACCOUNT (account), mech);

		key = g_strdup_printf ("%s/accounts/%s/options", gconf_path, id);
		options = gconf_client_get_list (priv->client, 
			(const gchar*) key, GCONF_VALUE_STRING, NULL);
		g_free (key);

		if (options)
		{
			GSList *tmp = options;
			while (options)
			{
				tny_camel_account_add_option (TNY_CAMEL_ACCOUNT (account), options->data);
				g_free (options->data);
				options = g_slist_next (options);
			}
			g_slist_free (tmp);
		}

		if (is_known_proto(proto))
		{
			gchar *user, *hostname;

			key = g_strdup_printf ("%s/accounts/%s/user", gconf_path, id);
			user = gconf_client_get_string (priv->client, 
				(const gchar*) key, NULL);

			g_free (key);
			tny_account_set_user (TNY_ACCOUNT (account), user);

			key = g_strdup_printf ("%s/accounts/%s/hostname", gconf_path, id);
			hostname = gconf_client_get_string (priv->client, 
				(const gchar*) key, NULL);
			g_free (key); 
			tny_account_set_hostname (TNY_ACCOUNT (account), 
				hostname);

			key = g_strdup_printf ("%s/accounts/%s/port", gconf_path, id);
			port = gconf_client_get_int (priv->client, 
				(const gchar*) key, NULL);
			g_free (key); 
			if (port != 0) tny_account_set_port (TNY_ACCOUNT (account), 
				port);

			g_free (hostname); g_free (user);
		} else {
			gchar *url_string;

			/* Un officially supported provider */
			/* Assuming there's a url_string in this case */

			key = g_strdup_printf ("%s/accounts/%s/url_string", gconf_path, id);
			url_string = gconf_client_get_string (priv->client, 
				(const gchar*) key, NULL);

			g_free (key);
			tny_account_set_url_string (TNY_ACCOUNT (account), url_string);
			g_free (url_string);
		}

		tny_account_set_id (TNY_ACCOUNT (account), id);

		tny_account_set_forget_pass_func (TNY_ACCOUNT (account),
			per_account_forget_pass_func);


		tny_account_set_pass_func (TNY_ACCOUNT (account),
			per_account_get_pass_func);

		g_hash_table_insert(priv->accounts, g_strdup(id), account);

		/* Add additional data to the hash table */
		AccountData *this_account_data = new_account_data();
		gchar *conf_value;
		
		key = g_strdup_printf("%s/accounts/%s/sent_folder", gconf_path, id);
		conf_value = gconf_client_get_string (priv->client, key, NULL);
		g_free(key);
	
		if (conf_value && *conf_value)
			this_account_data->sent = conf_value;
		
		key = g_strdup_printf("%s/accounts/%s/drafts_folder", gconf_path, id);
		conf_value = gconf_client_get_string (priv->client, key, NULL);
		g_free(key);
	
		if (conf_value && *conf_value)
			this_account_data->drafts = conf_value;
		
		key = g_strdup_printf("%s/accounts/%s/trash_folder", gconf_path, id);
		conf_value = gconf_client_get_string (priv->client, key, NULL);
		g_free(key);
	
		if (conf_value && *conf_value)
			this_account_data->trash = conf_value;
		

		key = g_strdup_printf("%s/accounts/%s/identity", gconf_path, id);
		conf_value = gconf_client_get_string (priv->client, key, NULL);
		g_free(key);
	
		if (conf_value && *conf_value)
			this_account_data->identity = conf_value;
		
		add_account_data(priv, account, this_account_data);

	}

	g_free (mech);
	g_free (proto);
}

static void
remove_account (TnyOpenMokoAccountStore *self, const gchar *id)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	if (!g_hash_table_remove (priv->accounts, id))
		g_warning("Tried to remove an invalid account from account store: %s", id);
}


static void
gconf_listener_account_changed (GConfClient *client, guint cnxn_id,
			GConfEntry *entry, gpointer user_data)
{
	/* TODO: I do not rely on listeners at the moment... I'll think about handling all of these later. */
#if 0
	TnyOpenMokoAccountStore *self = user_data;
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
	gint levels;

	
	gchar category[12] = {0, }; /* accounts or identities */
	gchar id[(sizeof(guint)*3+5)/10] = {0, }; /* In this implementation, account ids are integers.
                                                             10/3 is an approximation of log(10)/log(2) */
	gchar property[16] = {0, };
	gchar *pattern = g_strdup_printf("%s/%%%d[^/]/%%%d[^/]/%%%d[^/]", gconf_path,
	                                                                  sizeof(category)-1,
	                                                                  sizeof(id)-1,
	                                                                  sizeof(property)-1);

	levels = sscanf(entry->key, pattern, category, id, property);

	g_free(pattern);


	if (levels == 3) {
		if (!strcmp(category, "accounts") {
			/* If type was set, a new account was added. If type was removed, the account was removed. */

			if (!strcmp(property, "type")) {
				if (entry->value) {
					add_account(self, id);
					g_signal_emit (self, signals[ACCOUNT_ADDED], 0, 
						       g_hash_table_lookup(priv->accounts, id));
				}
				else {
					g_signal_emit (self, signals[ACCOUNT_REMOVED], 0, 
						       g_hash_table_lookup(priv->accounts, id));
					remove_account(self, account_id);
				}
			}
			else if (!strcmp(property, "disabled")) {
				if (entry->value && gconf_value_get_bool(entry->value)) {
					add_account(self, id);
					g_signal_emit (self, signals[ACCOUNT_ADDED], 0, 
						       g_hash_table_lookup(priv->accounts, id));
				}
				else {
					g_signal_emit (self, signals[ACCOUNT_REMOVED], 0, 
						       g_hash_table_lookup(priv->accounts, id));
					remove_account(self, id);
				}
			}
			else {
				//update_account(self, account_id, property, entry->value);
				//g_signal_emit(self, signals[ACCOUNT_CHANGED, 0,
				//              g_hash_table_lookup(priv->accounts, account_id));
				/* FIXME: With the current design, changing a property might lead to freeing
				   an internal value still processed by the GUI. Loading and unloading
				   the account is needed. */
				TnyAccount *account;
				if ((account = g_hash_table_lookup(priv->accounts, id))) {
					g_signal_emit (self, signals[ACCOUNT_REMOVED], 0, 
						       account);
					remove_account(self, id);
					add_account(self, id);
					g_signal_emit (self, signals[ACCOUNT_ADDED], 0, 
						       g_hash_table_lookup(priv->accounts, id));
				}
			}
		}
		else if (!strcmp(category, "identities") {
			if (!strcmp(property, "address")) {
			}
		}
		
	}
	
	
#endif
}

static void
load_accounts (TnyOpenMokoAccountStore *self)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
	GSList *accounts, *walk;

	gchar *key = g_strdup_printf("%s/accounts", gconf_path);
	accounts = gconf_client_all_dirs (priv->client, key, NULL);
	g_free(key);

	for (walk = accounts; walk; g_free(walk->data), walk = g_slist_next(walk))
	{
		
		gchar *id = strrchr(walk->data, '/')+1;
		if (id == (gpointer) 1) {
			g_warning("The path to the configuration was invalid: %s", (gchar*)walk->data);
			continue;
		}	
		add_account(self, id);
	}
	g_slist_free(accounts);

	tny_session_camel_set_initialized (priv->session);
}


static gboolean
account_in_hash_matches_url_string (gpointer key, TnyAccount* account, const gchar *url_string)
{
	return tny_account_matches_url_string(account, url_string);
}

static TnyAccount* 
tny_openmoko_account_store_find_account (TnyAccountStore *self, const gchar *url_string)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
	TnyAccount *found = g_hash_table_find(priv->accounts, (GHRFunc) account_in_hash_matches_url_string, (gpointer) url_string);
	if (found)
		found = TNY_ACCOUNT(g_object_ref(found));

	return found;
}

/**
 * tny_openmoko_account_store_get_account_by_id:
 * @self: The #TnyOpenMokoAccountStore instance
 * @id: The id of the account
 *
 * Gets the #TnyAccount from the account store with the specified ID.
 *
 * Return value: A #TnyAccount object, or NULL
 **/
TnyAccount* 
tny_openmoko_account_store_get_account_by_id (TnyOpenMokoAccountStore *self, const gchar *id)
{
	g_return_val_if_fail (TNY_IS_OPENMOKO_ACCOUNT_STORE (self), NULL);
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
	TnyAccount *found = g_hash_table_lookup(priv->accounts, id);
	if (found)
		found = TNY_ACCOUNT(g_object_ref(found));

	return found;
}



static const gchar*
tny_openmoko_account_store_get_cache_dir (TnyAccountStore *self)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	if (G_UNLIKELY (!priv->cache_dir))
	{
		/* Note that there's no listener for this key. If it changes,
		   the camelsession should be destroyed and rebuild from scratch.
		   Which basically means reloading the accounts aswell. 
		  
		   So say you're a nut who wants this key to be updatable at 
		   runtime, you'll have to unload all the accounts here, and of
		   course reload them. All the functionality for that is already
		   available. Perhaps I should just do it ... hmm, maybe another
		   day. Soon. Perhaps. I don't know. Probably . . . . bleh. 

		   Oh and, not to forget! You should probably also move the old
		   cache location to the new one. Or cleanup the old one. */

		gchar *cache_dir = gconf_client_get_string (priv->client, 
			"/apps/tinymail/cache_dir", NULL);
		priv->cache_dir = g_build_filename ("/",  cache_dir, NULL);
		g_free (cache_dir);
	}

	return priv->cache_dir;
}

static void
list_prepend_account_in_hash (gpointer key, TnyAccount *account, TnyList *list)
{
	tny_list_prepend (list, (GObject*)account);
}

static void
list_prepend_account_in_hash_if_store (gpointer key, TnyAccount *account, TnyList *list)
{
	if (TNY_IS_STORE_ACCOUNT (account))
		tny_list_prepend (list, (GObject*)account);
}


static void
list_prepend_account_in_hash_if_transport (gpointer key, TnyAccount *account, TnyList *list)
{
	if (TNY_IS_TRANSPORT_ACCOUNT (account))
		tny_list_prepend (list, (GObject*)account);
}


static void
tny_openmoko_account_store_get_accounts (TnyAccountStore *self, TnyList *list, TnyGetAccountsRequestType types)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	g_assert (TNY_IS_LIST (list));


	switch (types) {
		case TNY_ACCOUNT_STORE_STORE_ACCOUNTS:
			g_hash_table_foreach(priv->accounts, (GHFunc) list_prepend_account_in_hash_if_store, list);
			break;
		case TNY_ACCOUNT_STORE_TRANSPORT_ACCOUNTS:
			g_hash_table_foreach(priv->accounts, (GHFunc) list_prepend_account_in_hash_if_transport, list);
			break;
		case TNY_ACCOUNT_STORE_BOTH:
			g_hash_table_foreach(priv->accounts, (GHFunc) list_prepend_account_in_hash, list);
	}

}


static void
tny_openmoko_account_store_notify_add (TnyAccountStore *self)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
	gchar *key = g_strdup_printf("%s/accounts", gconf_path);
	priv->notify = gconf_client_notify_add (priv->client, 
		key, gconf_listener_account_changed,
		self, NULL, NULL);
	g_free(key);
	return;
}

static void
tny_openmoko_account_store_notify_remove (TnyAccountStore *self)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
	gconf_client_notify_remove (priv->client, priv->notify);
	return;
}

/*
	gconftool-2 -s /apps/tinymail/cache_dir -t string /media/card/.tinymail
	##gconftool-2 -s /apps/tinymail/accounts/count -t int COUNT
	gconftool-2 -s /apps/tinymail/accounts/0/proto -t string [smtp|imap|pop]
	gconftool-2 -s /apps/tinymail/accounts/0/type -t string [transport|store]
	gconftool-2 -s /apps/tinymail/accounts/0/user -t string username
	gconftool-2 -s /apps/tinymail/accounts/0/hostname -t string mailserver
or
	gconftool-2 -s /apps/tinymail/accounts/0/url_string -t string url_string


	gconftool-2 -s /apps/tinymail/accounts/0/sent_folder -t string 0/Inbox/Sent

*/


/* The return value should be freed */
static gchar *
get_path_to_folder(TnyFolder *folder) 
{
	GObject *object = G_OBJECT(folder);
	GSList *path_list = NULL; /* The path to the folder in TnyFolder
	                              and TnyAccount objects */ 
	const gchar **path_array;
	gchar *path = NULL;
	gint count = 0;

	g_object_ref(object);
	while (object != NULL) {
		path_list = g_slist_prepend(path_list, object);
		count++;

		if (TNY_IS_ACCOUNT(object))
			object = NULL;
		else if (G_LIKELY(TNY_IS_FOLDER(object)))
			object = G_OBJECT(tny_folder_get_folder_store(TNY_FOLDER(object)));
		else {
			g_warning("get_path_to_folder stumbled on an unknown "
			          "object in the path");
			object = NULL;
		}
	}
	if (G_LIKELY(TNY_IS_ACCOUNT(path_list->data) && count > 0)) {
		gint i;
		GSList *walk;

		path_array = g_malloc(sizeof(gpointer) * (count + 1));

		path_array[0] = tny_account_get_id(TNY_ACCOUNT(path_list->data));
		
		for (i = 1, walk = g_slist_next(path_list);
		     i < count && walk; /* Any of them should be sufficient, I know */
		     i++,  walk = g_slist_next(walk))
			path_array[i] = tny_folder_get_name(TNY_FOLDER(walk->data));
		path_array[i] = NULL;

		path = g_strjoinv ("/", (gchar **) path_array);
		g_free(path_array);

	}
	g_slist_foreach(path_list, (GFunc) g_object_unref, NULL);
	g_slist_free(path_list);
	return path;
}

gboolean
tny_openmoko_account_store_add_account_custom (TnyOpenMokoAccountStore *self, TnyOpenMokoAccountAttribute attr_type, ...)
{

	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	va_list ap;
	va_start(ap, attr_type);
	
	TnyAccount *account = NULL;
	const gchar *type = NULL, *proto = NULL, *mech = NULL, *name = NULL;
	const gchar *user = NULL, *hostname = NULL;
	const gchar *url_string = NULL, *identity = NULL;
	TnyFolder *sent_folder = NULL, *drafts_folder = NULL, *trash_folder = NULL;
	gint port = 0;
	GSList *options = NULL;
	gboolean disabled = FALSE;
	
	/* Some unused values */
	gpointer pointer_value;
	gint int_value;

	while (attr_type != -1) {
		switch (attr_type) {
			case TNY_OPENMOKO_ACCOUNT_OBJECT:
				account = va_arg(ap, TnyAccount*);
				break;
			case TNY_OPENMOKO_ACCOUNT_DISABLED:
				disabled = va_arg(ap, gboolean);
				break;
			case TNY_OPENMOKO_ACCOUNT_TYPE:
				type = va_arg(ap, const gchar *);
				break;
			case TNY_OPENMOKO_ACCOUNT_PROTO:
				proto = va_arg(ap, const gchar *);
				break;
			case TNY_OPENMOKO_ACCOUNT_MECH:
				mech = va_arg(ap, const gchar *);
				break;
			case TNY_OPENMOKO_ACCOUNT_NAME:
				name = va_arg(ap, const gchar *);
				break;
			case TNY_OPENMOKO_ACCOUNT_OPTIONS:
				options = va_arg(ap, GSList *);
				break;
			case TNY_OPENMOKO_ACCOUNT_USER:
				user = va_arg(ap, const gchar *);
				break;
			case TNY_OPENMOKO_ACCOUNT_HOSTNAME:
				hostname = va_arg(ap, const gchar *);
				break;
			case TNY_OPENMOKO_ACCOUNT_PORT:
				port = va_arg(ap, gint);
				break;
			case TNY_OPENMOKO_ACCOUNT_URL_STRING:
				url_string = va_arg(ap, const gchar *);
				break;
			case TNY_OPENMOKO_ACCOUNT_SENT_FOLDER:
				sent_folder = va_arg(ap, TnyFolder *);
				break;
			case TNY_OPENMOKO_ACCOUNT_DRAFTS_FOLDER:
				drafts_folder = va_arg(ap, TnyFolder *);
				break;
			case TNY_OPENMOKO_ACCOUNT_TRASH_FOLDER:
				trash_folder = va_arg(ap, TnyFolder *);
				break;
			case TNY_OPENMOKO_ACCOUNT_IDENTITY:
				identity = va_arg(ap, const gchar *);
				break;
			case TNY_OPNEMOKO_ACCOUNT_SKIP_POINTER:
				pointer_value = va_arg(ap, gpointer);
				break;
			case TNY_OPENMOKO_ACCOUNT_SKIP_INT:
				int_value = va_arg(ap, gint);
				break;
			default:
				g_warning("Unknown attribute type %d passed to tny_openmoko_account_store_add_account_custom", attr_type);
		}
		attr_type = va_arg(ap, TnyOpenMokoAccountAttribute);
	}
	
	va_end(ap);

	if (account) {
		
		if (!TNY_IS_ACCOUNT(account)) {
			g_warning("The passed object is not a valid account object");
			account = NULL;
		}
		else {
		
			if (G_UNLIKELY(type))
				g_warning("Both account and type passed to tny_openmoko_account_store_add_account_custom");

			if (TNY_IS_STORE_ACCOUNT(account))
				type = "store";
			else if (TNY_IS_TRANSPORT_ACCOUNT(account))
				type = "transport";
			else
				type = NULL;

			if (G_UNLIKELY(proto))
				g_warning("Both account and proto passed to tny_openmoko_account_store_add_account_custom");
			proto = tny_account_get_proto(account);
			if (type && is_known_proto(proto)) {
				if(G_UNLIKELY(mech || name || user || hostname || port))
					g_warning("Both account and custom attributes passed to tny_openmoko_account_store_add_account_custom");
			/* FIXME: I don't know which of thse are covered by the url string.
			          That won't get used in my client, anyway. 
			 */
				port = tny_account_get_port(account);
				mech = tny_account_get_secure_auth_mech(account);
				name = tny_account_get_name(account);
				user = tny_account_get_user(account);
				hostname = tny_account_get_hostname(account);	
			}
			else {
				url_string = tny_account_get_url_string(account);
				proto = NULL;
			}
		
			if (G_LIKELY(url_string || (type && proto && name && hostname))) {
				g_mutex_lock(priv->new_accounts_lock);
				priv->new_accounts = g_slist_prepend(priv->new_accounts, g_object_ref(account));
				g_mutex_unlock(priv->new_accounts_lock);
			}
		}

	}
	if (G_UNLIKELY(!url_string && (!type || !proto || !name || !hostname))) {
		g_warning("Arguments passed to not enough to create a new account");
		return FALSE;
	}
	
	gint i;
	gchar *key, *value;

	g_mutex_lock(priv->gconf_lock);
	/* First we need to find a free account name */
	for (i = 0; i <= G_MAXINT; i++)
	{
		key = g_strdup_printf("%s/accounts/%d", gconf_path, i);
		if (!gconf_client_dir_exists (priv->client, key, NULL)) {
			g_free(key);
			break;
		}
		g_free(key);
	}
	if (G_UNLIKELY(i < 0)) {
		g_warning("Tried to create more than the maximum possible accounts.");
		g_mutex_unlock(priv->gconf_lock);
		return FALSE;
	}

	if (account) {
		value = g_strdup_printf("%d", i);
		tny_account_set_id (TNY_ACCOUNT (account), value);
		g_free(value);

		tny_account_set_forget_pass_func (TNY_ACCOUNT (account), per_account_forget_pass_func);
		tny_account_set_pass_func (TNY_ACCOUNT (account), per_account_get_pass_func);
	}

	if (disabled) {
		key = g_strdup_printf ("%s/accounts/%d/disabled", gconf_path, i);
		gconf_client_set_bool (priv->client, (const gchar*) key, TRUE, NULL);
		g_free (key); 
	}


	if (url_string) {
		key = g_strdup_printf ("%s/accounts/%d/user", gconf_path, i);
		gconf_client_set_string (priv->client, (const gchar*) key, url_string, NULL);
		g_free (key); 
	}
	else {
		key = g_strdup_printf ("%s/accounts/%d/name", gconf_path, i);
		gconf_client_set_string (priv->client, (const gchar*) key, name, NULL);
		g_free (key); 

		key = g_strdup_printf ("%s/accounts/%d/hostname", gconf_path, i);
		gconf_client_set_string (priv->client, (const gchar*) key, hostname, NULL);
		g_free (key); 
	
		key = g_strdup_printf ("%s/accounts/%d/proto", gconf_path, i);
		gconf_client_set_string (priv->client, (const gchar*) key, proto, NULL);
		g_free (key); 


		if (port) {
			key = g_strdup_printf ("%s/accounts/%d/port", gconf_path, i);
			gconf_client_set_int (priv->client, (const gchar*) key, port, NULL);
			g_free (key); 
		}

	}
	/* FIXME: I don't know which of thse are covered by the url string.
	          That won't get used in my client, anyway. 
	 */
	if (user) {
		key = g_strdup_printf ("%s/accounts/%d/user", gconf_path, i);
		gconf_client_set_string (priv->client, (const gchar*) key,  user, NULL);
		g_free (key); 
	}
	if (mech) {
		key = g_strdup_printf ("%s/accounts/%d/mech", gconf_path, i);
		gconf_client_set_string (priv->client, (const gchar*) key,  mech, NULL);
		g_free (key); 
	}

	if (options) {
		key = g_strdup_printf("%s/accounts/%d/options", gconf_path, i);
		gconf_client_set_list (priv->client, key, GCONF_VALUE_STRING, options, NULL);
		g_free(key);
	}

	if (sent_folder) {
		value = get_path_to_folder(sent_folder);
		if (value) {
			key = g_strdup_printf ("%s/accounts/%d/sent_folder", gconf_path, i);
			gconf_client_set_string (priv->client, (const gchar*) key,  value, NULL);
			g_free (key); 
			g_free (value);
		}
	}

	if (drafts_folder) {
		value = get_path_to_folder(drafts_folder);
		if (value) {
			key = g_strdup_printf ("%s/accounts/%d/drafts_folder", gconf_path, i);
			gconf_client_set_string (priv->client, (const gchar*) key,  value, NULL);
			g_free (key); 
			g_free (value);
		} 
	}

	if (trash_folder) {
		value = get_path_to_folder(trash_folder);
		if (value) {
			key = g_strdup_printf ("%s/accounts/%d/trash_folder", gconf_path, i);
			gconf_client_set_string (priv->client, (const gchar*) key,  value, NULL);
			g_free (key); 
			g_free (value);
		} 
	}

	if (identity) {
		key = g_strdup_printf ("%s/accounts/%d/identity", gconf_path, i);
		gconf_client_set_string (priv->client, (const gchar*) key,  identity, NULL);
		g_free (key); 
	}

	/* We set type last for our listeners, though the locks should be enough. */
	key = g_strdup_printf ("%s/accounts/%d/type", gconf_path, i);
	gconf_client_set_string (priv->client, (const gchar*) key, (type?type:"unknown"), NULL);
	g_free (key); 

	/* TODO: Use the listener instead. */
	add_account(self, id);
	g_signal_emit (self, signals[ACCOUNT_ADDED], 0, g_hash_table_lookup(priv->accounts, id));

	g_mutex_unlock(priv->gconf_lock);

	return TRUE;
}


#if 0
static void
tny_openmoko_account_store_add_account (TnyAccountStore *self, TnyAccount *account, const gchar *type)
{
	tny_openmoko_account_store_add_account_custom (TNY_OPENMOKO_ACCOUNT_STORE(self),
	                                               TNY_OPENMOKO_ACCOUNT_OBJECT, account,
	                                               TNY_OPENMOKO_ACCOUNT_TYPE, type,
	                                               -1)
}
#endif

void
tny_openmoko_account_store_update_account (TnyOpenMokoAccountStore *self, const gchar *id, TnyOpenMokoAccountAttribute attr_type, ...)
{

	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	va_list ap;

	
	const gchar *string_value;
	TnyFolder *folder_value;
	gpointer pointer_value;
	gint int_value;
	GSList *options = NULL;
	gboolean disabled = FALSE;
	gchar *key, *value;

	TnyAccount *account;

	g_mutex_lock(priv->gconf_lock);
	key = g_strdup_printf ("%s/accounts/%s", gconf_path, id);
	if (G_UNLIKELY(!gconf_client_dir_exists (priv->client, key, NULL) ||
	               !(account = g_hash_table_lookup(priv->accounts, id))))
	{
		g_mutex_unlock(priv->gconf_lock);
		g_free(key);
		g_warning("Tried to update nonexistant account %s in account store.", id);
		return;
	}

	va_start(ap, attr_type);
	while (attr_type != -1) {
		switch (attr_type) {
			case TNY_OPENMOKO_ACCOUNT_OBJECT:
				/* TODO: When an account is passed, compare all the
				   properties and see if something has changed.
				   If the objects are different, remove old one
				   and add this one. */
				pointer_value = va_arg(ap, gpointer);
				g_warning("Passing an account here not supported.");
				break;
			case TNY_OPENMOKO_ACCOUNT_DISABLED:
				disabled = va_arg(ap, gboolean);
				key = g_strdup_printf ("%s/accounts/%s/disabled", gconf_path, id);
				gconf_client_set_bool (priv->client, (const gchar*) key, disabled != NULL, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_TYPE:
				string_value = va_arg(ap, const gchar *);
				key = g_strdup_printf ("%s/accounts/%s/type", gconf_path, id);
				gconf_client_set_string (priv->client, (const gchar*) key, string_value, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_PROTO:
				string_value = va_arg(ap, const gchar *);
				key = g_strdup_printf ("%s/accounts/%s/proto", gconf_path, id);
				gconf_client_set_string (priv->client, (const gchar*) key, string_value, NULL);
				g_free (key);
				break;
			case TNY_OPENMOKO_ACCOUNT_MECH:
				string_value = va_arg(ap, const gchar *);
				key = g_strdup_printf ("%s/accounts/%s/mech", gconf_path, id);
				if (string_value)
					gconf_client_set_string (priv->client, key,  string_value, NULL);
				else
					gconf_client_unset (priv->client, key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_NAME:
				string_value = va_arg(ap, const gchar *);
				key = g_strdup_printf ("%s/accounts/%s/name", gconf_path, id);
				gconf_client_set_string (priv->client, (const gchar*) key, string_value, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_OPTIONS:
				options = va_arg(ap, GSList *);
				key = g_strdup_printf("%s/accounts/%s/options", gconf_path, id);
				if (options)
					gconf_client_set_list (priv->client, key, GCONF_VALUE_STRING, options, NULL);
				else
					gconf_client_unset (priv->clinet, key, NULL);
				g_free(key);
				break;
			case TNY_OPENMOKO_ACCOUNT_USER:
				string_value = va_arg(ap, const gchar *);
				key = g_strdup_printf ("%s/accounts/%s/user", gconf_path, id);
				if (string_value)
					gconf_client_set_string (priv->client, key,  string_value, NULL);
				else
					gconf_client_unset (priv->client, key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_HOSTNAME:
				string_value = va_arg(ap, const gchar *);
				key = g_strdup_printf ("%s/accounts/%s/hostname", gconf_path, id);
				gconf_client_set_string (priv->client, (const gchar*) key, string_value, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_PORT:
				int_value = va_arg(ap, gint);
				key = g_strdup_printf ("%s/accounts/%s/port", gconf_path, id);
				if (int_value)
					gconf_client_set_int (priv->client, key, int_value, NULL);
				else
					gconf_client_unset (priv->client, key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_URL_STRING:
				string_value = va_arg(ap, const gchar *);
				key = g_strdup_printf ("%s/accounts/%s/user", gconf_path, id);
				gconf_client_set_string (priv->client, (const gchar*) key, string_value, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_SENT_FOLDER:
				folder_value = va_arg(ap, TnyFolder *);
				if (folder_value)
					value = get_path_to_folder(folder_value);
				else
					value = NULL;
				
				key = g_strdup_printf ("%s/accounts/%s/sent_folder", gconf_path, id);
				if (value && *value)
					gconf_client_set_string (priv->client, key,  value, NULL);
				else
					gconf_client_unset(priv->client, key, NULL);
				g_free (key); 
				g_free (value);
				break;
			case TNY_OPENMOKO_ACCOUNT_DRAFTS_FOLDER:
				folder_value = va_arg(ap, TnyFolder *);
				if (folder_value)
					value = get_path_to_folder(folder_value);
				else
					value = NULL;
				
				key = g_strdup_printf ("%s/accounts/%s/drafts_folder", gconf_path, id);
				if (value && *value)
					gconf_client_set_string (priv->client, key,  value, NULL);
				else
					gconf_client_unset(priv->client, key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_TRASH_FOLDER:
				folder_value = va_arg(ap, TnyFolder *);
				if (folder_value)
					value = get_path_to_folder(folder_value);
				else
					value = NULL;
				
				key = g_strdup_printf ("%s/accounts/%s/trash_folder", gconf_path, id);
				if (value)
					gconf_client_set_string (priv->client, key,  value, NULL);
				else
					gconf_client_unset(priv->client, key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_IDENTITY:
				string_value = va_arg(ap, const gchar *);
				key = g_strdup_printf ("%s/accounts/%s/identity", gconf_path, id);
				if (string_value)
					gconf_client_set_string (priv->client, key,  string_value, NULL);
				else
					gconf_client_unset (priv->client, key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_SKIP_POINTER:
				pointer_value = va_arg(ap, gpointer);
				break;
			case TNY_OPENMOKO_ACCOUNT_SKIP_INT:
				int_value = va_arg(ap, gint);
				break;
			default:
				g_warning("Unknown attribute type %d passed to tny_openmoko_account_store_add_account_custom", attr_type);
		}
		attr_type = va_arg(ap, TnyOpenMokoAccountAttribute);
	}

	/* TODO: Use the listener instead */

	g_signal_emit (self, signals[ACCOUNT_REMOVED], 0, account);
	remove_account(self, id);
	add_account(self, id);
	g_signal_emit (self, signals[ACCOUNT_ADDED], 0, g_hash_table_lookup(priv->accounts, id));

	g_mutex_unlock(priv->gconf_lock);
	
	va_end(ap);
}


void
tny_openmoko_account_store_get_account_info (TnyOpenMokoAccountStore *self, const gchar *id, TnyOpenMokoAccountAttribute attr_type, ...)
{

	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	va_list ap;

	
	gpointer value;
	gchar *key;

	g_mutex_lock(priv->gconf_lock);
	key = g_strdup_printf ("%s/accounts/%s", gconf_path, id);
	if (G_UNLIKELY(!gconf_client_dir_exists (priv->client, key, NULL))) {
		g_mutex_unlock(priv->gconf_lock);
		g_free(key);
		g_warning("Tried to update nonexistant account %s in account store.", id);
		return;
	}
	g_free(key);

	va_start(ap, attr_type);
	while (attr_type != -1) {
		value = va_arg(ap, gpointer);
		switch (attr_type) {			
			case TNY_OPENMOKO_ACCOUNT_OBJECT:
				*((TnyAccount**) value) = g_object_ref(g_hash_table_lookup(priv->accounts, id));
				break;
			case TNY_OPENMOKO_ACCOUNT_DISABLED:
				key = g_strdup_printf ("%s/accounts/%s/disabled", gconf_path, id);
				*((gboolean*) value) = gconf_client_get_bool (priv->client, (const gchar*) key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_TYPE:
				key = g_strdup_printf ("%s/accounts/%s/type", gconf_path, id);
				*((gchar**) value) = gconf_client_get_string (priv->client, (const gchar*) key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_PROTO:
				key = g_strdup_printf ("%s/accounts/%s/proto", gconf_path, id);
				*((gchar**) value) = gconf_client_get_string (priv->client, (const gchar*) key, NULL);
				g_free (key);
				break;
			case TNY_OPENMOKO_ACCOUNT_MECH:
				key = g_strdup_printf ("%s/accounts/%s/mech", gconf_path, id);
				*((gchar**) value) = gconf_client_get_string (priv->client, (const gchar*) key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_NAME:
				key = g_strdup_printf ("%s/accounts/%s/name", gconf_path, id);
				*((gchar**) value) = gconf_client_get_string (priv->client, (const gchar*) key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_OPTIONS:
				key = g_strdup_printf("%s/accounts/%s/options", gconf_path, id);
				*((GSList**) value) = gconf_client_get_list (priv->client, (const gchar*) key, GCONF_VALUE_STRING, NULL);
				g_free(key);
				break;
			case TNY_OPENMOKO_ACCOUNT_USER:
				key = g_strdup_printf ("%s/accounts/%s/user", gconf_path, id);
				*((gchar**) value) = gconf_client_get_string (priv->client, (const gchar*) key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_HOSTNAME:
				key = g_strdup_printf ("%s/accounts/%s/hostname", gconf_path, id);
				*((gchar**) value) = gconf_client_get_string (priv->client, (const gchar*) key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_PORT:
				key = g_strdup_printf ("%s/accounts/%s/port", gconf_path, id);
				*((gint*) value) = gconf_client_get_int (priv->client, (const gchar*) key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_URL_STRING:
				key = g_strdup_printf ("%s/accounts/%s/user", gconf_path, id);
				*((gchar**) value) = gconf_client_get_string (priv->client, (const gchar*) key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_SENT_FOLDER:
			case TNY_OPENMOKO_ACCOUNT_DRAFTS_FOLDER:
			case TNY_OPENMOKO_ACCOUNT_TRASH_FOLDER:
				g_warning("Getting a configured folder with get_account_info not implemented.");
				*((void**)value) = NULL;
				break;
			case TNY_OPENMOKO_ACCOUNT_IDENTITY:
				key = g_strdup_printf ("%s/accounts/%s/identity", gconf_path, id);
				*((gchar**) value) = gconf_client_get_string (priv->client, (const gchar*) key, NULL);
				g_free (key); 
				break;
			case TNY_OPENMOKO_ACCOUNT_SKIP_POINTER:
			case TNY_OPENMOKO_ACCOUNT_SKIP_INT:
				break;
			default:
				g_warning("Unknown attribute type %d passed to tny_openmoko_account_store_get_account_info", attr_type);
		}
		attr_type = va_arg(ap, TnyOpenMokoAccountAttribute);
	}
	g_mutex_unlock(priv->gconf_lock);
	
	va_end(ap);
}

void		tny_openmoko_account_store_get_identities	(TnyOpenMokoAccountStore *self,
								 GetIdentitiesFunc func,
								 gpointer user_data)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
	gchar *key = g_strdup_printf("%s/identities", gconf_path);
	GSList *walk, *identities = gconf_client_all_dirs (client, key, NULL);
	g_free(key);
	
	for (walk = identities; walk; g_free(walk->data), walk = g_slist_next(walk))
	{
		
		gchar *name, *identity, *address, *replyto, *account_id, *signature;

		const gchar *id = strrchr((gchar*)walk->data, '/')+1;

		key = g_strdup_printf ("%s/name", (gchar*)walk->data);
		name = gconf_client_get_string (client, key, NULL);
		g_free(key);

		key = g_strdup_printf ("%s/address", (gchar*)walk->data);
		address = gconf_client_get_string (client, key, NULL);
		g_free(key);

		key = g_strdup_printf ("%s/replyto", (gchar*)walk->data);
		replyto = gconf_client_get_string (client, key, NULL);
		g_free(key);

		key = g_strdup_printf ("%s/account_id", (gchar*)walk->data);
		account_id = gconf_client_get_string (client, key, NULL);
		g_free(key);
		

		key = g_strdup_printf ("%s/signature", (gchar*)walk->data);
		signature = gconf_client_get_string (client, key, NULL);
		g_free(key);
		

		if (replyto && !replyto[0]) {
			g_free(replyto);
			replyto = NULL;
		}

		if (signature && !signature[0]) {
			g_free(signature);
			signature = NULL;
		}

		Identity identity = { id, name, address, replyto, account_id, signature };

		func(self, FALSE, &identity, user_data);

		g_free(name);
		g_free(address);
		g_free(account_id);
		g_free(signature);
		g_free(replyto);
	}
	g_slist_free(identities);
}


void
tny_openmoko_account_store_remove_account (TnyOpenMokoAccountStore *self, const gchar *id)
{

	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	gchar *key;

	g_mutex_lock(priv->gconf_lock);
	key = g_strdup_printf ("%s/accounts/%s", gconf_path, id);
	if (G_UNLIKELY(!gconf_client_dir_exists (priv->client, key, NULL))) {
		g_mutex_unlock(priv->gconf_lock);
		g_free(key);
		g_warning("Tried to remove nonexistant account %s in account store.", id);
		return;
	}

	/* FIXME: I have no fucking idea what the GCONF_UNSET_INCLUDING_SCHEMA_NAMES flag does,
	   but I have a feeling that it does what I want it to do. But I'm too lazy to try.
	*/
	gconf_client_recursive_unset (priv->client, key, 0, NULL);
	g_free(key);

	/* TODO: Use the listener instead **/
	g_signal_emit (self, signals[ACCOUNT_REMOVED], 0, g_hash_table_lookup(priv->accounts, id));
	remove_account(self, account_id);

	g_mutex_unlock(priv->gconf_lock);
	
}

gboolean	tny_openmoko_account_store_add_identity		(TnyOpenMokoAccountStore *self,
								 const gchar *name,
								 const gchar *address,
								 const gchar *replyto,
								 const gchar *account_id,
								 const gchar *signature)
{
	if (G_UNLIKELY(!name || !*name || !address || !*address)) {
		g_warning("Arguments passed not enough to create an identity.");
		return FALSE;
	}
	

	Identity identity = { "", name, address, replyto, account_id, signature };
	gchar *key;
	gint i;
	gchar id[20] = {0, };

	g_mutex_lock(priv->gconf_lock);
	/* First we need to find a free account name */
	for (i = 0; i <= G_MAXINT; i++)
	{
		key = g_strdup_printf("%s/identities/%d", gconf_path, i);
		if (!gconf_client_dir_exists (priv->client, key, NULL)) {
			g_free(key);
			break;
		}
		g_free(key);
	}
	if (G_UNLIKELY(i < 0)) {
		g_warning("Tried to create more than the maximum possible identities.");
		g_mutex_unlock(priv->gconf_lock);
		return FALSE;
	}

	snprintf(id, sizeof(id), "%d", i);
	identity.id = id;

	key = g_strdup_printf ("%s/identities/%d/name", gconf_path, i);
	gconf_client_set_string (priv->client, (const gchar*) key, name, NULL);
	g_free (key); 


	key = g_strdup_printf ("%s/identities/%d/adress", gconf_path, i);
	gconf_client_set_string (priv->client, (const gchar*) key, address, NULL);
	g_free (key); 

	if (replyto && *replyto) {
		key = g_strdup_printf ("%s/identities/%d/replyto", gconf_path, i);
		gconf_client_set_string (priv->client, (const gchar*) key, replyto, NULL);
		g_free (key); 
	}

	if (account_id && *account_id) {
		key = g_strdup_printf ("%s/identities/%d/account_id", gconf_path, i);
		gconf_client_set_string (priv->client, (const gchar*) key, account_id, NULL);
		g_free (key); 
	}

	if (signature && *signature) {
		key = g_strdup_printf ("%s/identities/%d/signature", gconf_path, i);
		gconf_client_set_string (priv->client, (const gchar*) key, signature, NULL);
		g_free (key); 
	}

	/* TODO: Use the listener instead. */
	g_signal_emit (self, signals[IDENTITY_ADDED], 0, identity);

	g_mutex_unlock(priv->gconf_lock);

	return TRUE;
}



static TnyDevice*
tny_openmoko_account_store_get_device (TnyAccountStore *self)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	return g_object_ref (G_OBJECT (priv->device));
}


/**
 * tny_openmoko_account_store_new:
 *
 * Create a #TnyAccountStore for OpenMoko-based devices that uses GConf for storing 
 * the account information and detecting changes to that configuration.
 *
 * Return value: A new #TnyAccountStore instance
 **/
TnyAccountStore*
tny_openmoko_account_store_new (void)
{
	TnyOpenMokoAccountStore *self = g_object_new (TNY_TYPE_OPENMOKO_ACCOUNT_STORE, NULL);
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	priv->session = tny_session_camel_new (TNY_ACCOUNT_STORE (self));
	tny_session_camel_set_ui_locker (priv->session, tny_gtk_lockable_new ());
	
	return TNY_ACCOUNT_STORE (self);
}


static void
tny_openmoko_account_store_init (TnyOpenMokoAccountStore *self)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
	TnyPlatformFactory *platfact;

	priv->accounts = g_hash_table_new_full (g_str_hash, g_str_equal, 
	                                        (GDestroyNotify) g_free, (GDestroyNotify) g_object_unref);
	/* We create a hash table for the private data */
	priv->account_data = g_hash_table_new_full (NULL, NULL, NULL, 
	                                           (GDestroyNotify) free_account_data);

	priv->new_accounts = NULL;

	priv->new_accounts_lock = g_mutex_new();
	priv->gconf_lock = g_mutex_new();

	priv->client = gconf_client_get_default ();

	gconf_client_add_dir (priv->client, "/apps/tinymail", 
		GCONF_CLIENT_PRELOAD_RECURSIVE, NULL);

	load_accounts(self);
	tny_openmoko_account_store_notify_add (TNY_ACCOUNT_STORE (self));


	platfact = TNY_PLATFORM_FACTORY (tny_openmoko_platform_factory_get_instance ());
	priv->device = tny_platform_factory_new_device (platfact);

	return;
}


static void
tny_openmoko_account_store_finalize (GObject *object)
{
	TnyOpenMokoAccountStore *self = (TnyOpenMokoAccountStore *)object;	
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);

	tny_openmoko_account_store_notify_remove (TNY_ACCOUNT_STORE (self));
	g_object_unref (G_OBJECT (priv->client));

	kill_stored_accounts (priv);

	if (G_LIKELY (priv->cache_dir))
		g_free (priv->cache_dir);

	g_hash_table_unref(priv->accounts);

	g_hash_table_unref (priv->account_data);
	
	g_mutex_lock(priv->new_accounts_lock);
	g_slist_foreach(priv->new_accounts, (GFunc) g_object_unref, NULL);
	g_slist_free(priv->new_accounts);

	priv->new_accounts = NULL;

	g_mutex_unlock(priv->new_accounts_lock);

	g_mutex_free(priv->new_accounts_lock);
	g_mutex_free(priv->gconf_lock);
	(G_OBJECT_CLASS(tny_openmoko_account_store_parent_class)->finalize) (object);

	return;
}


/**
 * tny_openmoko_account_store_get_session:
 * @self: The #TnyOpenMokoAccountStore instance
 *
 * Get the #CamelSession being used
 *
 * Return value: A #TnySessionCamel instance
 **/
TnySessionCamel*
tny_openmoko_account_store_get_session (TnyOpenMokoAccountStore *self)
{
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
    
	return priv->session;
}

GConfClient* 
tny_openmoko_account_store_get_gconf_client (TnyOpenMokoAccountStore *self)
{
	g_return_val_if_fail (TNY_IS_OPENMOKO_ACCOUNT_STORE (self), NULL);
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
	return priv->client;
}

static void 
tny_openmoko_account_store_class_init (TnyOpenMokoAccountStoreClass *class)
{
	GObjectClass *object_class;

	signals[ACCOUNT_ADDED] =
		g_signal_new ("account-added",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (TnyOpenMokoAccountStoreClass, account_added),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1, G_TYPE_OBJECT);

	signals[ACCOUNT_REMOVED] =
		g_signal_new ("account-removed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (TnyOpenMokoAccountStoreClass, account_removed),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1, G_TYPE_OBJECT);

	signals[ACCOUNT_CHANGED] =
		g_signal_new ("account-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (TnyOpenMokoAccountStoreClass, account_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1, G_TYPE_OBJECT);

	object_class = (GObjectClass*) class;

	object_class->finalize = tny_openmoko_account_store_finalize;

	g_type_class_add_private (object_class, sizeof (TnyOpenMokoAccountStorePriv));


	return;
}

static void
tny_account_store_init (gpointer g, gpointer iface_data)
{
	TnyAccountStoreIface *klass = (TnyAccountStoreIface *)g;

	klass->get_accounts_func = tny_openmoko_account_store_get_accounts;
	klass->get_cache_dir_func = tny_openmoko_account_store_get_cache_dir;
	klass->get_device_func = tny_openmoko_account_store_get_device;
	klass->alert_func = tny_openmoko_account_store_alert;
	klass->find_account_func = tny_openmoko_account_store_find_account;

	return;
}


/* The following functions are used to search for folder by path in the store */
typedef struct {
	TnyOpenMokoAccountStore *account_store;
	gchar **folder_path;
	gint level;
	gboolean create;
	TnyOpenMokoFindFolderCallback callback;
	TnyStatusCallback status_callback;
	gpointer user_data;
} FindFolderData;

static void
free_find_folder_data(FindFolderData *data)
{
	g_object_unref(data->account_store);
	g_strfreev(data->folder_path);
	g_free(data);
}

static void
find_folder_status (GObject *sender, TnyStatus *status, FindFolderData *data)
{
	data->status_callback(sender, status, data->user_data);
}



static void
find_folder_callback (TnyFolderStore *self, gboolean cancelled,  
                      TnyList *list, GError *err, FindFolderData *data)
{
	/* We move one level down in the path */
	data->level++;

	if (!cancelled && !err && list && data->folder_path && data->folder_path[data->level]) {
		TnyIterator *iter;
		TnyFolder *folder;
		iter = tny_list_create_iterator (list);
		while (!tny_iterator_is_done (iter)) {
			folder = TNY_FOLDER (tny_iterator_get_current (iter));
			if (!strcmp(data->folder_path[data->level], tny_folder_get_name (folder))) {
				if (!data->folder_path[data->level + 1]) {
					data->callback (data->account_store,
					                FALSE,
					                folder,
					                NULL,
					                data->user_data);
					free_find_folder_data(data);
					g_object_unref(G_OBJECT(folder));
					g_object_unref(G_OBJECT(iter));
					return;
				}
				else if (TNY_IS_FOLDER_STORE(folder)) {
					TnyList *folders;

					folders = tny_simple_list_new ();
					tny_folder_store_get_folders_async  (TNY_FOLDER_STORE(folder),
							folders, 
							(TnyGetFoldersCallback) find_folder_callback,
							NULL,
							(TnyStatusCallback) find_folder_status,
							 data);
					g_object_unref(G_OBJECT(folders));
					g_object_unref(G_OBJECT(folder));
					g_object_unref(G_OBJECT(iter));
					return;
				}
			}
			g_object_unref (G_OBJECT (folder));
			tny_iterator_next (iter);
		}
		g_object_unref(iter);

		/* The folder was not found. Try to create it here */
		/* TODO: I don't know if using async here will make a difference, but as it
		         didn't work for me, I place a TODO here as everywhere I used the following
		         function */
		folder = tny_folder_store_create_folder(self, data->folder_path[data->level], &err);
		if (!err && folder) {
			TnyList *folders;
			folders = tny_simple_list_new ();
			tny_folder_store_get_folders_async  (TNY_FOLDER_STORE(folder),
					folders, 
					(TnyGetFoldersCallback) find_folder_callback,
					NULL,
					(TnyStatusCallback) find_folder_status,
					 data);
			g_object_unref(G_OBJECT(folders));
			g_object_unref(G_OBJECT(folder));
			return;	
		}
		if (folder)
			g_object_unref(G_OBJECT(folder));
	}

	
	if (err || cancelled) {
		
		data->callback (data->account_store, cancelled, NULL, err, data->user_data);
		
	}
	else {
		gchar *folder_name = *data->folder_path ? g_strjoinv("/", data->folder_path + 1) : "";
		g_set_error (&err, TNY_ACCOUNT_STORE_ERROR, TNY_ACCOUNT_STORE_ERROR_UNKNOWN_ALERT,
		             _("Folder %s not found"), folder_name);
		data->callback (data->account_store, cancelled, NULL, err, data->user_data);
		g_free(folder_name);
		g_error_free(err);
	}
	
	free_find_folder_data(data);

}


/**
 * tny_openmoko_acount_store_find_folder_async:
 * @self: The #TnyOpenMokoAccountStore instance
 * @account: The #TnyAccount with which the folder is associated
 * @folder_type: The #TnyFolderType which you are looking for
 * @callback: The #TnyOpenMokoFindFolderCallback called when the folder is found
 * @create: When true, the folder will be created if does not exist
 * @status_callback: #TnyStatusCallback
 * @user_data: A pointer to user data
 *
 * Try to find the folder_type #TnyFolder associated with account, and call the callback function.
 * The folder will be NULL, if not available. If there was an error when searching for the folder,
 * err will be set.
 *
 **/
void
tny_openmoko_acount_store_find_folder_async	(TnyOpenMokoAccountStore *self,
						 TnyAccount *account,
						 TnyFolderType folder_type,
						 gboolean create,
						 TnyOpenMokoFindFolderCallback callback,
						 TnyStatusCallback status_callback,
						 gpointer user_data)
{
	g_return_if_fail (TNY_IS_OPENMOKO_ACCOUNT_STORE (self));
	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);


	AccountData *this_account_data = g_hash_table_lookup(priv->account_data, account);

	/* I consider not valid account a bug in the application, if you find a case,
	   when this should happen in correct code, change and/or notify the mantainer :) */
	if (G_UNLIKELY(!this_account_data)) {
		g_warning("The account %lx doesn't exist in the account store.", (long) account);
		callback(self, FALSE, NULL, FALSE, user_data);
		return;
	}

	const gchar* folder_path;
	switch (folder_type) {
		case TNY_FOLDER_TYPE_SENT:
			folder_path = this_account_data->sent;
			break;
		case TNY_FOLDER_TYPE_DRAFTS:
			folder_path = this_account_data->drafts;
			break;
		case TNY_FOLDER_TYPE_TRASH:
			folder_path = this_account_data->trash;
			break;
		default:
			folder_path = NULL;
	}

	if (!folder_path) {
		callback(self, FALSE, NULL, FALSE, user_data);
		return;
	}

	
	FindFolderData *data = g_malloc0(sizeof(FindFolderData));

	data->account_store = g_object_ref(self);
	data->folder_path = g_strsplit_set(folder_path, "/", -1);
	data->level = 0; /* The folders start from 1, but the callback increments this on call */
	data->create = create;
	data->callback = callback;
	data->status_callback = status_callback;
	data->user_data = user_data;
	
	account = NULL;

	if (data->folder_path[0] && data->folder_path[1]) {
		account = tny_openmoko_account_store_get_account_by_id(self, data->folder_path[0]);
		if (account) {
			TnyList *folders = tny_simple_list_new();
			tny_folder_store_get_folders_async  (TNY_FOLDER_STORE(account),
						folders, 
						(TnyGetFoldersCallback) find_folder_callback,
						NULL,
						(TnyStatusCallback) find_folder_status,
						 data);
			g_object_unref(folders);
		}
	}
	 
		
	free_find_folder_data(data);
	callback(self, FALSE, NULL, FALSE, user_data);

	
}



/**
 * tny_openmoko_account_store_get_identity:
 * @self: The #TnyOpenMokoAccountStore instance
 * @account: The #TnyAccount for which to get the default identity
 *
 * Gets the default identity associated with a #TnyAccount, this should be a 
 * #TnyStoreAccount in general.
 *
 * Return value: A constant gchar array with the identity ID. The returned value
 * should not be freed, modified or stored.
 **/
const gchar*
tny_openmoko_account_store_get_identity	(TnyOpenMokoAccountStore *self,
						 TnyAccount *account)
{
	g_return_val_if_fail (TNY_IS_OPENMOKO_ACCOUNT_STORE (self), NULL);

	TnyOpenMokoAccountStorePriv *priv = TNY_OPENMOKO_ACCOUNT_STORE_GET_PRIVATE (self);
	

	AccountData *this_account_data = g_hash_table_lookup(priv->account_data, account);

	if (G_UNLIKELY(!this_account_data)) {

		g_warning("The account %lx doesn't exist in the account store", (long) account);
		return NULL;
	}
	
	return this_account_data->identity;

	
}

