/* 
 * OpenMoko functions for libtinymail
 * Copyright (C) 2006-2007 Philip Van Hoof <pvanhoof@gnome.org>
 * Copyright (C) 2007 Milko Krachounov <exabyte@3mhz.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <glib/gi18n-lib.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>

#include <tny-openmoko-device.h>


static void tny_device_init (gpointer g, gpointer iface_data);

G_DEFINE_TYPE_WITH_CODE(TnyOpenMokoDevice, tny_openmoko_device, G_TYPE_OBJECT,
			G_IMPLEMENT_INTERFACE(TNY_TYPE_DEVICE, tny_device_init))


#include "tny-openmoko-device-priv.h"

static void tny_openmoko_device_on_online (TnyDevice *self);
static void tny_openmoko_device_on_offline (TnyDevice *self);
static gboolean tny_openmoko_device_is_online (TnyDevice *self);

static gboolean
emit_status (TnyDevice *self)
{
	if (tny_openmoko_device_is_online (self))
		tny_openmoko_device_on_online (self);
	else
		tny_openmoko_device_on_offline (self);

	return FALSE;
}



static void 
tny_openmoko_device_reset (TnyDevice *self)
{
	TnyOpenMokoDevicePriv *priv = TNY_OPENMOKO_DEVICE_GET_PRIVATE (self);

	const gboolean status_before = tny_openmoko_device_is_online (self);

	priv->fset = FALSE;
	priv->forced = FALSE;

	/* Signal if it changed: */
	if (status_before != tny_openmoko_device_is_online (self))
		g_idle_add_full (G_PRIORITY_DEFAULT, 
			(GSourceFunc) emit_status, 
			g_object_ref (self), 
			(GDestroyNotify) g_object_unref);

}

static void 
tny_openmoko_device_force_online (TnyDevice *self)
{
	TnyOpenMokoDevicePriv *priv = TNY_OPENMOKO_DEVICE_GET_PRIVATE (self);

	const gboolean already_online = tny_openmoko_device_is_online (self);

	priv->fset = TRUE;
	priv->forced = TRUE;

	/* Signal if it changed: */
	if (!already_online)
		g_idle_add_full (G_PRIORITY_DEFAULT, 
			(GSourceFunc) emit_status, 
			g_object_ref (self), 
			(GDestroyNotify) g_object_unref);

}


static void
tny_openmoko_device_force_offline (TnyDevice *self)
{
	TnyOpenMokoDevicePriv *priv = TNY_OPENMOKO_DEVICE_GET_PRIVATE (self);

	const gboolean already_offline = !tny_openmoko_device_is_online (self);

	priv->fset = TRUE;
	priv->forced = FALSE;

	/* Signal if it changed: */
	if (!already_offline)
		g_idle_add_full (G_PRIORITY_DEFAULT, 
			(GSourceFunc) emit_status, 
			g_object_ref (self), 
			(GDestroyNotify) g_object_unref);

}

static void
tny_openmoko_device_on_online (TnyDevice *self)
{
	gdk_threads_enter ();
	g_signal_emit (self, tny_device_signals [TNY_DEVICE_CONNECTION_CHANGED], 0, TRUE);
	gdk_threads_leave ();

	return;
}

static void
tny_openmoko_device_on_offline (TnyDevice *self)
{
	gdk_threads_enter ();
	g_signal_emit (self, tny_device_signals [TNY_DEVICE_CONNECTION_CHANGED], 0, FALSE);
	gdk_threads_leave ();

	return;
}

static gboolean
tny_openmoko_device_is_online (TnyDevice *self)
{
	TnyOpenMokoDevicePriv *priv = TNY_OPENMOKO_DEVICE_GET_PRIVATE (self);
	gboolean retval = priv->forced;

	if (!priv->fset)
	{
		/* TODO: Implement. There should be a gconf configuration setting, which
		         tells, whether we should check our mail on gprs by default, or not.
                         That is, if GPRS should be considered "online".
			 For example:
		gint state = some_function(...);
		gint gprs_ok = 	count = gconf_client_get_int (priv->client, 
			"/apps/tinymail/check_on_gprs", NULL);
		switch (state)
		{
			case OPENMOKO_ACTIVE_WIFI_CONNECTION:
			case OPENMOKO_ACTIVE_OTHER_CONNECTION:
				retval = TRUE;
				break;
			case OPENMOKO_ACTIVE_GPRS_CONNECTION:
				retval = gprs_ok;
				break;
			default:
				retval = FALSE;
			
		*/

		retval = TRUE;
	}

	return retval;
}

/* #define IMMEDIATE_ONLINE_TEST */

static void
tny_openmoko_device_init (TnyOpenMokoDevice *self)
{
	TnyOpenMokoDevicePriv *priv = TNY_OPENMOKO_DEVICE_GET_PRIVATE (self);

	priv->fset = FALSE;
	priv->forced = FALSE;

#ifdef IMMEDIATE_ONLINE_TEST
	priv->fset = TRUE;
	priv->forced = TRUE;
#endif

/* Take this as a reference:

#ifdef GNOME
	priv->invnm = FALSE;
	priv->nm_ctx = libnm_glib_init ();
#ifndef IMMEDIATE_ONLINE_TEST
	priv->callback_id = libnm_glib_register_callback 
		(priv->nm_ctx, nm_callback, self, NULL);
#endif
#endif
*/
	return;
}



/**
 * tny_openmoko_device_new:
 *
 * Create a #TnyDevice for OpenMoko based phones. It checks if the phone has an active 
 * internet connection. It can ignore the GPRS connection, if configured.
 * 
 * Return value: A new #TnyDevice instance
 **/
TnyDevice*
tny_openmoko_device_new (void)
{
	TnyOpenMokoDevice *self = g_object_new (TNY_TYPE_OPENMOKO_DEVICE, NULL);

	return TNY_DEVICE (self);
}


static void
tny_openmoko_device_finalize (GObject *object)
{
//	TnyOpenMokoDevice *self = (TnyOpenMokoDevice *)object;	
//	TnyOpenMokoDevicePriv *priv = TNY_OPENMOKO_DEVICE_GET_PRIVATE (self);

/*
#ifdef GNOME
	if (!priv->invnm)
	{
		libnm_glib_unregister_callback (priv->nm_ctx, priv->callback_id);
		libnm_glib_shutdown (priv->nm_ctx);
	}
#endif
*/
	(G_OBJECT_CLASS(tny_openmoko_device_parent_class)->finalize) (object);

	return;
}


static void
tny_device_init (gpointer g, gpointer iface_data)
{
	TnyDeviceIface *klass = (TnyDeviceIface *)g;

	klass->is_online_func = tny_openmoko_device_is_online;
	klass->reset_func = tny_openmoko_device_reset;
	klass->force_offline_func = tny_openmoko_device_force_offline;
	klass->force_online_func = tny_openmoko_device_force_online;

	return;
}



static void 
tny_openmoko_device_class_init (TnyOpenMokoDeviceClass *class)
{
	GObjectClass *object_class;

	object_class = (GObjectClass*) class;

	object_class->finalize = tny_openmoko_device_finalize;

	g_type_class_add_private (object_class, sizeof (TnyOpenMokoDevicePriv));

	return;
}


