/* 
 * Attachment Display Widget for OpenMoko and libtinymail
 * Copyright (C) 2007 Milko Krachounov <exabyte@3mhz.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "tny-openmoko-attachment-view.h"
#include <tny-gtk-attach-list-model.h>
#include <tny-stream.h>
#include <glib/gi18n.h>

/* TODO: I might move the functions here, later */
#include "mail-common.h"
#include "mokofieldlabel.h"


typedef struct _TnyOpenMokoAttachmentViewPrivate TnyOpenMokoAttachmentViewPrivate;
struct _TnyOpenMokoAttachmentViewPrivate
{
	TnyGtkAttachListModel *model;
	GtkWidget *vbox;
	GList *attachment_widgets;
	gint n_attachments;

	gssize total_size;

	/* Needed for calculating the attachment size */
	gssize bytes_written;

};

/* NOTE: The TnyStream interface is used as a hack to calculate the size of an attachment.
 *       It is not very nice, but ...
 */
static void tny_stream_init (gpointer g, gpointer iface_data);
G_DEFINE_TYPE_WITH_CODE (TnyOpenMokoAttachmentView, tny_openmoko_attachment_view, GTK_TYPE_EXPANDER,
			 G_IMPLEMENT_INTERFACE(TNY_TYPE_STREAM, tny_stream_init))

#define TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(o) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((o), TNY_TYPE_OPENMOKO_ATTACHMENT_VIEW, \
	 TnyOpenMokoAttachmentViewPrivate))



enum {
	SAVE_REQUEST,
	DELETE_REQUEST,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

static void
tny_openmoko_attachment_view_class_init (TnyOpenMokoAttachmentViewClass *class)
{
	g_type_class_add_private (class, sizeof (TnyOpenMokoAttachmentViewPrivate));


	signals[SAVE_REQUEST] =
		g_signal_new ("save-request",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (TnyOpenMokoAttachmentViewClass, save_request),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1, G_TYPE_OBJECT);

	signals[DELETE_REQUEST] =
		g_signal_new ("delete-request",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (TnyOpenMokoAttachmentViewClass, delete_request),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1, G_TYPE_OBJECT);
}


static void
tny_openmoko_attachment_view_init (TnyOpenMokoAttachmentView *attachment_view)
{
	
	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(attachment_view);
	priv->model = NULL;

	priv->vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(priv->vbox);
	gtk_container_add(GTK_CONTAINER(attachment_view), priv->vbox);

	priv->attachment_widgets = NULL;
	priv->total_size = 0;
	

}
GtkWidget*
tny_openmoko_attachment_view_new()
{
	return g_object_new (TNY_TYPE_OPENMOKO_ATTACHMENT_VIEW, "label", "", NULL);
}

static void
save_button_pressed (GtkButton *button, TnyOpenMokoAttachmentView *attachment_view)
{
	TnyMimePart *part;

	g_object_get (G_OBJECT(button), "user-data", &part, NULL);
	g_signal_emit (attachment_view, signals[SAVE_REQUEST], 0, part);
}

static void
delete_button_pressed (GtkButton *button, TnyOpenMokoAttachmentView *attachment_view)
{
	TnyMimePart *part;

	g_object_get (G_OBJECT(button), "user-data", &part, NULL);
	g_signal_emit (attachment_view, signals[DELETE_REQUEST], 0, part);

}


static void 
button_destroyed (GtkButton *button, TnyOpenMokoAttachmentView *attachment_view)
{
	GObject *user_data;
	g_object_get (G_OBJECT(button), "user-data", &user_data, NULL);
	g_object_unref(user_data);
}


static void
update_expander_label(TnyOpenMokoAttachmentView *attachment_view)
{
	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(attachment_view);
	gchar buf[128];
	const gchar *unit_name;
	snprintf(buf, sizeof(buf), 
		ngettext("%d attachment (%.2f %s)", "%d attachments (%.2f %s)", priv->n_attachments),
		priv->n_attachments,
		priv->total_size / (double)get_size_display_factor(priv->total_size, &unit_name),
		_(unit_name));
	gtk_expander_set_label(GTK_EXPANDER(attachment_view), buf);
}


static void
attachment_set_info (TnyOpenMokoAttachmentView *attachment_view, 
		     GtkWidget *hbox, const gchar *filename, TnyMimePart *part)
{

	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(attachment_view);


	char buf[32];
	const gchar *unit;
	GtkWidget *label;
	GtkWidget *button;	
	PangoAttrList *bold;
	PangoAttribute *attr;

	
	/* Bold attribute */
	bold = pango_attr_list_new();	
	attr = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
	attr->start_index = 0;
	attr->end_index = G_MAXUINT;
	pango_attr_list_insert(bold, attr);
	
	/* We get the filesize here. We use our attachment_view as a void TnyStream,  which counts the bytes written
	   to it. Decoding the whole attachment is a slow way to get the size, but I cannot think of any other way
	   right now (FIXME). We then save the size in the hbox widget, so we have it stored. */
	tny_mime_part_decode_to_stream (part, TNY_STREAM(attachment_view));
	/* We hope that attachments will be <4GB on 32 bit architectures */
	g_object_set(G_OBJECT(hbox), "user-data", GSIZE_TO_POINTER((gsize)priv->bytes_written), NULL);
	priv->total_size += priv->bytes_written;


	
	/* Save button */
	button = gtk_button_new();
	gtk_button_set_image(GTK_BUTTON(button), gtk_image_new_from_icon_name (GTK_STOCK_SAVE, GTK_ICON_SIZE_MENU));
	g_object_set(G_OBJECT(button), "user-data", g_object_ref(part), NULL);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(save_button_pressed), attachment_view);
	g_signal_connect(G_OBJECT(button), "destroy", G_CALLBACK(button_destroyed), attachment_view);
	gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, TRUE, 0);

	/* Delete button */
	button = gtk_button_new();
	gtk_button_set_image(GTK_BUTTON(button), gtk_image_new_from_icon_name (GTK_STOCK_DELETE, GTK_ICON_SIZE_MENU));
	g_object_set(G_OBJECT(button), "user-data", g_object_ref(part), NULL);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(delete_button_pressed), attachment_view);
	g_signal_connect(G_OBJECT(button), "destroy", G_CALLBACK(button_destroyed), attachment_view);
	gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, TRUE, 0);


	label = moko_field_label_new(filename);
	gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);
	gtk_label_set_attributes(GTK_LABEL(label), bold);
	gtk_misc_set_alignment (GTK_MISC (label), 0.0f, 0.5f);
	gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
	
	

	snprintf(buf, sizeof(buf), "%.2f %s", priv->bytes_written / (double)get_size_display_factor(priv->bytes_written, &unit), unit);
	label = gtk_label_new(buf);
	gtk_misc_set_alignment (GTK_MISC (label), 1.0f, 0.5f);
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, TRUE, 0);



	pango_attr_list_unref (bold);
	gtk_widget_show_all(hbox);
	
}
static void
attachment_changed (GtkTreeModel              *model,
		    GtkTreePath               *path,
		    GtkTreeIter               *iter,
		    TnyOpenMokoAttachmentView *attachment_view)
{
	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(attachment_view);
	gpointer size;

	gchararray filename;
	TnyMimePart *part;

	GtkWidget *hbox;
	gint index;

	index = gtk_tree_path_get_indices(path)[0];

	gtk_tree_model_get (model, iter,
			    TNY_GTK_ATTACH_LIST_MODEL_FILENAME_COLUMN, &filename,
			    TNY_GTK_ATTACH_LIST_MODEL_INSTANCE_COLUMN, &part,
			    -1);

	/* NOTE: The user-data attribute of the hbox carries the attachment size. */
	hbox = g_list_nth_data(priv->attachment_widgets, index);

	/* We subtract the old size */
	g_object_get(G_OBJECT(hbox), "user-data", &size, NULL);
	priv->total_size -= GPOINTER_TO_SIZE(size);
	
	/* We remove the old widgets */
	container_remove_all_widgets(GTK_CONTAINER(hbox));

	if (part)
	{
		attachment_set_info (attachment_view, hbox, filename, part);
	}

	if (part)
		g_object_unref(part);
	if (filename)
		g_free(filename);
	update_expander_label(attachment_view);
}

static void
attachment_added (GtkTreeModel              *model,
		  GtkTreePath               *path,
		  GtkTreeIter               *iter,
		  TnyOpenMokoAttachmentView *attachment_view)
{
	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(attachment_view);

	gchararray filename;
	TnyMimePart *part;

	GtkWidget *hbox;
	gint index;

	index = gtk_tree_path_get_indices(path)[0];


	/* We get the info from the model */
	gtk_tree_model_get (model, iter,
			    TNY_GTK_ATTACH_LIST_MODEL_FILENAME_COLUMN, &filename,
			    TNY_GTK_ATTACH_LIST_MODEL_INSTANCE_COLUMN, &part,
			    -1);

	/* NOTE: The user-data attribute of the hbox carries the attachment size. */
	hbox = gtk_hbox_new(FALSE, 3);
	if (part)
	{
		attachment_set_info (attachment_view, hbox, filename, part);
	}
	else {
		g_object_set(G_OBJECT(hbox), "user-data", GSIZE_TO_POINTER(0), NULL);
	}

	

	if (part)
		g_object_unref(part);
	if (filename)
		g_free(filename);
	
	priv->attachment_widgets = g_list_insert(priv->attachment_widgets, hbox, index);
	priv->n_attachments++;


	gtk_box_pack_start (GTK_BOX (priv->vbox), hbox, FALSE, TRUE, 0);
	gtk_box_reorder_child (GTK_BOX(priv->vbox), hbox, index);


	if (priv->model)
		update_expander_label(attachment_view);
	
	
}

static void
attachment_removed (GtkTreeModel              *model,
		    GtkTreePath               *path,
		    TnyOpenMokoAttachmentView *attachment_view)
{
	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(attachment_view);
	gpointer size;
	GList *attachment;
	gint index;

	index = gtk_tree_path_get_indices(path)[0];
	attachment = g_list_nth(priv->attachment_widgets, index);
	

	g_object_get(G_OBJECT(attachment), "user-data", &size, NULL);

	priv->n_attachments--;
	priv->total_size -= GPOINTER_TO_SIZE(size);

	gtk_widget_destroy(GTK_WIDGET(attachment->data));

	priv->attachment_widgets = g_list_remove_link (priv->attachment_widgets, attachment);
	g_list_free_1 (attachment);

	update_expander_label(attachment_view);
	
}

/* I hope this never ever happens to be called */
static void
attachment_reorder (GtkTreeModel              *model,
		    GtkTreePath               *parent,
		    GtkTreeIter               *iter,
		    gint                      *new_order,
		    TnyOpenMokoAttachmentView *attachment_view)
{
	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(attachment_view);
	gint i;
	GList *list;
	gpointer *item_array;

	/* We save the current order in an array */
	item_array = g_new(gpointer, priv->n_attachments);
	for (list = g_list_first(priv->attachment_widgets), i = 0;
	     list && i < priv->n_attachments ; 
	     list = g_list_next(list), i++)
	{
		item_array[i] = list->data;
	}
	
	/* We create a new list with the new order and reorder the widgets*/
	for (list = NULL, i = priv->n_attachments - 1; i >= 0; i--) {
		list = g_list_prepend(list, item_array[new_order[i]]);
		gtk_box_reorder_child (GTK_BOX(priv->vbox), item_array[new_order[i]], i);
	}
		
	g_list_free(priv->attachment_widgets);
	priv->attachment_widgets = list;

	g_free(item_array);
}

GtkTreeModel*
tny_openmoko_attachment_view_get_model (TnyOpenMokoAttachmentView *attachment_view)
{
	g_return_val_if_fail (TNY_IS_OPENMOKO_ATTACHMENT_VIEW (attachment_view), NULL);
	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(attachment_view);
	return GTK_TREE_MODEL(priv->model);
}
void
tny_openmoko_attachment_view_set_model		(TnyOpenMokoAttachmentView *attachment_view,
				 TnyGtkAttachListModel *model)
{
	g_return_if_fail (TNY_IS_OPENMOKO_ATTACHMENT_VIEW (attachment_view));
	if (model)
		g_return_if_fail (TNY_IS_GTK_ATTACH_LIST_MODEL (model));

	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(attachment_view);


	if (priv->model) {	
		g_signal_handlers_disconnect_by_func (priv->model, attachment_added, attachment_view);
		g_signal_handlers_disconnect_by_func (priv->model, attachment_removed, attachment_view);
		g_signal_handlers_disconnect_by_func (priv->model, attachment_changed, attachment_view);
		g_signal_handlers_disconnect_by_func (priv->model, attachment_reorder, attachment_view);

		g_object_unref(priv->model);
		priv->model = NULL;
	}
	if (priv->attachment_widgets) {
		container_remove_all_widgets(GTK_CONTAINER(priv->vbox));
		g_list_free(priv->attachment_widgets);
		priv->attachment_widgets = NULL;
	}
	if (model) {
		g_object_ref(model);
		gtk_tree_model_foreach (GTK_TREE_MODEL(model), (GtkTreeModelForeachFunc) attachment_added, attachment_view);
		priv->model = model;
		update_expander_label(attachment_view);
		g_signal_connect (model, "row-inserted", G_CALLBACK (attachment_added), attachment_view);
		g_signal_connect (model, "row-deleted", G_CALLBACK (attachment_removed), attachment_view);
		g_signal_connect (model, "row-changed", G_CALLBACK (attachment_changed), attachment_view);
		g_signal_connect (model, "rows-reordered", G_CALLBACK (attachment_reorder), attachment_view);
	}
	
}



/* TnyStream interface hack follows ... */
static gssize
tny_void_stream_read(TnyStream *self, char *buffer, gsize n)
{
	return 0;
}

static gssize
tny_void_stream_write(TnyStream *self, const char *buffer, gsize n)
{
	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(self);
	priv->bytes_written += n;
	return n;
}

static gboolean
tny_void_stream_is_eos (TnyStream *self)
{
	return TRUE;
}

static gssize
tny_void_stream_write_to_stream (TnyStream *self, TnyStream *melf)
{
	return 0;
}

static gint
tny_void_stream_do_nothing(TnyStream *self)
{
	return 0;
}

static gint
tny_void_stream_reset(TnyStream *self)
{
	TnyOpenMokoAttachmentViewPrivate *priv = TNY_OPENMOKO_ATTACHMENT_VIEW_GET_PRIVATE(self);
	priv->bytes_written = 0;
	return 0;
}


static void
tny_stream_init (gpointer g, gpointer iface_data)
{
	TnyStreamIface *klass = (TnyStreamIface *)g;

	klass->read_func = tny_void_stream_read;
	klass->write_func = tny_void_stream_write;
	klass->flush_func = tny_void_stream_do_nothing;
	klass->close_func = tny_void_stream_do_nothing;
	klass->reset_func = tny_void_stream_reset;
	klass->is_eos_func = tny_void_stream_is_eos;
	klass->write_to_stream_func = tny_void_stream_write_to_stream;

	return;
}
