/*
 * Mail for OpenMoko
 * Copyright (C) 2007 Milko Krachounov <exabyte@3mhz.net>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */


#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <time.h>
#include <moko-finger-scroll.h>
#include <moko-search-bar.h>
#include <moko-stock.h>
#include <tny-folder.h>
#include <tny-header.h>
#include <tny-msg.h>
#include <tny-simple-list.h>
#include <tny-account-store.h>
#include <tny-store-account.h>
#include <tny-folder-monitor.h>
#include <tny-folder-store.h>
#include <tny-gtk-folder-store-tree-model.h>
#include <tny-gtk-header-list-model.h>


#include "mail-defines.h"
#include "mail-common.h"
#include "mail-access.h"
#include "mail-read.h"
#include "mail-write.h"
#include "mail-list.h"
#include "mail-config.h"

#include "headercellrenderer.h"


#define SEARCH_BAR_TIMEOUT 500

/* Some prototypes */
static inline gboolean has_current_message(AppData *data);
static void messages_load_message_list (AppData* app_data);

static inline gboolean
is_outgoing_folder_type(TnyFolderType type)
{
	return (type == TNY_FOLDER_TYPE_DRAFTS || 
	        type == TNY_FOLDER_TYPE_SENT || 
	        type == TNY_FOLDER_TYPE_OUTBOX);
}


/* Callback on folder selection */

static void
cb_folder_select (MokoSearchBar *bar, GtkComboBox *folder_combo, AppData *data)
{	
	GtkTreeIter iter;
	TnyFolder *new_folder;
	gint element_type;

	if (!gtk_combo_box_get_active_iter(folder_combo, &iter))
		return;

	
	gtk_tree_model_get(data->folder_list,
			   &iter,
			   TNY_GTK_FOLDER_STORE_TREE_MODEL_TYPE_COLUMN,
			   &element_type,
			   TNY_GTK_FOLDER_STORE_TREE_MODEL_INSTANCE_COLUMN,
			   &new_folder,
			   -1);

	

	if (element_type != TNY_FOLDER_TYPE_ROOT && new_folder != data->current_folder) {
		if (data->current_folder)
			g_object_unref(G_OBJECT(data->current_folder));
	
		if (data->current_header) {
			g_object_unref(G_OBJECT(data->current_header));
			data->current_header = NULL;
		}

		data->current_folder = new_folder;
		data->folder_type = element_type;
		g_object_set(data->renderer, 
			     "display-to", is_outgoing_folder_type(element_type) != 0,
			     NULL);
		messages_load_message_list(data);
	}
	else {
		g_object_unref(G_OBJECT(new_folder));
	}
}

/* We will wait SEARCH_BAR_TIMEOUT milliseconds before refiltering the list, 
   so that that entering the search string could be faster
*/
typedef struct {
	gint *source;
	GtkEditable *entry;
	AppData *app_data;
} SearchTimeoutData;

static void
destroy_search_timeout_data (SearchTimeoutData *data)
{
	*(data->source) = 0;
	g_object_unref(data->entry);
	g_free(data);
}

static gboolean
execute_refilter (SearchTimeoutData *data)
{
	const gchar *string;
	
	if (data->app_data->message_filter_text)
		g_free(data->app_data->message_filter_text);

	string = gtk_entry_get_text(GTK_ENTRY(data->entry));

	if (string[0])
		data->app_data->message_filter_text = g_utf8_strup (string, -1);
	else
		data->app_data->message_filter_text = NULL;

	gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(data->app_data->message_filter));
	return FALSE;
}

static void
cb_search_filter (MokoSearchBar *bar, GtkEditable *entry, AppData *app_data)
{
	static gint hold_timeout_source = 0;
	SearchTimeoutData *data;	

	if (hold_timeout_source) {
		g_source_remove (hold_timeout_source);
		hold_timeout_source = 0;
	}
	
	data = g_malloc(sizeof(SearchTimeoutData));
	data->source = &hold_timeout_source;
	data->entry = g_object_ref(entry);
	data->app_data = app_data;

	hold_timeout_source = g_timeout_add_full (G_PRIORITY_DEFAULT, SEARCH_BAR_TIMEOUT, (GSourceFunc) execute_refilter, data, (GDestroyNotify) destroy_search_timeout_data);
}


static void
cb_search_toggled (MokoSearchBar *bar, gboolean visible, AppData *data)
{
	data->message_filter_active = visible;
	if (data->message_filter_text)
		gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(data->message_filter));
}
#if 0
static const gchar *
get_folder_icon (gint id, gint flags) {
	return "gtk-directory";
}
#endif

/* Functions dealing with the account/folder tree model */
static void
add_account_to_model (TnyAccountStore *store, TnyAccount *account, GtkTreeModel *folder_list)
{
	if (TNY_IS_STORE_ACCOUNT(account))
		tny_gtk_folder_store_tree_model_append(
				TNY_GTK_FOLDER_STORE_TREE_MODEL(folder_list),
				TNY_FOLDER_STORE(account),
				tny_account_get_name(account));
}

static void
remove_account_from_model (TnyAccountStore *store, TnyAccount *account, GtkTreeModel *folder_list)
{
	tny_list_remove(TNY_LIST(folder_list), G_OBJECT(account));
}

static void
messages_init_folder_list (AppData *data)
{
	TnyAccount *account;
	TnyIterator *iter;
	TnyList *list;

	data->folder_list = tny_gtk_folder_store_tree_model_new  (NULL);

	/* TODO: New accounts? */

	list = tny_simple_list_new ();
	if (data->account_store)
		tny_account_store_get_accounts (data->account_store, list, TNY_ACCOUNT_STORE_STORE_ACCOUNTS);

	iter = tny_list_create_iterator (list);
	while (!tny_iterator_is_done (iter)) {
		account = TNY_ACCOUNT (tny_iterator_get_current (iter));
		tny_gtk_folder_store_tree_model_append(
					TNY_GTK_FOLDER_STORE_TREE_MODEL(data->folder_list),
					TNY_FOLDER_STORE(account),
					tny_account_get_name(account));
		g_object_unref (G_OBJECT (account));
		tny_iterator_next (iter);
	}
	g_signal_connect(G_OBJECT(data->account_store), "account-added",
	                 G_CALLBACK(add_account_to_model), data->folder_list);
	g_signal_connect(G_OBJECT(data->account_store), "account-removed",
	                 G_CALLBACK(remove_account_from_model), data->folder_list);

	g_object_unref (G_OBJECT (iter));
	g_object_unref (G_OBJECT (list));

	

}


static GtkWidget *
messages_create_folder_list (AppData* app_data)
{
	GtkCellRenderer     *renderer;
	GtkWidget           *search_bar, *folder_combo;

	
	folder_combo = gtk_combo_box_new_with_model (app_data->folder_list);

	/* TODO: Create a new renderer, with an icon and bold on folders with new
	         messages, depending on TNY_GTK_FOLDER_STORE_TREE_MODEL_TYPE_COLUMN
	         and  TNY_GTK_FOLDER_STORE_TREE_MODEL_UNREAD_COLUMN */

	renderer = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (folder_combo), renderer, TRUE);
	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (folder_combo),
	                                            renderer,
	                                            "text", TNY_GTK_FOLDER_STORE_TREE_MODEL_NAME_COLUMN,
	                                            NULL);

	/* Set the current folder and initialize the current folder iter */
	gtk_combo_box_set_active(GTK_COMBO_BOX(folder_combo), 0);


	search_bar = moko_search_bar_new_with_combo(GTK_COMBO_BOX(folder_combo));

	g_signal_connect (G_OBJECT(search_bar), "combo-changed", G_CALLBACK(cb_folder_select), app_data);
	g_signal_connect (G_OBJECT(search_bar), "text-changed", G_CALLBACK(cb_search_filter), app_data);
	g_signal_connect (G_OBJECT(search_bar), "toggled", G_CALLBACK(cb_search_toggled), app_data);

	app_data->folder_combo = GTK_COMBO_BOX(folder_combo);

	return search_bar;
}

/* Message list */
/* Callback called, when a message is selected */
static void
cb_message_select (GtkTreeSelection *selection, AppData *data)
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GList *selected;
	/* If there is a selected message, we should free it */
	if (data->current_header) {
		g_object_unref(G_OBJECT(data->current_header));
		data->current_header = NULL;
	}

	selected = gtk_tree_selection_get_selected_rows (selection, &model);
	if (selected && !g_list_next(selected) && gtk_tree_model_get_iter (model, &iter, selected->data)) {

		gtk_tree_model_get(model,
				  &iter,
				  TNY_GTK_HEADER_LIST_MODEL_INSTANCE_COLUMN,
				  &(data->current_header),
				  -1);

	}
	g_list_foreach (selected, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (selected);
}


static gboolean
message_list_filter (GtkTreeModel *model, GtkTreeIter *iter, AppData *data)
{
	TnyHeaderFlags flags;
	gtk_tree_model_get(model,
			   iter,
			   TNY_GTK_HEADER_LIST_MODEL_FLAGS_COLUMN,
			   &flags,
			   -1);

	/* Ughh??! What the fuck?! There should be no such headers in here? */
	if (flags & TNY_HEADER_FLAG_EXPUNGED)
		return FALSE;

	if (!data->message_filter_active || !data->message_filter_text)
		return TRUE;

	gchararray peer, subject, compare;
	gboolean ret_val = FALSE;

	gtk_tree_model_get(model,
			   iter,
			   is_outgoing_folder_type(data->folder_type) ?
				TNY_GTK_HEADER_LIST_MODEL_TO_COLUMN :
				TNY_GTK_HEADER_LIST_MODEL_FROM_COLUMN,
			   &peer,
			   TNY_GTK_HEADER_LIST_MODEL_SUBJECT_COLUMN,
			   &subject,
			   -1);
	if (peer) {
		compare = g_utf8_strup(peer, -1);
		ret_val |= g_strrstr(compare, data->message_filter_text) != NULL;
		g_free(compare);
		g_free(peer);
	}
		
	if (subject) {
		compare = g_utf8_strup(subject, -1);
		ret_val |= g_strrstr(compare, data->message_filter_text) != NULL;
		g_free(compare);
		g_free(subject);
	}
	return ret_val;
}

static void
headers_loaded (TnyFolder *self, gboolean cancelled, TnyList *headers, GError *err, AppData *app_data)
{
	GtkWidget *dialog;
	gint dialog_result;

	if (err) {
		quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
		                 _("Error loading folder: %s"), err->message);
	}
}



static void
messages_load_message_list (AppData* app_data)
{
	

	GtkTreeModel *messages;
	GtkTreeModel *messages_sortable;
	GtkTreeModel *messages_filter;

	if (!app_data->current_folder) {
		gtk_tree_view_set_model (GTK_TREE_VIEW (app_data->message_view), NULL);
		return;
	}

	messages = tny_gtk_header_list_model_new();

	tny_gtk_header_list_model_set_folder (TNY_GTK_HEADER_LIST_MODEL(messages),
				      app_data->current_folder,
				      FALSE, /* Refresh */
				      (TnyGetHeadersCallback) headers_loaded,
				      (TnyStatusCallback) status_update,
				      app_data);

	if (app_data->folder_monitor) {
		tny_folder_monitor_stop (TNY_FOLDER_MONITOR(app_data->folder_monitor));
		g_object_unref (G_OBJECT (app_data->folder_monitor));
	}

	app_data->folder_monitor = tny_folder_monitor_new (app_data->current_folder);
	tny_folder_monitor_add_list (TNY_FOLDER_MONITOR(app_data->folder_monitor), TNY_LIST (messages));
	tny_folder_monitor_start (TNY_FOLDER_MONITOR(app_data->folder_monitor));


	/* We set up a GtkTreeModelSortable */

	messages_sortable = gtk_tree_model_sort_new_with_model(messages);

	/* Sort messages by time */

	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE(messages_sortable),
					 TNY_GTK_HEADER_LIST_MODEL_DATE_RECEIVED_COLUMN,
					 tny_gtk_header_list_model_received_date_sort_func,
					 NULL,
					 NULL);

	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE(messages_sortable),
					 TNY_GTK_HEADER_LIST_MODEL_DATE_SENT_COLUMN,
					 tny_gtk_header_list_model_sent_date_sort_func,
					 NULL,
					 NULL);

	/* TODO */
	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (messages_sortable),
					      TNY_GTK_HEADER_LIST_MODEL_DATE_SENT_COLUMN,
					      GTK_SORT_DESCENDING);
	


	messages_filter = gtk_tree_model_filter_new(messages_sortable, NULL);
	
	gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER(messages_filter), 
						(GtkTreeModelFilterVisibleFunc) message_list_filter,
						app_data,
						NULL);
	
	gtk_tree_view_set_model (GTK_TREE_VIEW (app_data->message_view), messages_filter);
	
	
	app_data->message_filter = messages_filter;
	
	gtk_widget_show_all(GTK_WIDGET(app_data->message_view));
	g_object_unref(messages);
	g_object_unref(messages_filter);
	g_object_unref(messages_sortable); 

	tny_folder_refresh_async (app_data->current_folder, NULL, 
				  (TnyStatusCallback) status_update, app_data);
}

static inline gboolean has_current_message(AppData *data)
{
	return data->current_header != NULL;
}



/* This callback catches a key press or release and adds modifiers to it */
gboolean
cb_list_click  (GtkWidget *widget, GdkEventButton *event, AppData *data)
{
	event->state |= data->modifiers;
	return FALSE;
}

static GtkWidget *
messages_create_message_list (AppData* app_data)
{
	GtkCellRenderer     *renderer;
	GtkWidget           *message_view;
	GtkTreeSelection    *selection;

	message_view = gtk_tree_view_new ();
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW (message_view), FALSE);
	//gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (message_view), TRUE);

	/* FIXME: The following is an ugly hack, which improves performance when using 
           the finger scroll by killing prelight redrawing of the entries. */
	g_signal_connect (G_OBJECT(message_view), "enter-notify-event", G_CALLBACK (gtk_true), NULL);
	g_signal_connect (G_OBJECT(message_view), "motion-notify-event", G_CALLBACK (gtk_true), NULL);

	/* Another hack that catches mouse presses and adds modifiers, so that we can select 
	   multiple messages */
	g_signal_connect (G_OBJECT(message_view), "button-press-event", G_CALLBACK (cb_list_click), app_data);
	g_signal_connect (G_OBJECT(message_view), "button-release-event", G_CALLBACK (cb_list_click), app_data);

	renderer = moko_header_cell_renderer_new();

	/* FIXME: Consider this a bug in the renderer */
	g_object_set(G_OBJECT(renderer), "text-spacing", 5, 
					 "icon-spacing", 15, NULL);


	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (message_view),
	                                            -1,
	                                            _("Message"),
	                                            renderer,
	                                            "subject", TNY_GTK_HEADER_LIST_MODEL_SUBJECT_COLUMN,
	                                            "from", TNY_GTK_HEADER_LIST_MODEL_FROM_COLUMN,
	                                            "to", TNY_GTK_HEADER_LIST_MODEL_TO_COLUMN,
	                                            "flags", TNY_GTK_HEADER_LIST_MODEL_FLAGS_COLUMN,
	                                            NULL);
	
	gtk_tree_view_set_model (GTK_TREE_VIEW (message_view), NULL);
	

	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (message_view));
	g_signal_connect (G_OBJECT(selection), "changed", G_CALLBACK(cb_message_select), app_data);

	gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);


	app_data->message_view = GTK_TREE_VIEW(message_view);
	app_data->renderer = renderer;
	app_data->message_selection = selection;

	return message_view;
}

#if 0
/* Callback that puts the gutter between the panes in a better position :) */
void cb_paned_size_allocate (GtkWidget *widget, GtkAllocation *allocation, gpointer data)
{
	/* FIXME: This is a crude hack */
	static gint last_height = 0;
	if (allocation->height > 0 && last_height != allocation->height) {
		gtk_paned_set_position(GTK_PANED(widget), allocation->height * 4 / 10);
		last_height = allocation->height;
	}
}
#endif

#if 0
/* Button callbacks */
/* These could be assigned to a GtkToolButton, or a GtkMenuItem */
static void
cb_message_new_sms (GtkWidget *widget, AppData *data)
{

}
#endif

static void
cb_message_new_mail (GtkWidget *widget, AppData *data)
{
	gint page_num;
	GtkWidget *page;
	page = messages_create_write_email_page(data, 0, NULL, NULL);
	page_num = notebook_add_page_with_icon (data->notebook, page, MOKO_STOCK_MAIL_NEW, 0);
	gtk_notebook_set_current_page (data->notebook, page_num);
}


static void
cb_message_reply_sender (GtkWidget *widget, AppData *data)
{
	if (has_current_message(data)) {
		messages_create_write_email_page_no_msg(data, MESSAGE_WRITE_REPLY_SENDER, data->current_folder, data->current_header);
	}
	return;
}

static void
cb_message_reply_all (GtkWidget *widget, AppData *data)
{
	if (has_current_message(data)) {
		messages_create_write_email_page_no_msg(data, MESSAGE_WRITE_REPLY_ALL, data->current_folder, data->current_header);
	}
	return;
}

/* After asking several people, inline forwarding is prefered to attachment,
 * and also, I believe that it is better for a phone */
static void
cb_message_forward_inline (GtkWidget *widget, AppData *data)
{
	if (has_current_message(data)) {
		messages_create_write_email_page_no_msg(data, MESSAGE_WRITE_FORWARD_INLINE, data->current_folder, data->current_header);
	}
	return;
}

static void
cb_message_forward_attachment (GtkWidget *widget, AppData *data)
{

	if (has_current_message(data)) {
		messages_create_write_email_page_no_msg(data, MESSAGE_WRITE_FORWARD_ATTACHMENT, data->current_folder, data->current_header);
	}
	return;
}

static void
cb_message_open (GtkToolButton *button, AppData *data)
{
	gint page_num;
	GtkWidget *page;
	if (has_current_message(data)) {
		if (tny_header_get_flags(data->current_header) & TNY_HEADER_FLAG_DRAFT) {
			messages_create_write_email_page_no_msg(data, MESSAGE_WRITE_OPEN_DRAFT, data->current_folder, data->current_header);
		}
		else {
			page = messages_create_read_page(data, data->current_folder, data->current_header);
			page_num = notebook_add_page_with_icon (data->notebook, page, "mail-read", 6);
		}
		
	}
}


/* This callback is used when creating a TnyList with the selected messages,
   for deletion, copying, moving, and so on. */
static void 
add_header_to_tnylist (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
                        TnyList *list)
{
	TnyHeader* header;
	gtk_tree_model_get(model,
			  iter,
			  TNY_GTK_HEADER_LIST_MODEL_INSTANCE_COLUMN,
			  &header,
			  -1);
	if (header) {
		tny_list_append(list, G_OBJECT(header));
		g_object_unref(header);
	}
}

/* These callbacks are used to mass-set/unset flags */
static void
set_header_flag (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer flag)
{
	TnyHeader* header;
	gtk_tree_model_get(model,
			  iter,
			  TNY_GTK_HEADER_LIST_MODEL_INSTANCE_COLUMN,
			  &header,
			  -1);
	if (header) {
		tny_header_set_flag (header, GPOINTER_TO_INT(flag));
		g_object_unref(G_OBJECT(header));
		gtk_tree_model_row_changed (model, path, iter);
	}
}

static void
unset_header_flag (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer flag)
{
	TnyHeader* header;
	gtk_tree_model_get(model,
			  iter,
			  TNY_GTK_HEADER_LIST_MODEL_INSTANCE_COLUMN,
			  &header,
			  -1);
	if (header) {
		tny_header_unset_flag (header, GPOINTER_TO_INT(flag));
		g_object_unref(G_OBJECT(header));
		gtk_tree_model_row_changed (model, path, iter);
	}
}


static void
cb_message_delete (GtkWidget *widget, AppData *data)
{
	gtk_tree_selection_selected_foreach (data->message_selection,
	                                     (GtkTreeSelectionForeachFunc) set_header_flag,
	                                     GINT_TO_POINTER(TNY_HEADER_FLAG_DELETED));
	tny_folder_sync_async (data->current_folder, FALSE, NULL, 
			       (TnyStatusCallback) status_update, data);
}


static void
cb_message_undelete (GtkWidget *widget, AppData *data)
{
	gtk_tree_selection_selected_foreach (data->message_selection,
	                                     (GtkTreeSelectionForeachFunc) unset_header_flag,
	                                     GINT_TO_POINTER(TNY_HEADER_FLAG_DELETED));
	tny_folder_sync_async (data->current_folder, FALSE, NULL, 
			       (TnyStatusCallback) status_update, data);
}

static void
cb_message_set_read (GtkWidget *widget, AppData *data)
{
	gtk_tree_selection_selected_foreach (data->message_selection,
	                                     (GtkTreeSelectionForeachFunc) set_header_flag,
	                                     GINT_TO_POINTER(TNY_HEADER_FLAG_SEEN));
	tny_folder_sync_async (data->current_folder, FALSE, NULL, 
			       (TnyStatusCallback) status_update, data);
}


static void
cb_message_set_unread (GtkWidget *widget, AppData *data)
{
	gtk_tree_selection_selected_foreach (data->message_selection,
	                                     (GtkTreeSelectionForeachFunc) unset_header_flag,
	                                     GINT_TO_POINTER(TNY_HEADER_FLAG_SEEN));
	tny_folder_sync_async (data->current_folder, FALSE, NULL, 
			       (TnyStatusCallback) status_update, data);
}

static void
cb_message_destroy (GtkWidget *widget, AppData *data)
{
	TnyList *list;
	GtkWidget *dialog;
	gint dialog_result;
	GError *err;



	quick_dialog_run (dialog, dialog_result, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
			  _("Permanently destroy the selected %d messages?"),
			  gtk_tree_selection_count_selected_rows(data->message_selection));
	

	if (dialog_result == GTK_RESPONSE_YES) {
		list = tny_simple_list_new();
		gtk_tree_selection_selected_foreach (data->message_selection,
		                                     (GtkTreeSelectionForeachFunc) add_header_to_tnylist,
		                                     list);

		tny_folder_remove_msgs (data->current_folder, list, &err);
		if (err) {
			quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
			                 _("Message destruction failed: %s"), err->message);
			g_error_free(err);
		}
		g_object_unref(list);
		tny_folder_sync_async (data->current_folder, FALSE, NULL, 
				       (TnyStatusCallback) status_update, data);
	}

}


static void
cb_message_expunge (GtkWidget *widget, AppData *data)
{
	GtkWidget *dialog;
	gint dialog_result;


	quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
				 _("Expunge the deleted messages?"));

	if (dialog_result == GTK_RESPONSE_YES)
		tny_folder_sync_async (data->current_folder, TRUE, NULL, 
				       (TnyStatusCallback) status_update, data);
}
#if 0
static void
folder_created (TnyFolderStore *self, gboolean cancelled, TnyFolder *new_folder,
                GError *err, AppData *data)
{
	GtkWidget *dialog;
	gint dialog_result;

	if (err) {
		quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
		                 _("Folder creation failed: %s"), err->message);
	}
}
#endif

static void
cb_folder_new (GtkWidget *widget, AppData *data)
{
	GtkTreeIter iter;
	TnyFolder *new_folder;

	gchararray folder_name;
	TnyFolderStore *store;
	TnyFolder *selected;
	gint response;

	GtkWidget *dialog;
	gint dialog_result;

	GError *err;
	

	/* Note that we get the selected folder from the combo, not from our AppData structure,
	   because an account can be selected. This is not very good style, but I'll leave it 
	   this way until the first code cleanup. */
	if (!gtk_combo_box_get_active_iter(data->folder_combo, &iter))
		return;

	
	gtk_tree_model_get(data->folder_list,
			   &iter,
			   TNY_GTK_FOLDER_STORE_TREE_MODEL_INSTANCE_COLUMN,
			   &selected,
			   -1);

	if (!selected)
		return;
	
	/* We create the folder as a child of this folder, if possible, otherwise, we
	   create it at the same level. Not very intuitive, but it works. */

	if (G_LIKELY(TNY_IS_FOLDER_STORE(data->current_folder)))
		store = TNY_FOLDER_STORE(selected);
	else { 
		store = tny_folder_get_folder_store(selected);
		g_object_unref(selected);
	}

	if (!store)
		return;

	response = text_entry_dialog (_("Enter the name of the new folder"), &folder_name, "New Folder");
	if (response == GTK_RESPONSE_OK) {
/* TODO: Switch to async version, when available */
#if 0
		tny_folder_store_create_folder_async (store,
		                                      folder_name,
		                                      (TnyCreateFolderCallback) folder_created,
		                                      (TnyStatusCallback) status_callback,
		                                      data);
#endif
		new_folder = tny_folder_store_create_folder(store, folder_name, &err);
		if (err) {
			quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
			                 _("Folder creation failed: %s"), err->message);
			g_error_free(err);
		}
		if (new_folder)
			g_object_unref(new_folder);
	}
	g_free(folder_name);
	g_object_unref(G_OBJECT(store));

}


static void
cb_folder_delete (GtkWidget *widget, AppData *data)
{
	GtkWidget *dialog;
	gint dialog_result;
	TnyFolderStore *store;
	GError *err;


	quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO,
	                 _("Are you sure you want to permanently delete this folder?"));
	if (dialog_result == GTK_RESPONSE_YES) {
		store = tny_folder_get_folder_store (data->current_folder);
		if (!store) {
			quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
			                 _("This folder cannot be deleted"));
			return;
		}
		tny_folder_store_remove_folder(store, data->current_folder, &err);
		if (err) {
			quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
			                 _("Folder deletion failed: %s"), err->message);
			g_error_free(err);
		}
		/* TODO: I guess that here the folder is automatically changed, and therefore unref'd, but check */
		g_object_unref(store);
	}
}


static void
cb_properties (GtkWidget *widget, AppData *data)
{
	gint page_num;
	GtkWidget *page;
	page = messages_create_config_page(data);
	page_num = notebook_add_page_with_icon (data->notebook, page, GTK_STOCK_PROPERTIES, 6);
	gtk_notebook_set_current_page (data->notebook, page_num);

}

/* Callback for the async creation of a folder */
static void 
messages_transfered (TnyFolder *folder, gboolean cancelled, GError *err, AppData *data)
{
	GtkWidget *dialog;
	gint dialog_result;

	if (err) {
		quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
		                 _("Message transfer failed: %s"), err->message);
	}
}


/* TODO: The following two functions are almost the same ... */
static void
cb_message_move_to_folder (GtkWidget *widget, AppData *data)
{
	TnyFolder *folder;
	TnyList *list;



	if (folder_select_dialog (_("Destination"), &folder, data) == GTK_RESPONSE_OK && folder) {
		list = tny_simple_list_new();
		gtk_tree_selection_selected_foreach (data->message_selection,
		                                     (GtkTreeSelectionForeachFunc) add_header_to_tnylist,
		                                     list);
		tny_folder_transfer_msgs_async (data->current_folder,
		                                list,
		                                folder,
		                                TRUE,  /* Delete originals */
                                                (TnyTransferMsgsCallback) messages_transfered,
                                                (TnyStatusCallback) status_update,
                                                data);
		g_object_unref(list);
		
	}
	if (folder)
		g_object_unref(folder);

	
}

static void
cb_message_copy_to_folder (GtkWidget *widget, AppData *data)
{
	TnyFolder *folder;
	TnyList *list;



	if (folder_select_dialog (_("Destination"), &folder, data) == GTK_RESPONSE_OK && folder) {
		list = tny_simple_list_new();
		gtk_tree_selection_selected_foreach (data->message_selection,
		                                     (GtkTreeSelectionForeachFunc) add_header_to_tnylist,
		                                     list);
		tny_folder_transfer_msgs_async (data->current_folder,
		                                list,
		                                folder,
		                                FALSE,  /* Delete originals */
                                                (TnyTransferMsgsCallback) messages_transfered,
                                                (TnyStatusCallback) status_update,
                                                data);
		g_object_unref(list);
		
	}
	if (folder)
		g_object_unref(folder);

	
}

static void
cb_message_select_toggle (GtkCheckMenuItem *item, AppData *data)
{
	if (gtk_check_menu_item_get_active(item)) {
		//gtk_tree_selection_set_mode(data->message_selection, GTK_SELECTION_MULTIPLE);
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(data->shift_toggle), FALSE);
		data->modifiers = GDK_CONTROL_MASK;
		gtk_widget_show(data->selector);
	}
	else {	
		//gtk_tree_selection_set_mode(data->message_selection, GTK_SELECTION_SINGLE);
		data->modifiers = 0;
		gtk_widget_hide(data->selector);
	}
}


static GtkWidget*
get_menu_message_new (AppData* app_data)
{
	GtkWidget *menu, *item;
	menu = gtk_menu_new();

	item = gtk_menu_item_new_with_label (_("New mail"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_new_mail), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());


	item = gtk_menu_item_new_with_label (_("Reply to Sender"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_reply_sender), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	item = gtk_menu_item_new_with_label (_("Reply to All"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_reply_all), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
	
	item = gtk_menu_item_new_with_label (_("Forward as Inline"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_forward_inline), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	item = gtk_menu_item_new_with_label (_("Forward as Attachment"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_forward_attachment), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_widget_show_all(menu);
	return menu;
}
/* TODO: These menus are not very intuitive */
/* FIXME: Not working options are currently disabled */
static GtkWidget*
get_menu_message_delete (AppData* app_data)
{
	GtkWidget *menu, *item;
	menu = gtk_menu_new();


	item = gtk_menu_item_new_with_label (_("Delete selected messages"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_delete), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);


	item = gtk_menu_item_new_with_label (_("Undelete selected messages"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_undelete), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);


#if 0
	item = gtk_menu_item_new_with_label (_("Destroy selected messages"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_destroy), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);
#endif
	item = gtk_menu_item_new_with_label (_("Expunge deleted messages"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_expunge), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());

	item = gtk_menu_item_new_with_label (_("Set as read"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_set_read), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);


	item = gtk_menu_item_new_with_label (_("Set as unread"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_set_unread), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());

	item = gtk_check_menu_item_new_with_label (_("Select mode"));
	g_signal_connect(G_OBJECT(item), "toggled", G_CALLBACK(cb_message_select_toggle), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_widget_show_all(menu);
	return menu;
}


static GtkWidget*
get_menu_properties (AppData* app_data)
{
	GtkWidget *menu, *item;
	menu = gtk_menu_new();

	item = gtk_menu_item_new_with_label (_("Properties"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_properties), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());

	item = gtk_menu_item_new_with_label (_("New folder"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_folder_new), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	item = gtk_menu_item_new_with_label (_("Delete folder"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_folder_delete), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
	
	item = gtk_menu_item_new_with_label (_("Copy to folder"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_copy_to_folder), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);


	item = gtk_menu_item_new_with_label (_("Move to folder"));
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(cb_message_move_to_folder), app_data);
	gtk_menu_shell_append (GTK_MENU_SHELL(menu), item);

	gtk_widget_show_all(menu);
	return menu;
}


/* Small virtual scroller and keyboard buttons */
static void
cb_virt_up (GtkButton *button, AppData *data)
{
	GtkAdjustment *adj = gtk_tree_view_get_vadjustment (data->message_view);
	gtk_adjustment_set_value(adj, adj->value - adj->page_increment);
	gtk_adjustment_value_changed(adj);
}

static void
cb_virt_down (GtkButton *button, AppData *data)
{
	GtkAdjustment *adj = gtk_tree_view_get_vadjustment (data->message_view);
	gtk_adjustment_set_value(adj, adj->value + adj->page_increment);
	gtk_adjustment_value_changed(adj);
}

/* Enable selecting contagious selection regions, that is, holds "shift" */
static void
cb_virt_continuous (GtkToggleButton *button, AppData *data)
{
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(button)))
		data->modifiers = GDK_SHIFT_MASK;
	else
		data->modifiers = GDK_CONTROL_MASK;
}

GtkWidget*
messages_create_list_page (AppData* app_data)
{
	GtkWidget *page, *toolbar, *button, *folders, *messages, *list_scroll;

	page = gtk_vbox_new (FALSE, 0);

	/* We create the toolbar */
	toolbar = gtk_toolbar_new ();


	toolbar_insert_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (MOKO_STOCK_MAIL_READ, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_message_open), app_data);
	toolbar_insert_menu_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (MOKO_STOCK_MAIL_NEW, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_message_new_mail), app_data, get_menu_message_new(app_data));
	toolbar_insert_menu_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (MOKO_STOCK_MAIL_DELETE, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_message_delete), app_data, get_menu_message_delete(app_data));
	toolbar_insert_menu_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (GTK_STOCK_PROPERTIES, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_properties), app_data, get_menu_properties(app_data));
	gtk_box_pack_start (GTK_BOX (page), toolbar, FALSE, TRUE, 0);


	/* We create the folder list */
	messages_init_folder_list(app_data);
	folders = messages_create_folder_list(app_data);
	gtk_box_pack_start (GTK_BOX(page), folders, FALSE, TRUE, 0);

	/* We create the message list */
	messages = messages_create_message_list(app_data);
	


	//list_viewport = gtk_viewport_new(NULL, NULL);
	//gtk_viewport_set_shadow_type (GTK_VIEWPORT (list_viewport), GTK_SHADOW_NONE);

	
	list_scroll = moko_finger_scroll_new();
	//list_scroll = gtk_scrolled_window_new (NULL, NULL);
	//gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (list_scroll), GTK_SHADOW_NONE);
	//gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (list_scroll), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	//gtk_container_add (GTK_CONTAINER (list_viewport), messages);	
	gtk_container_add (GTK_CONTAINER (list_scroll), messages);
	
	

	gtk_widget_show_all(list_scroll);
	gtk_widget_show_all(page);
	
	toolbar = gtk_hbox_new(FALSE, 0);

	button = gtk_toggle_button_new_with_label(_("Continuous"));
	gtk_button_set_focus_on_click (GTK_BUTTON(button), FALSE);
	g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(cb_virt_continuous), app_data);
	gtk_box_pack_start(GTK_BOX(toolbar), button, TRUE, TRUE, 0);
	gtk_widget_show(button);
	

	app_data->selector = toolbar;
	app_data->shift_toggle = button;

	button = gtk_button_new();
	gtk_button_set_image(GTK_BUTTON(button), 
			     gtk_image_new_from_icon_name (GTK_STOCK_GO_UP,
							   GTK_ICON_SIZE_MENU));
	gtk_button_set_focus_on_click (GTK_BUTTON(button), FALSE);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(cb_virt_up), app_data);
	gtk_box_pack_start(GTK_BOX(toolbar), button, FALSE, TRUE, 0);
	gtk_widget_show(button);


	button = gtk_button_new();
	gtk_button_set_image(GTK_BUTTON(button), 
			     gtk_image_new_from_icon_name (GTK_STOCK_GO_DOWN,
							   GTK_ICON_SIZE_MENU));
	gtk_button_set_focus_on_click (GTK_BUTTON(button), FALSE);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(cb_virt_down), app_data);
	gtk_box_pack_start(GTK_BOX(toolbar), button, FALSE, TRUE, 0);
	gtk_widget_show(button);

	
	
	
	gtk_box_pack_end(GTK_BOX(page), toolbar, FALSE, TRUE, 0);
	gtk_box_pack_end (GTK_BOX (page), list_scroll, TRUE, TRUE, 0);

	return page;

	

}
