/*
 * Mail for OpenMoko
 * Copyright (C) 2007 Milko Krachounov <exabyte@3mhz.net>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <unistd.h>
#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <glib/gstdio.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <strings.h>
#include <string.h>
#include <stdarg.h>
#include <moko-finger-scroll.h>
#include <moko-stock.h>

#include <tny-gtk-text-buffer-stream.h>
#include <tny-gtk-account-list-model.h>
#include <tny-header.h>
#include <tny-folder.h>
#include <tny-msg.h>
#include <tny-simple-list.h>
#include <tny-fs-stream.h>
#include <tny-camel-send-queue.h>
#include <tny-folder-emitter.h>
#include <tny-openmoko-account-store.h>

#include "mokofieldlabel.h"
#include "mail-defines.h"
#include "mail-common.h"
#include "mail-write.h"


enum {
	ATTACHMENT_INTERNAL = 1,
	ATTACHMENT_FILE
};
typedef struct {
	/* Internal attachment is one contained in a TnyMimeType */
	gint type;
	union {
		TnyMimePart *part;	
		struct {
			gchararray path;
			gchararray filename;
			gssize size;
		};
	};
} MailAttachment;

static void
free_mail_attachment (MailAttachment *attachment, void *ptr)
{
	switch(attachment->type) {
		case ATTACHMENT_INTERNAL:
			if (attachment->part)
				g_object_unref(attachment->part);
			break;
	
		case ATTACHMENT_FILE:
			if (attachment->path)
				g_free(attachment->path);
			if (attachment->filename)
				g_free(attachment->filename);	
			break;
	}
	g_free(attachment);
}

static MailAttachment*
new_mail_attachment (gint type, ...)
{
	va_list ap;
	va_start(ap, type);
	
	MailAttachment *ret_val;
	
	ret_val = g_malloc0(sizeof(ret_val));
	ret_val->type = type;
	switch (type) {
		case ATTACHMENT_INTERNAL:
			ret_val->part = g_object_ref(va_arg(ap, TnyMimePart*));
			break;
		case ATTACHMENT_FILE:
			ret_val->path = g_strdup(va_arg(ap, const gchar*));
			ret_val->filename = g_strdup(va_arg(ap, const gchar*));
			break;
	}
	va_end(ap);
	return ret_val;
}

typedef struct {
	AppData *app_data;
	GtkTextBuffer *text;
	TnyStream *text_stream;
	GList *attachments;
	GtkWidget *page;
	union {
		struct {
			GtkWidget *tofrom;
		};
		struct {
			GtkWidget *from;
			GtkWidget *account;
			GtkWidget *to;
			GtkWidget *subject;
			GtkWidget *cc;
			GtkWidget *bcc;
			GtkWidget *attachment_box;
		};
			
	} widgets;

	TnyFolder *draft_folder;
	TnyHeader *draft_header;

} MessagesWritePageData;

/* NOTE: Call this with button = NULL on successfull sent! */
static void
cb_message_close (GtkToolButton *button, MessagesWritePageData *data)
{
	GtkWidget *dialog;
	gint dialog_result;

	/* TODO: Entering data into to and from fields should be considered too, but I'm too lazy to code it */
	if (button && (gtk_text_buffer_get_char_count (data->text) > 0 || data->attachments)) {
		quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
		_("Close the message?"));
		if (dialog_result != GTK_RESPONSE_YES)
			return;
	}
	/* If button is NULL than this wasn't called from the user, and the message was sent. Delete the draft. */
	if (!button && data->draft_folder && data->draft_header)
	{
		tny_folder_remove_msg(data->draft_folder, data->draft_header, NULL);
	}

	gtk_widget_destroy(data->page);

	g_object_unref(data->text);
	g_object_unref(data->text_stream);

	g_list_foreach (data->attachments, (GFunc) free_mail_attachment, NULL);
	g_list_free(data->attachments);

	g_free(data);
}



static const gchar*
get_mime_type (const gchar *filename)
{
	return "application/octet-stream";
}

/* 0. Helper functions for processing draft folder */
/* This function sets the TNY_HEADER_FLAG_DRAFT on the saved message */
static void
fix_saved_message (TnyFolderObserver *self, TnyFolderChange *change, AppData *data)
{
	if (tny_folder_change_get_changed(change) & TNY_FOLDER_CHANGE_CHANGED_ADDED_HEADERS)  {
		TnyFolder *folder;
		TnyList *headers;
		TnyIterator *iter;

		folder = tny_folder_change_get_folder(change);

		headers = tny_simple_list_new();

		tny_folder_change_get_added_headers (change, headers);
		
		iter = tny_list_create_iterator (headers);
		while (!tny_iterator_is_done (iter)) {
			TnyHeader *header = TNY_HEADER(tny_iterator_get_current(iter));
			tny_header_set_flag(header, TNY_HEADER_FLAG_DRAFT);
	
			g_object_unref (header);
			tny_iterator_next (iter);
		}
		g_object_unref(iter);
		g_object_unref(headers);
		

		tny_folder_sync_async (folder, FALSE, NULL, 
			               (TnyStatusCallback) status_update, data);
		g_object_unref(folder);

		/* We unreference the observer, so that memory is freed if unused. 
		   FIXME: Two issues: 1. Keeping the observer might be more optimal.
		                      2. It is actually NOT guaranteed that it is actually unref'd
		*/
		g_object_unref(self);
		
	}
}

/* This function sets a folder observer setting the TNY_HEADER_FLAG_DRAFT draft on each message. */
static void
prepare_draft_folder (TnyFolder *folder, AppData *data)
{
	FolderObserverData *observer_data;

	observer_data = g_hash_table_lookup(data->observers, folder);

	if (!observer_data) {
		observer_data = g_malloc(sizeof(FolderObserverData));
		observer_data->observer = tny_folder_emitter_new();
	
		g_hash_table_insert(data->observers, g_object_ref(folder), observer_data);

		g_signal_connect(G_OBJECT(observer_data->observer), "folder-changed",
				 G_CALLBACK(fix_saved_message), data);

		g_object_weak_ref(G_OBJECT(observer_data->observer), 
		                  (GWeakNotify) g_hash_table_remove, data->observers);
		
			
	}
	else {
		g_object_ref(G_OBJECT(observer_data->observer));
	}
	
}

/* 1. Set of functions and declarations for saving a message to a folder (for drafts) */
typedef struct {
	AppData *app_data;
	TnyMsg *msg;
} SaveMsgData;

static void
free_save_msg_data(SaveMsgData *data)
{
	g_object_unref(data->msg);
	g_free(data);
}

static void
message_save_status (GObject *sender, TnyStatus *status, SaveMsgData *data)
{
	status_update(sender, status, data->app_data);
}

static void
message_save_callback (TnyFolder *self, gboolean cancelled, GError *err, SaveMsgData *data)
{
	
	if (err) {
		GtkWidget *dialog;
		gint result;
		quick_dialog_run(dialog, result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
				 _("Saving to the sent folder failed: %s\n"), err->message);
	}
	
	free_save_msg_data(data);
}

/* NOTE: This is written for drafts, DO NOT use for anything else */
static void
message_save_to_folder	(TnyOpenMokoAccountStore *self,
			 gboolean cancelled,
			 TnyFolder *folder,
			 GError *err,
			 SaveMsgData *data)
{
	if (!err && !cancelled && folder) {

		prepare_draft_folder(folder, data->app_data);
		tny_folder_add_msg_async (folder,
 					  data->msg,
					  (TnyFolderCallback) message_save_callback,
					  (TnyStatusCallback) message_save_status,
					  data);
		return;
	}
	if (err) {
		GtkWidget *dialog;
		gint result;
		quick_dialog_run(dialog, result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
				 _("Saving message to draft folder failed: %s\n"), err->message);
	}
	free_save_msg_data(data);

}


/* 2. Set of functions and declarations for moving a message to a folder (for sent mail) */
typedef struct {
	AppData *app_data;
	TnyFolder *src;
	TnyList *headers;
} MoveMsgData;

static void
free_move_msg_data(MoveMsgData *data)
{
	g_object_unref(data->src);
	g_object_unref(data->headers);
	g_free(data);
}

static void
message_move_status (GObject *sender, TnyStatus *status, MoveMsgData *data)
{
	status_update(sender, status, data->app_data);
}

static void
message_move_callback (TnyFolder *self, gboolean cancelled, GError *err, MoveMsgData *data)
{
	
	if (err) {
		GtkWidget *dialog;
		gint result;
		quick_dialog_run(dialog, result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
				 _("Saving to the sent folder failed: %s\n"), err->message);
	}
	
	free_move_msg_data(data);
}

static void
message_move_to_folder	(TnyOpenMokoAccountStore *self,
			 gboolean cancelled,
			 TnyFolder *folder,
			 GError *err,
			 MoveMsgData *data)
{
	if (!err && !cancelled && folder) {
		tny_folder_transfer_msgs_async      (data->src,
						     data->headers,
						     folder,
						     TRUE,
						     (TnyTransferMsgsCallback) message_move_callback,
						     (TnyStatusCallback) message_move_status,
						     data);
		return;
	}
	if (err) {
		GtkWidget *dialog;
		gint result;
		quick_dialog_run(dialog, result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
				 _("Getting the sent folder failed: %s\n"), err->message);
	}
	free_move_msg_data(data);

}


/* 3. Actually moving a send mail... */

/* Move the sent message from the internal folder to the configured folder */
static void
move_sent_message (TnyFolderObserver *self, TnyFolderChange *change, AppData *data)
{
	if (tny_folder_change_get_changed(change) & TNY_FOLDER_CHANGE_CHANGED_ADDED_HEADERS)  {
		MoveMsgData *move_data = g_malloc0(sizeof(MoveMsgData));
		FolderObserverData *observer_data;

		move_data->app_data = data;
		move_data->src = tny_folder_change_get_folder(change);
		move_data->headers = tny_simple_list_new();

		tny_folder_change_get_added_headers (change, move_data->headers);

		/* Note that we carry the account as an additional data with the observer. */
		observer_data = g_hash_table_lookup(data->observers, move_data->src);	

		tny_openmoko_acount_store_find_folder_async 
					(TNY_OPENMOKO_ACCOUNT_STORE(data->account_store),
					 TNY_ACCOUNT(observer_data->data),
					 TNY_FOLDER_TYPE_SENT,
					 TRUE,
					 (TnyOpenMokoFindFolderCallback) message_move_to_folder,
					 (TnyStatusCallback) message_move_status,
					  move_data);
		/* We unreference the observer, so that memory is freed if unused. 
		   FIXME: Two issues: 1. Keeping the observer might be more optimal.
		                      2. It is actually NOT guaranteed that it is actually unref'd
		*/
		g_object_unref(observer_data->observer);
		
	}
}

/* We wait for the message to arrive in the sentbox of the queue and move it to our sent
   folder */
static void 
redirect_sent_folder (TnyAccount *account, TnySendQueue *queue, AppData *data)
{
	TnyFolder *folder;
	FolderObserverData *observer_data;
	
	
	folder = tny_send_queue_get_sentbox(queue);

	observer_data = g_hash_table_lookup(data->observers, folder);

	if (!observer_data) {
		observer_data = g_malloc(sizeof(FolderObserverData));
		observer_data->observer = tny_folder_emitter_new();
		observer_data->data = account;
	
		g_hash_table_insert(data->observers, g_object_ref(folder), observer_data);

		g_signal_connect(G_OBJECT(observer_data->observer), "folder-changed",
			 G_CALLBACK(move_sent_message), data);
		
		g_object_weak_ref(G_OBJECT(observer_data->observer), 
		                  (GWeakNotify) g_hash_table_remove, data->observers);
	}
	else {
		g_object_ref(G_OBJECT(observer_data->observer));
	}
	

	g_object_unref(G_OBJECT(folder));
}
/* End of save and move functions */

/* If send_message is TRUE, send message, else save as a draft */
static void
construct_and_process_message (MessagesWritePageData *data, gboolean send_message)
{
	gchararray from, replyto, signature;
	const gchar *to, *cc, *bcc, *subject;
	const gchar *mime_type = "text/plain; charset=UTF-8";

	TnyTransportAccount *account;
	TnySendQueue *queue;
	TnyMsg *msg;
	TnyHeader *header;
	TnyMimePart *part;
	

	GList *walk;

	int fd;		
	TnyStream *stream;	

	gint oldend = -1, dialog_result;
	GtkTextIter start, end;

	GtkTreeIter iter;

	GtkWidget *dialog;

	gboolean close_page = FALSE;

	GError *err = NULL;


	/* 1. We read the data from our widgets, perform checking and print warning and errors */
	if (G_UNLIKELY(!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(data->widgets.from), &iter))) {
		quick_dialog_run (dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
				  _("You have not choosen an identity"));
		goto abort_send_no_id;
	}
	
	/* FIXME: We take for granted that these are correct */
	gtk_tree_model_get(data->app_data->identities,
			   &iter,
			   IDENTITIES_COLUMN_NAME,	&from,
			   IDENTITIES_COLUMN_REPLYTO,	&replyto,
			   IDENTITIES_COLUMN_SIGNATURE,	&signature,
			   -1);


	if (G_UNLIKELY(!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(data->widgets.account), &iter))) {
		quick_dialog_run (dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
				  _("You have not choosen an outgoing account"));
		goto abort_send_no_account;
	}

	to = gtk_entry_get_text(GTK_ENTRY(data->widgets.to));
	cc = gtk_entry_get_text(GTK_ENTRY(data->widgets.cc));
	bcc = gtk_entry_get_text(GTK_ENTRY(data->widgets.bcc));
	subject = gtk_entry_get_text(GTK_ENTRY(data->widgets.subject));

	if (!is_address_field_valid(to, FALSE)) {
		invalid_field_dialog(dialog, dialog_result, _("To"));
		goto abort_send_fields_error;
	}
	
	if (cc && *cc && !is_address_field_valid(cc, FALSE)) {
		invalid_field_dialog(dialog, dialog_result, _("Cc"));
		goto abort_send_fields_error;
	}
	
	if (bcc && *bcc && !is_address_field_valid(bcc, FALSE)) {
		invalid_field_dialog(dialog, dialog_result, _("Bcc"));
		goto abort_send_fields_error;
	}
	

	/* TODO: We should do more checking here. */
	if (!subject || !*subject) {
		quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO,
				 _("Are you sure you want to send the message with an empty subject?"));
		if (dialog_result != GTK_RESPONSE_YES)
			goto abort_send_fields_error;
	}

	/* 2. We begin to construct our message */

	msg = tny_platform_factory_new_msg(data->app_data->platfact);
	header = tny_msg_get_header(msg);

	tny_header_set_from(header, from);
	tny_header_set_to(header, to);
	if (cc && *cc)
		tny_header_set_cc(header, cc);
	if (bcc && *bcc)
		tny_header_set_bcc(header, bcc);
	if (replyto && *replyto)
		tny_header_set_replyto(header, replyto);
	if (subject && *subject)
		tny_header_set_subject(header, subject);

	/* We append the signature to the text buffer. On abort, we'll simply remove it :) */
	if (signature && *signature) {
		gtk_text_buffer_get_end_iter (data->text, &end);
		/* We save the position at which we inserted the signature */
		oldend = gtk_text_iter_get_offset (&end);
		gtk_text_buffer_insert (data->text, &end, "\n\n-- \n", -1);
		gtk_text_buffer_insert (data->text, &end, signature, -1);
	}

	tny_stream_reset(data->text_stream);

	if (data->attachments)	{
		part = tny_platform_factory_new_mime_part(data->app_data->platfact);
		if (tny_mime_part_construct_from_stream (part, data->text_stream, mime_type) == -1)
		{
			quick_dialog_run (dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
					  _("An error occured while constructing the message"));
			g_object_unref(part);
			goto abort_message_error;
		}
		tny_mime_part_add_part(TNY_MIME_PART(msg), part);
		g_object_unref(part);
		for (walk = data->attachments; walk; walk = g_list_next(walk)) {
			MailAttachment *attachment = walk->data;
			switch (attachment->type) {
				case ATTACHMENT_INTERNAL:
					tny_mime_part_add_part(TNY_MIME_PART(msg), attachment->part);
					break;
				case ATTACHMENT_FILE:
					part = tny_platform_factory_new_mime_part(data->app_data->platfact);
					fd = g_open(attachment->path, O_RDONLY);
					if (fd == -1) {
						g_object_unref(part);
						quick_dialog_run(dialog, dialog_result, 
								 GTK_MESSAGE_ERROR, GTK_BUTTONS_YES_NO,
								 _("Opening %s for reading failed. Do you want send the mail without this attachment?"), attachment->filename);
						if (dialog_result != GTK_RESPONSE_YES)
							goto abort_message_error;
						break;
						
					}
					stream = tny_fs_stream_new (fd);
					if (tny_mime_part_construct_from_stream (part, stream, get_mime_type(attachment->path)) == -1)
					{
						close(fd);
						g_object_unref(stream);
						g_object_unref(part);
						quick_dialog_run(dialog, dialog_result, 
								 GTK_MESSAGE_ERROR, GTK_BUTTONS_YES_NO,
								 _("An error occured while adding %s to the message. Do you want send the mail without this attachment?"), attachment->filename);
						if (dialog_result != GTK_RESPONSE_YES)
							goto abort_message_error;
						break;
					}

					tny_mime_part_set_filename(part, attachment->filename);
					tny_mime_part_add_part(TNY_MIME_PART(msg), part);

					close(fd);
					g_object_unref(G_OBJECT(stream));
					g_object_unref(G_OBJECT(part));			
					break;
			}
		}
	}
	else {
		if (tny_mime_part_construct_from_stream (TNY_MIME_PART(msg), data->text_stream, mime_type) == -1)
		{
			quick_dialog_run (dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
					  _("An error occured while constructing the message"));
			goto abort_message_error;
		}
	}
	
	if (send_message) {
		gtk_tree_model_get(data->app_data->transport_accounts,
				   &iter,
				   TNY_GTK_ACCOUNT_LIST_MODEL_INSTANCE_COLUMN,
				   &account,
				   -1);
	
		queue = get_send_queue(account, data->app_data);
	
		/* We want to redirect the sent folder to the user configured location */
		redirect_sent_folder (TNY_ACCOUNT(account), queue, data->app_data);
		/* TODO: Camel async adding? */
		tny_send_queue_add(queue, msg, &err);
		if (err) {
			quick_dialog_run(dialog, dialog_result, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
					 _("An error occured, while adding message to queue: %s"),
					 err->message);
			g_error_free(err);
			goto abort_message_error;
		}
		
		/* FIXME: This camel should be wrapped somewhere, I guess */
		tny_camel_send_queue_flush(TNY_CAMEL_SEND_QUEUE(queue));
	}
	else {
		SaveMsgData *save_data = g_malloc0(sizeof(SaveMsgData));
		save_data->app_data = data->app_data;
		save_data->msg = g_object_ref(msg);
		tny_openmoko_acount_store_find_folder_async 
				(TNY_OPENMOKO_ACCOUNT_STORE(data->app_data->account_store),
				 TNY_ACCOUNT(account),
				 TNY_FOLDER_TYPE_DRAFTS,
				 TRUE,
				 (TnyOpenMokoFindFolderCallback) message_save_to_folder,
				 (TnyStatusCallback) message_save_status,
				 save_data);
	}

	/* If we reached this point, we successfully put the message on the queue, so we close the page */
	close_page = TRUE;


abort_message_error:
	g_object_unref(G_OBJECT(msg));
	if (!close_page && oldend >= 0) {
		gtk_text_buffer_get_iter_at_offset  (data->text, &start, oldend);
		gtk_text_buffer_get_end_iter (data->text, &end);
		gtk_text_buffer_delete (data->text, &start, &end);
	}
abort_send_fields_error:
	g_object_unref(G_OBJECT(account));
abort_send_no_account:
	g_free(from);
	g_free(replyto);
	g_free(signature);
abort_send_no_id:
	if (close_page)
		cb_message_close(NULL, data);
}

static void
cb_send_email (GtkToolButton *button, MessagesWritePageData *data)
{
	construct_and_process_message (data, TRUE);
}

static void
cb_save_draft (GtkToolButton *button, MessagesWritePageData *data)
{
	construct_and_process_message (data, FALSE);
}


static void
cb_id_select (GtkComboBox *combo, MessagesWritePageData *data)
{	
	GtkTreeIter iter;
	TnyTransportAccount *default_account, *account;
	
	if (!gtk_combo_box_get_active_iter(combo, &iter))
		return;

	
	gtk_tree_model_get(data->app_data->identities,
			   &iter,
			   IDENTITIES_COLUMN_ACCOUNT,
			   &default_account,
			   -1);

	if (gtk_tree_model_get_iter_first (data->app_data->transport_accounts, &iter)) do
	{
		gtk_tree_model_get(data->app_data->transport_accounts,
				   &iter,
				   TNY_GTK_ACCOUNT_LIST_MODEL_INSTANCE_COLUMN,
				   &account,
				   -1);
		
		if (default_account == account)	{	
			gtk_combo_box_set_active_iter(GTK_COMBO_BOX(data->widgets.account), &iter);
			g_object_unref(account);
			g_object_unref(default_account);
			return;
		}
		g_object_unref(account);
	} while (gtk_tree_model_iter_next (data->app_data->transport_accounts, &iter));

	g_object_unref(default_account);
}

/* A function that automatically chooses an identity for the initial selection 
 * If an account is selected, the default identity for the account is choosen,
 * otherwise the default identity is selected, if not available, the identity for
 * the currently open one in the list page is choosen.
 * Then if there is a hint specified, try to find a better identity based on the hint.
 * If no identity was found, try to find one solely based on the hint.
 * If still no identity was found, select one if exists
 */

static void
auto_select_identity (TnyAccount *account, const gchar *hint, MessagesWritePageData *data, gboolean hint_only)
{
	TnyAccount *transport_account = NULL;
	const gchar *account_identity = NULL, *default_identity;
	gboolean prefer_account = account != NULL;
	GtkTreeIter iter;
	gboolean selected = FALSE;
	
	if (!hint_only) {
		if (account) 
			g_object_ref(account);
		else if (data->app_data->current_folder) {
			account = tny_folder_get_account(data->app_data->current_folder);
		}
	
		if (account)
			account_identity = tny_openmoko_account_store_get_identity(TNY_OPENMOKO_ACCOUNT_STORE(data->app_data->account_store), account);
		
		default_identity = data->app_data->default_identity;
	
		/* Not exactly needed, but saves us from some unneeded checks :) */
		if (account_identity && default_identity && !strcmp(default_identity, account_identity))
			account_identity = NULL;
		
		if (!account_identity)
			prefer_account = FALSE;
		
		/* We walk through the identites and choose our prefered in the combo, and save its transport account */
		if (gtk_tree_model_get_iter_first (data->app_data->identities, &iter)) do
		{
			gchar *identity;
			TnyAccount *walk_account;
	
			gtk_tree_model_get(data->app_data->identities,
					   &iter,
					   IDENTITIES_COLUMN_ID,
					   &identity,
					   IDENTITIES_COLUMN_ACCOUNT,
					   &walk_account,
					   -1);
			if (!identity) {
				if (walk_account)
					g_object_unref(walk_account);
				continue;
			}
			if (default_identity && !strcmp(identity, default_identity)) {	
				gtk_combo_box_set_active_iter(GTK_COMBO_BOX(data->widgets.from), &iter);
				selected = TRUE;
				if (transport_account)
					g_object_unref(transport_account);
				if (walk_account)
					transport_account = g_object_ref(walk_account);
				else
					transport_account = NULL;
	
				if (!prefer_account) {
					if (walk_account)
						g_object_unref(walk_account);
					g_free(identity);
					break;
				}
			}
			if (account_identity && !strcmp(identity, account_identity)) {
				gtk_combo_box_set_active_iter(GTK_COMBO_BOX(data->widgets.from), &iter);
				selected = TRUE;
				if (transport_account)
					g_object_unref(transport_account);
				if (walk_account)
					transport_account = g_object_ref(walk_account);
				else
					transport_account = NULL;
	
				if (prefer_account) {
					if (walk_account)
						g_object_unref(walk_account);
					g_free(identity);
					break;
				}
			}
			if (walk_account)
				g_object_unref(walk_account);
			g_free(identity);
		} while (gtk_tree_model_iter_next (data->app_data->identities, &iter));
	}

	/* If we have a hint, we search for an identity matching hint, but having the same send account */
	if (hint && gtk_tree_model_get_iter_first (data->app_data->identities, &iter)) do
	{
		gchar *name;
		TnyAccount *walk_account;

		gtk_tree_model_get(data->app_data->identities,
				   &iter,
				   IDENTITIES_COLUMN_NAME,
				   &name,
				   IDENTITIES_COLUMN_ACCOUNT,
				   &walk_account,
				   -1);
		if (!name) {
			if (walk_account)
				g_object_unref(walk_account);
			continue;
		}
		if ((!transport_account || walk_account == transport_account) && !strcmp(name, hint)) {	
			gtk_combo_box_set_active_iter(GTK_COMBO_BOX(data->widgets.from), &iter);
			selected = TRUE;
			if (walk_account)
				g_object_unref(walk_account);
			g_free(name);
			break;
		}
		if (walk_account)
			g_object_unref(walk_account);
		g_free(name);
	} while (gtk_tree_model_iter_next (data->app_data->identities, &iter));

	if (transport_account)
		g_object_unref(transport_account);

	if (account)
		g_object_unref(account);
	/* Finally, if nothing was selected so far select the first one */
	if (!selected && gtk_tree_model_get_iter_first (data->app_data->identities, &iter)) {
		gtk_combo_box_set_active_iter(GTK_COMBO_BOX(data->widgets.from), &iter);
	}
}


/* Mail function */

/* The following functions are used to quote a message when forwarding or replying,
   they are currently usable only in plain text mails?
*/


/* This must be freed */
/* This function creates a short header when forwarding a message */
static gchararray
forward_mail_header(TnyHeader *header)
{
#if 0
	return g_strdup_printf("\n\n%s %s %s\n%s %s\n%s %s\n%s %s\n%s %s\n%s %s\n",
			"--------", _("Original letter"), "--------",
			_("From:"), tny_header_get_from (header),
			_("Subject:"), tny_header_get_subject (header),
			_("Date:"), date_human_readable(tny_header_get_date_sent(header)),
			_("To:"), tny_header_get_to (header),
			_("Cc:"), tny_header_get_cc (header));
#endif
	gsize i, j, len;
	gchararray result;

	

	const gchar* label_original = N_("Original letter");
	const gchar* bar = "--------";
	
	struct {
		const gchar* label;
		const gchar* text;
	} fields[] = {
		{ N_("From:"), tny_header_get_from (header) },
		{ N_("Subject:"), tny_header_get_subject (header) },
		{ N_("Date:"), date_human_readable(tny_header_get_date_sent(header)) },
		{ N_("To:"), tny_header_get_to (header) },
		//{ N_("Cc:"), tny_header_get_cc (header) },
		{ NULL, NULL }
	};
	len = strlen(bar)*2 + strlen(_(label_original)) + 5; /* 5 = 2 spaces + 3 newlines */
	for (i = 0; fields[i].label; i++)
		if (fields[i].text && fields[i].text[0])
			len += strlen(_(fields[i].label)) + strlen(fields[i].text) + 2;
	len = len + 1; /* The \0 character at the end */
	/* Be careful here! */
	result = g_malloc(sizeof(gchar)*len);
	j = snprintf(result, len, "\n\n%s %s %s\n", bar, _(label_original), bar);
	for (i = 0; fields[i].label; i++)
		if (fields[i].text && fields[i].text[0])
			j += snprintf(result + j, len - j, "%s %s\n", fields[i].label, fields[i].text);
	/* Just to be sure */
	if(result[len-1] != '\0')
		result[len-1] = '\0';

	return result;
}

static inline gchararray
reply_mail_header (TnyHeader* header)
{
	return g_strdup_printf(_("\n\n%s wrote: \n"), tny_header_get_from (header));
}

static void
text_counter(const gchar * text, gsize *len, gsize *lines)
{
	gsize i, count = 0;
	for (i = 0; text[i]; i++)
		if (text[i] == '\n') count++;
	*len = i;
	*lines = count;

}

/* This must be freed */
static gchararray 
quote_and_free_text(gchararray text, const gchar* header)
{
	gsize len, lines, i, j;
	gchararray result;
	text_counter(text, &len, &lines);
	j = strlen(header);
	len = len+lines*2+j+3; /* Two characters plus null char! */
	/* Be careful here ! */
	result = g_malloc(sizeof(gchar)*len);
	strncpy(result, header, j);
	if (j < len-1)
		result[j++] = '>';
	if (j < len-1)
		result[j++] = ' ';
	for (i = 0; text[i] && j < len-1; i++, j++)
	{
		result[j] = text[i];
		if (text[i] == '\n') {
			if (text[i+1])
				result[++j] = '>';
				result[++j] = ' ';
		}
	}
	result[j] = '\0';

	g_free(text);
	return result;
}

static void
decode_parts(TnyMimePart *part, TnyStream *text_stream, GList **attachments)
{
	TnyIterator *iter;
	TnyList *part_list;
	GList *att_list = NULL;
	
	part_list = tny_simple_list_new ();
	
	tny_mime_part_get_parts (part, part_list);

	iter = tny_list_create_iterator (part_list);
	for (iter = tny_list_create_iterator (part_list);
	     !tny_iterator_is_done(iter);
	     tny_iterator_next(iter))
	{
		part = TNY_MIME_PART(tny_iterator_get_current (iter));
		if (tny_mime_part_is_attachment(part)) {
			if (attachments)
				att_list = g_list_prepend(att_list,
						 new_mail_attachment(ATTACHMENT_INTERNAL,
								     part));
		}
		/* TODO: Think of all possible cases and the best ways to handle them */
		else if(tny_mime_part_content_type_is (part, "multipart/alternative")) {
			decode_parts(part, text_stream, NULL);
		}
		else if(tny_mime_part_content_type_is (part, "text/plain") && text_stream)
			tny_mime_part_decode_to_stream(part, text_stream);
		g_object_unref (G_OBJECT (part));

	}
	g_object_unref (G_OBJECT (iter));
	g_object_unref (G_OBJECT (part_list));

	if (attachments) 
		*attachments = g_list_reverse(att_list);

}


static void
forward_inline (MessagesWritePageData *data, TnyHeader *header, TnyMimePart *part)
{
	TnyAccount *account = NULL;
	TnyFolder *folder;
	gchararray message_text, header_text, quoted_text;

	GtkTextIter start = {0, }, end = {0, };
	gchararray subject;

	/* Choose an identity */

	folder = tny_header_get_folder(header);	
	if (folder)
		account = tny_folder_get_account(folder);

	auto_select_identity(account, tny_header_get_to(header), data, FALSE);

	if (account)
		g_object_unref(account);
	if (folder)
		g_object_unref(folder);

	/* Headers */
	/* NOTE: This is not localized, is that OK ? */
	subject = g_strdup_printf("[Fwd: %s]", tny_header_get_subject(header));
	gtk_entry_set_text(GTK_ENTRY(data->widgets.subject), subject);
	g_free(subject);
	
	header_text = forward_mail_header(header);

	if (tny_mime_part_content_type_is (part, "message/rfc822") || tny_mime_part_content_type_is (part, "multipart/*"))
		decode_parts(part, data->text_stream, &(data->attachments));
	else
		tny_mime_part_decode_to_stream(part, data->text_stream);

	gtk_text_buffer_get_start_iter (data->text, &start);
	gtk_text_buffer_get_end_iter (data->text, &end);

	message_text = gtk_text_buffer_get_text(data->text, &start, &end, FALSE);
	
	quoted_text = quote_and_free_text(message_text, header_text);
	g_free(header_text);

	gtk_text_buffer_set_text(data->text, quoted_text, -1);
	g_free(quoted_text);
	
}

static void
forward_attachment (MessagesWritePageData *data, TnyHeader *header, TnyMimePart *part)
{
	TnyAccount *account = NULL;
	TnyFolder *folder;
	gchararray subject;

	/* Choose an identity */
	folder = tny_header_get_folder(header);	
	if (folder)
		account = tny_folder_get_account(folder);

	auto_select_identity(account, tny_header_get_to(header), data, FALSE);

	if (account)
		g_object_unref(account);
	if (folder)
		g_object_unref(folder);

	/* Headers */
	/* NOTE: This is not localized, is that OK ? */
	subject = g_strdup_printf("[Fwd: %s]", tny_header_get_subject(header));
	gtk_entry_set_text(GTK_ENTRY(data->widgets.subject), subject);
	g_free(subject);


	subject = g_strdup_printf("Fwd: %s.eml", tny_header_get_subject(header));
	tny_mime_part_set_filename(part, subject);
	g_free(subject);

	data->attachments = g_list_append(data->attachments,
					   new_mail_attachment(ATTACHMENT_INTERNAL,
							       part));
	

	
}


static void
reply (MessagesWritePageData *data, TnyHeader *header, TnyMimePart *part, gboolean all)
{
	TnyAccount *account = NULL;
	TnyFolder *folder;
	gchararray message_text, header_text, quoted_text;

	GtkTextIter start = {0, }, end = {0, };
	gchararray subject;
	const gchar *old_subject;

	/* Choose an identity */
	folder = tny_header_get_folder(header);	
	if (folder)
		account = tny_folder_get_account(folder);

	auto_select_identity(account, tny_header_get_to(header), data, FALSE);

	if (account)
		g_object_unref(account);
	if (folder)
		g_object_unref(folder);

	/* Headers */
	old_subject = tny_header_get_subject(header);
	if (strncmp("Re: ", old_subject, 4))
		subject = g_strdup_printf("Re: %s", tny_header_get_subject(header));
	else
		subject = g_strdup_printf("%s", tny_header_get_subject(header));

	gtk_entry_set_text(GTK_ENTRY(data->widgets.subject), subject);
	g_free(subject);

	gtk_entry_set_text(GTK_ENTRY(data->widgets.to), tny_header_get_from(header));
	if (all) {
		gtk_entry_set_text(GTK_ENTRY(data->widgets.cc), tny_header_get_cc(header));
	}

	header_text = reply_mail_header(header);

	if (tny_mime_part_content_type_is (part, "message/rfc822") || tny_mime_part_content_type_is (part, "multipart/*"))
		decode_parts(part, data->text_stream, NULL);
	else
		tny_mime_part_decode_to_stream(part, data->text_stream);




	gtk_text_buffer_get_start_iter (data->text, &start);
	gtk_text_buffer_get_end_iter (data->text, &end);

	

	message_text = gtk_text_buffer_get_text(data->text, &start, &end, FALSE);
	
	quoted_text = quote_and_free_text(message_text, header_text);
	g_free(header_text);

	gtk_text_buffer_set_text(data->text, quoted_text, -1);
	g_free(quoted_text);

	
}


static void
open_draft (MessagesWritePageData *data, TnyHeader *header, TnyMimePart *part)
{
	/* Choose an identity */

	/* FIXME: Currently, if no identity was found, it uses the first, not the account */
	auto_select_identity(NULL, tny_header_get_from(header), data, TRUE);
	gtk_entry_set_text(GTK_ENTRY(data->widgets.subject), tny_header_get_subject(header));
	gtk_entry_set_text(GTK_ENTRY(data->widgets.to), tny_header_get_to(header));
	gtk_entry_set_text(GTK_ENTRY(data->widgets.cc), tny_header_get_cc(header));
	gtk_entry_set_text(GTK_ENTRY(data->widgets.bcc), tny_header_get_bcc(header));

	
	

	if (tny_mime_part_content_type_is (part, "message/rfc822") || tny_mime_part_content_type_is (part, "multipart/*"))
		decode_parts(part, data->text_stream, &(data->attachments));
	else
		tny_mime_part_decode_to_stream(part, data->text_stream);
	
}


static void
cb_remove_attachment(GtkButton* button, MessagesWritePageData *data)
{
	
	GtkWidget *box;
	MailAttachment *attachment;
	g_object_get(G_OBJECT(button), "user-data", &box, NULL);
	g_object_get(G_OBJECT(box), "user-data", &attachment, NULL);
	
	data->attachments = g_list_remove(data->attachments, attachment);

	free_mail_attachment(attachment, NULL);
	gtk_widget_destroy(box);
}

static void
display_attachment (MailAttachment *attachment, MessagesWritePageData *data)
{
	char buf[32];
	const gchar *unit, *filename;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *button;
	PangoAttrList *bold;
	PangoAttribute *attr;

	/* Bold attribute */
	bold = pango_attr_list_new();	
	attr = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
	attr->start_index = 0;
	attr->end_index = G_MAXUINT;
	pango_attr_list_insert(bold, attr);

	/* We carry a pointer to the attachment element with the hbox. */
	hbox = gtk_hbox_new(FALSE, 3);
	g_object_set(G_OBJECT(hbox), "user-data", attachment, NULL);
	

	button = gtk_button_new();
	gtk_button_set_image(GTK_BUTTON(button), gtk_image_new_from_icon_name (GTK_STOCK_DELETE, GTK_ICON_SIZE_MENU));
	/* We carry a pointer to the hbox with the button */
	g_object_set(G_OBJECT(button), "user-data", hbox, NULL);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(cb_remove_attachment), data);
	gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, TRUE, 0);

	switch (attachment->type) {
		case ATTACHMENT_INTERNAL:
			if (tny_mime_part_is_attachment(attachment->part))
				filename = tny_mime_part_get_filename(attachment->part);
			else
				/* Well, I don't know what should we display here ... */
				filename = tny_mime_part_get_content_id (attachment->part);
			break;
		case ATTACHMENT_FILE:
			filename = attachment->filename;
			break;
		default:
			filename = "";
	}
	

	label = moko_field_label_new(filename);
	gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);
	gtk_label_set_attributes(GTK_LABEL(label), bold);
	gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
	
	
	if (attachment->type == ATTACHMENT_FILE) {
		snprintf(buf, sizeof(buf), "%.2f %s", attachment->size / (double)get_size_display_factor(attachment->size, &unit), unit);
		label = gtk_label_new(buf);
		gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, TRUE, 0);
	}
	
	gtk_widget_show_all(hbox);
	gtk_box_pack_start (GTK_BOX (data->widgets.attachment_box), hbox, FALSE, TRUE, 0);

	
	
}

static void
cb_notebook_prev(GtkButton *button, GtkNotebook *notebook)
{
	gtk_notebook_prev_page(notebook);
}

static void
cb_notebook_next(GtkButton *button, GtkNotebook *notebook)
{
	gtk_notebook_next_page(notebook);
}



static void
cb_attachment_toggle(GtkToggleToolButton *button, GtkNotebook *notebook)
{
	if (gtk_toggle_tool_button_get_active(button))
		gtk_notebook_next_page(notebook);
	else
		gtk_notebook_prev_page(notebook);
}

static void
draw_attachments(MessagesWritePageData *data)
{
	g_list_foreach(data->attachments, 
		       (GFunc) display_attachment,
		       data);
	gtk_widget_show_all(data->widgets.attachment_box);
}

static void
cb_attach (GtkButton *button, MessagesWritePageData *data)
{
	gint response;
	gchar *path;
	
	/* TODO: Implement a file chooser */
	response = text_entry_dialog (_("Enter the path to the file to attach"), &path, "/media/card/");
	if (!g_file_test(path, G_FILE_TEST_IS_REGULAR)) {
		GtkWidget *dialog;
		gint response;

		quick_dialog_run(dialog, response, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
		 		 _("File does not exist: %s\n"), path);
		g_free(path);
		return;
	}

	MailAttachment *attachment;
	const gchar *filename = strrchr(path, '/');
			
	if (!filename)
		filename = path;
	else
		filename = filename + 1;
	
	attachment = new_mail_attachment(ATTACHMENT_FILE, path, filename);
	data->attachments = g_list_append(data->attachments, attachment);	  
	display_attachment(attachment, data);
	g_free(path);
}

GtkWidget*
create_little_header_table(const gchar *label1, GtkWidget *value1, 
                           const gchar *label2, GtkWidget *value2)
{
	GtkWidget *table, *label_widget;
	static PangoAttrList *bold = NULL;
	PangoAttribute *attr;

	/* We initialise the bold attribute that we will use. */	
	if (!bold) {
		bold = pango_attr_list_new();	
		attr = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
		attr->start_index = 0;
		attr->end_index = G_MAXUINT;
		pango_attr_list_insert(bold, attr);
	}


	/* The table */
	table = gtk_table_new (2, 2, FALSE);
	gtk_table_set_row_spacings(GTK_TABLE(table), 0);
	gtk_table_set_col_spacings(GTK_TABLE(table), 0);
	

	label_widget = gtk_label_new(label1);	
	gtk_label_set_attributes(GTK_LABEL(label_widget), bold);
	gtk_misc_set_alignment(GTK_MISC(label_widget), 1.0f, 0.5f);

	gtk_table_attach(GTK_TABLE(table), label_widget, 0, 1, 0, 1, GTK_FILL | GTK_SHRINK, GTK_FILL | GTK_EXPAND, 3, 3);
	gtk_table_attach(GTK_TABLE(table), value1, 1, 2, 0, 1, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 3, 3);
	
	label_widget = gtk_label_new(label2);	
	gtk_label_set_attributes(GTK_LABEL(label_widget), bold);
	gtk_misc_set_alignment(GTK_MISC(label_widget), 1.0f, 0.5f);

	gtk_table_attach(GTK_TABLE(table), label_widget, 0, 1, 1, 2, GTK_FILL | GTK_SHRINK, GTK_FILL | GTK_EXPAND, 3, 3);
	gtk_table_attach(GTK_TABLE(table), value2, 1, 2, 1, 2, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 3, 3);


	//pango_attr_list_unref(bold);


	return table;

	
}



GtkWidget*
messages_create_write_email_page (AppData* app_data, gint action, TnyHeader *header, TnyMsg *msg)
{
	
	GtkWidget *page, *toolbar, *scroll, *table, *text;
	GtkWidget *notebook, *vbox, *hbox, *button, *viewport;
	GtkCellRenderer  *renderer;

	/* We initialise the data */

	MessagesWritePageData *data = g_malloc0(sizeof(MessagesWritePageData));
	data->text = gtk_text_buffer_new(NULL);
	data->text_stream = tny_gtk_text_buffer_stream_new (data->text);
	data->attachments = NULL;
	data->app_data = app_data;




	/* Initialising widgets */
	page = gtk_vbox_new (FALSE, 0);

	toolbar = gtk_toolbar_new ();
	toolbar_insert_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (GTK_STOCK_CLOSE, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_message_close), data);
	toolbar_insert_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (MOKO_STOCK_MAIL_SEND, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_send_email), data);
	toolbar_insert_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (GTK_STOCK_SAVE, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_save_draft), data);
	
	gtk_box_pack_start (GTK_BOX (page), toolbar, FALSE, TRUE, 0);

	/* We create a vertical notebook with the different input fields */
	/* We place the notebook in a hbox with two buttons right to it, for
	   switching */

	hbox = gtk_hbox_new(FALSE, 0);
	notebook = gtk_notebook_new();
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
	gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);


	/* 1. Account and identity page */
	data->widgets.from = gtk_combo_box_new_with_model(data->app_data->identities);

	renderer = gtk_cell_renderer_text_new();
	g_object_set(G_OBJECT(renderer), "ellipsize", PANGO_ELLIPSIZE_END, NULL);
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (data->widgets.from), renderer, TRUE);
	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (data->widgets.from),
	                                renderer,
	                                "text", IDENTITIES_COLUMN_NAME,
	                                NULL);


	g_signal_connect (G_OBJECT(data->widgets.from), "changed", G_CALLBACK(cb_id_select), data);

	data->widgets.account = gtk_combo_box_new_with_model(data->app_data->transport_accounts);
	
	renderer = gtk_cell_renderer_text_new();
	g_object_set(G_OBJECT(renderer), "ellipsize", PANGO_ELLIPSIZE_END, NULL);
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (data->widgets.account), renderer, TRUE);
	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (data->widgets.account),
	                                renderer,
	                                "text", TNY_GTK_ACCOUNT_LIST_MODEL_NAME_COLUMN,
	                                NULL);

	
	
	table = create_little_header_table(_("From:"), data->widgets.from,
	                                   _("Via:"), data->widgets.account);


	gtk_widget_show_all(table);
	gtk_notebook_append_page (GTK_NOTEBOOK(notebook), table, NULL);
	

	/* 2. To and subject (main) page */
	data->widgets.to = gtk_entry_new();
	data->widgets.subject = gtk_entry_new();

	table = create_little_header_table(_("To:"), data->widgets.to,
	                                   _("Subject:"), data->widgets.subject);

	gtk_widget_show_all(table);
	gtk_notebook_append_page (GTK_NOTEBOOK(notebook), table, NULL);
	

	/* 3. Cc and bcc page */

	data->widgets.cc = gtk_entry_new();
	data->widgets.bcc = gtk_entry_new();

	table = create_little_header_table(_("Cc:"), data->widgets.cc,
	                                   _("Bcc:"), data->widgets.bcc);

	gtk_widget_show_all(table);
	gtk_notebook_append_page (GTK_NOTEBOOK(notebook), table, NULL);



	/* The default page is the page with the From and Subject fields */
	gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook), 1);

	gtk_box_pack_start (GTK_BOX (hbox), notebook, TRUE, TRUE, 0);


	
	/* Initialising data */

	if (action) {
		if (header && msg)
			switch (action) {
		
				case MESSAGE_WRITE_REPLY_ALL:
					reply(data, header, TNY_MIME_PART(msg), TRUE);
					break;
				case MESSAGE_WRITE_REPLY_SENDER:
					reply(data, header, TNY_MIME_PART(msg), FALSE);
					break;
				case MESSAGE_WRITE_FORWARD_INLINE:
					forward_inline(data, header, TNY_MIME_PART(msg));
					break;
				case MESSAGE_WRITE_FORWARD_ATTACHMENT:
					forward_attachment(data, header, TNY_MIME_PART(msg));
					break;
				case MESSAGE_WRITE_OPEN_DRAFT:
					open_draft(data, header, TNY_MIME_PART(msg));
					data->draft_folder = tny_header_get_folder(header);
					data->draft_header = g_object_ref(header);
				default:
					auto_select_identity(NULL, NULL, data, FALSE);
			}
		else
			g_warning("The program tried to forward or reply to a message, without specific one.");
	}
	else
		auto_select_identity(NULL, NULL, data, FALSE);

	/* Notebook switching */
	vbox = gtk_vbox_new(FALSE, 0);
	button = gtk_button_new();
	gtk_button_set_image(GTK_BUTTON(button), 
			     gtk_image_new_from_icon_name (GTK_STOCK_GO_UP,
							   GTK_ICON_SIZE_MENU));
	gtk_button_set_focus_on_click (GTK_BUTTON(button), FALSE);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(cb_notebook_prev), notebook);
	gtk_box_pack_start (GTK_BOX (vbox), button, TRUE, TRUE, 0);
		
	button = gtk_button_new();
	gtk_button_set_image(GTK_BUTTON(button), 
			     gtk_image_new_from_icon_name (GTK_STOCK_GO_DOWN,
							   GTK_ICON_SIZE_MENU));
	gtk_button_set_focus_on_click (GTK_BUTTON(button), FALSE);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(cb_notebook_next), notebook);
	gtk_box_pack_start (GTK_BOX (vbox), button, TRUE, TRUE, 0);
	
	gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (page), hbox, FALSE, TRUE, 0);
	


	/* We also create a notebook with attachment and message text and 
	   insert a toolbar toggle for it */


	notebook = gtk_notebook_new();
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
	gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);

	toolbar_insert_toggle_button (GTK_TOOLBAR(toolbar), gtk_image_new_from_icon_name (GREP_FOR_THIS_ATTACHMENT, GTK_ICON_SIZE_LARGE_TOOLBAR), G_CALLBACK(cb_attachment_toggle), notebook);

	/* 1. The message */
	text = gtk_text_view_new_with_buffer(data->text);
	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(text), GTK_WRAP_WORD_CHAR);
	gtk_container_set_border_width(GTK_CONTAINER(text), 1);
	

	scroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scroll), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);

	gtk_container_add (GTK_CONTAINER (scroll), text);
	gtk_notebook_append_page (GTK_NOTEBOOK(notebook), scroll, NULL);
	
	/* 2. The attachments */
	
	vbox = gtk_vbox_new(FALSE, 0);
	/* We create an attachment box, where they are saved. */
	data->widgets.attachment_box = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), data->widgets.attachment_box, TRUE, TRUE, 0);
	draw_attachments(data);
	
	hbox = gtk_hbox_new(FALSE, 0);
	button = gtk_button_new_with_label(_("Attach..."));
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(cb_attach), data);

	gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, TRUE, 0);

	viewport = gtk_viewport_new(NULL, NULL);
	gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport), GTK_SHADOW_NONE);

	scroll = moko_finger_scroll_new();
	gtk_container_add (GTK_CONTAINER (viewport), vbox);	
	gtk_container_add (GTK_CONTAINER (scroll), viewport);
	gtk_notebook_append_page (GTK_NOTEBOOK(notebook), scroll, NULL);

	gtk_box_pack_end (GTK_BOX (page), notebook, TRUE, TRUE, 0);

	data->page = page;

	gtk_widget_show_all (page);

	return page;
}


struct delayed_page_data {
	AppData *app_data;
	gint action;
};

static void
cb_status_write_page (GObject *sender, TnyStatus *status, struct delayed_page_data *data)
{
	status_update(sender, status, data->app_data);
}

static void
pop_email_page (TnyFolder *folder, gboolean canceled, TnyMsg *msg, GError *err, struct delayed_page_data *data)
{
	TnyHeader *header;
	AppData *app_data = data->app_data;
	gint action = data->action;

	g_free(data);

	GtkWidget *w;
	gint r;

	const gchar *action_string;
	
	switch (action) {
		case MESSAGE_WRITE_REPLY_ALL:
		case MESSAGE_WRITE_REPLY_SENDER:
			action_string = N_("answer");
			break;
		case MESSAGE_WRITE_FORWARD_INLINE:
		case MESSAGE_WRITE_FORWARD_ATTACHMENT:
			action_string = N_("forward");
			break;
		default:
			action_string = N_("play with");
	}
	
	

	if (err) {
		quick_dialog_run(w, r, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
			 	 _("Could not open the message to %s: %s\n"), _(action_string), err->message);
		return;
	}

	
	if (!canceled) {
		header = tny_msg_get_header(msg);
		w = messages_create_write_email_page(app_data, action, header, msg);
		r = notebook_add_page_with_icon (app_data->notebook, w, "mail-message-new", 0);
		g_object_unref(header);
	}

}

void
messages_create_write_email_page_no_msg (AppData* app_data, gint action, TnyFolder *folder, TnyHeader *header)
{
	struct delayed_page_data *data;
	data = g_malloc(sizeof(*data));
	data->app_data = app_data;
	data->action = action;
	tny_folder_get_msg_async (folder, header, (TnyGetMsgCallback)pop_email_page,
				  (TnyStatusCallback) cb_status_write_page, data);
}
