/* 
 * Hold Menu Tool Button for GTK
 * Copyright (C) 2003 Ricardo Fernandez Pascual
 * Copyright (C) 2004 Paolo Borelli
 * Copyright (C) 2007 Milko Krachounov <exabyte@3mhz.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* This is a subclass of GtkToolButton, that is similar to GtkMenuToolButton, with the difference
   that the menu pops up on 

   To unmokovise, and rename it to "something" you can use:
          sed -i.bak '{s,moko,something,g;s,Moko,Something,g;s,MOKO,SOMETHING,g}' mokomenutoolbutton.{c,h}
   For example:
          sed -i.bak '{s,moko,gtk,g;s,Moko,GTK,g;s,MOKO,GTK,g}' mokomenutoolbutton.{c,h}

   TODO: Implement signals and stuff. Make more friendly.
 */

#include "menutoolbutton.h"
#include <gtk/gtkmenu.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkmain.h>
#define DEFAULT_MENU_INTERVAL 500

typedef struct _MokoMenuToolButtonPrivate MokoMenuToolButtonPrivate;
struct _MokoMenuToolButtonPrivate
{
	GtkMenu	*menu;
	guint	interval;
};

static void
menu_position_func (GtkMenu			*menu,
                    int				*x,
                    int				*y,
                    gboolean			*push_in,
                    GtkWidget	*button);
static gboolean
moko_menu_button_pressed (GtkButton *button, 
                          GdkEventButton *event,
                          MokoMenuToolButton *toolbutton);


G_DEFINE_TYPE (MokoMenuToolButton, moko_menu_tool_button, GTK_TYPE_TOOL_BUTTON)

#define MOKO_MENU_TOOL_BUTTON_GET_PRIVATE(o) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((o), MOKO_TYPE_MENU_TOOL_BUTTON, \
	 MokoMenuToolButtonPrivate))



static void
moko_menu_tool_button_class_init (MokoMenuToolButtonClass *class)
{
	g_type_class_add_private (class, sizeof (MokoMenuToolButtonPrivate));
	
}


static void
moko_menu_tool_button_init (MokoMenuToolButton *toolbutton)
{
	
	MokoMenuToolButtonPrivate *priv = MOKO_MENU_TOOL_BUTTON_GET_PRIVATE(toolbutton);
	priv->menu = NULL;
	priv->interval = DEFAULT_MENU_INTERVAL;

	/* We add connect our hold handler */
	GtkWidget *button = GTK_BIN(toolbutton)->child;
	g_signal_connect (G_OBJECT(button), "button_press_event", G_CALLBACK (moko_menu_button_pressed), toolbutton);
	g_signal_connect (G_OBJECT(button), "button_release_event", G_CALLBACK (moko_menu_button_pressed), toolbutton);
}

GtkToolItem *
moko_menu_tool_button_new (GtkWidget *icon_widget, const gchar *label)
{
	MokoMenuToolButton *button;

	button = g_object_new (MOKO_TYPE_MENU_TOOL_BUTTON, NULL);
	if (label)
		gtk_tool_button_set_label (GTK_TOOL_BUTTON (button), label);

	if (icon_widget)
		gtk_tool_button_set_icon_widget (GTK_TOOL_BUTTON (button), icon_widget);

	return GTK_TOOL_ITEM (button);
}

GtkToolItem *
gtk_tool_button_new_from_stock (const gchar *stock_id)
{
	MokoMenuToolButton *button;
	g_return_val_if_fail (stock_id != NULL, NULL);

	button = g_object_new (MOKO_TYPE_MENU_TOOL_BUTTON, "stock-id", stock_id, NULL);
	return GTK_TOOL_ITEM (button);
}
GtkWidget*
moko_menu_tool_button_get_menu (MokoMenuToolButton *button)
{
	MokoMenuToolButtonPrivate *priv;
	priv = MOKO_MENU_TOOL_BUTTON_GET_PRIVATE(button);
	return GTK_WIDGET(priv->menu);
}

void
moko_menu_tool_button_set_interval	(MokoMenuToolButton *button,
					 guint interval)
{
	MokoMenuToolButtonPrivate *priv;
	priv = MOKO_MENU_TOOL_BUTTON_GET_PRIVATE(button);
	priv->interval = interval;
}

guint
moko_menu_tool_button_get_interval	(MokoMenuToolButton *button)
{
	MokoMenuToolButtonPrivate *priv;
	priv = MOKO_MENU_TOOL_BUTTON_GET_PRIVATE(button);
	return priv->interval;
}

typedef struct {
	MokoMenuToolButton *toolbutton;
	GdkEventButton event;
} HoldButtonData;

static gboolean
popup_menu (HoldButtonData *data)
{
	MokoMenuToolButtonPrivate *priv = MOKO_MENU_TOOL_BUTTON_GET_PRIVATE(data->toolbutton);

	if (!priv->menu)
		return FALSE;

	gtk_menu_popup (priv->menu, NULL, NULL,
			(GtkMenuPositionFunc) menu_position_func,
			data->toolbutton,
			//0,
			//gtk_get_current_event_time ());
			data->event.button,
			data->event.time);
	return FALSE;
}

static gboolean
moko_menu_button_pressed (GtkButton *button, 
                          GdkEventButton *event,
                          MokoMenuToolButton *toolbutton)
{
	static gint hold_timeout_source = 0;
	MokoMenuToolButtonPrivate *priv = MOKO_MENU_TOOL_BUTTON_GET_PRIVATE(toolbutton);
	if (!priv->menu)
		return FALSE;
	if (event->type == GDK_BUTTON_PRESS) {
		HoldButtonData *data;
		data = g_malloc(sizeof(HoldButtonData));
		data->toolbutton = toolbutton;
		data->event = *event;
		hold_timeout_source = g_timeout_add_full (G_PRIORITY_DEFAULT, priv->interval, (GSourceFunc) popup_menu, data, (GDestroyNotify) g_free);
	}
	else if (event->type == GDK_BUTTON_RELEASE) {
		if (hold_timeout_source != 0) {
			if (!g_source_remove (hold_timeout_source)) {

				/* We don't want a regular click, when the timeout elapsed */
				return TRUE;
			}
		}
	}
	return FALSE;
}





/* The following is copied from the GTK source  */
static void
menu_position_func (GtkMenu			*menu,
                    int				*x,
                    int				*y,
                    gboolean			*push_in,
                    GtkWidget	*button)
{
  MokoMenuToolButtonPrivate *priv = MOKO_MENU_TOOL_BUTTON_GET_PRIVATE(button);
  GtkWidget *widget = GTK_WIDGET (button);
  GtkRequisition req;
  GtkRequisition menu_req;
  GtkOrientation orientation;
  GtkTextDirection direction;
  GdkRectangle monitor;
  gint monitor_num;
  GdkScreen *screen;

  gtk_widget_size_request (GTK_WIDGET (priv->menu), &menu_req);

  orientation = gtk_tool_item_get_orientation (GTK_TOOL_ITEM (button));
  direction = gtk_widget_get_direction (widget);

  screen = gtk_widget_get_screen (GTK_WIDGET (menu));
  monitor_num = gdk_screen_get_monitor_at_window (screen, widget->window);
  if (monitor_num < 0)
    monitor_num = 0;
  gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);

  if (orientation == GTK_ORIENTATION_HORIZONTAL)
    {
      gdk_window_get_origin (widget->window, x, y);
      *x += widget->allocation.x;
      *y += widget->allocation.y;

      if (direction == GTK_TEXT_DIR_LTR)
	*x += MAX (widget->allocation.width - menu_req.width, 0);
      else if (menu_req.width > widget->allocation.width)
        *x -= menu_req.width - widget->allocation.width;

      if ((*y + button->allocation.height + menu_req.height) <= monitor.y + monitor.height)
	*y += button->allocation.height;
      else if ((*y - menu_req.height) >= monitor.y)
	*y -= menu_req.height;
      else if (monitor.y + monitor.height - (*y + button->allocation.height) > *y)
	*y += button->allocation.height;
      else
	*y -= menu_req.height;
    }
  else 
    {
      gdk_window_get_origin (GTK_BUTTON (button)->event_window, x, y);
      gtk_widget_size_request (button, &req);

      if (direction == GTK_TEXT_DIR_LTR)
        *x += button->allocation.width;
      else 
        *x -= menu_req.width;

      if (*y + menu_req.height > monitor.y + monitor.height &&
	  *y + button->allocation.height - monitor.y > monitor.y + monitor.height - *y)
	*y += button->allocation.height - menu_req.height;
    }

  *push_in = FALSE;
}



static void
menu_detacher (GtkWidget *widget,
               GtkMenu   *menu)
{
  MokoMenuToolButtonPrivate *priv = MOKO_MENU_TOOL_BUTTON_GET_PRIVATE (widget);

  g_return_if_fail (priv->menu == menu);

  priv->menu = NULL;
}

void
moko_menu_tool_button_set_menu (MokoMenuToolButton *toolbutton,
                               GtkWidget         *menu)
{
  MokoMenuToolButtonPrivate *priv;
 
  g_return_if_fail (MOKO_IS_MENU_TOOL_BUTTON (toolbutton));
  g_return_if_fail (GTK_IS_MENU (menu) || menu == NULL);


  priv = MOKO_MENU_TOOL_BUTTON_GET_PRIVATE(toolbutton);;
  GtkWidget *button = GTK_BIN(toolbutton)->child;

  if (priv->menu != GTK_MENU (menu))
    {
      if (priv->menu && GTK_WIDGET_VISIBLE (priv->menu))
        gtk_menu_shell_deactivate (GTK_MENU_SHELL (priv->menu));

      if (priv->menu)
	{
#if 0
          g_signal_handlers_disconnect_by_func (priv->menu, 
						menu_deactivate_cb, 
						button);
#endif
	  gtk_menu_detach (priv->menu);
	}

      priv->menu = GTK_MENU (menu);

      if (priv->menu)
        {
          gtk_menu_attach_to_widget (priv->menu, GTK_WIDGET (button),
                                     menu_detacher);

          gtk_widget_set_sensitive (GTK_WIDGET(button), TRUE);
#if 0
          g_signal_connect (priv->menu, "deactivate",
                            G_CALLBACK (menu_deactivate_cb), button);
#endif
        }
      else
       gtk_widget_set_sensitive (GTK_WIDGET(button), FALSE);
    }
#if 0
  g_object_notify (G_OBJECT (button), "menu");
#endif
}

